﻿// <copyright file="AuthenticationManager.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> AuthenticationManager
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Collections;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Threading;

using Eunomia.SSH2.Crypto;
using Eunomia.SSH2.Log;
using Eunomia.SSH2.Packets;
using Eunomia.SSH2.Signature;
using Eunomia.SSH2.Transport;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.Auth
{
    /// <summary>
    /// Class used to manage ssh authentication.
    /// </summary>
    /// <example>
    /// </example>
    public class AuthenticationManager : MessageHandler
    {
        private TransportManager tm;

        private ArrayList packets;
        private bool connectionClosed;

        private string banner;

        private string[] remainingMethods;
        private bool partialSuccess;

        private bool authenticated;
        private bool initDone;
        
        /// <summary>
        /// Initializes a new instance of the AuthenticationManager class.
        /// </summary>
        /// <param name="tm">This is the TransportationManager that is doing authentication.</param>
        public AuthenticationManager(TransportManager tm)
        {
            this.packets = new ArrayList(10);
            this.remainingMethods = new string[0];
            this.tm = tm;
        }
        
        /// <summary>
        /// Gets a value indicating whether the authentication had partial success.
        /// </summary>
        public bool PartialSuccess
        {
            get
            {
                return this.partialSuccess;
            }
        }
        
        /// <summary>
        /// Gets the stored banner.
        /// </summary>
        public string Banner
        {
            get
            {
                return this.banner;
            }
        }

        /// <summary>
        /// Gets the next available packet/message (byte array).
        /// </summary>
        /// <returns>The next available packet/message (byte array).</returns>
        public byte[] GetNextMessage()
        {
            while (true)
            {
                byte[] msg = this.DeQueue();

                if (msg[0] != Packet.SSHMessageUserAuthorizationBanner)
                {
                    return msg;
                }

                PacketUserauthBanner sb = new PacketUserauthBanner(msg, 0, msg.Length);
                this.banner = sb.Banner;
            }
        }

        /// <summary>
        /// Gets any remaining authentication methods.
        /// </summary>
        /// <param name="user">The user we are trying to authenticate.</param>
        /// <returns>A string array of the remaining methods.</returns>
        public string[] GetRemainingMethods(string user)
        {
            this.Initialize(user);
            return this.remainingMethods;
        }

        /// <summary>
        /// Tries to authenticate with a public key.
        /// </summary>
        /// <param name="user">The user to authenticate.</param>
        /// <param name="pemPrivateKey">The private key to use.</param>
        /// <param name="password">The password to use.</param>
        /// <param name="random">A source for random bytes.</param>
        /// <returns>True if the authentication succeeds.</returns>
        [SuppressMessage("Microsoft.Performance", "CA1800:DoNotCastUnnecessarily",
                         Justification = "FxCop is confused.  The casting is appropriate.")]
        public bool AuthenticatePublicKey(
            string user,
            char[] pemPrivateKey,
            string password,
            SecureRandom random)
        {
            try
            {
                this.Initialize(user);

                if (this.MethodPossible("publickey") == false)
                {
                    throw new IOException("Authentication method publickey not supported by the server at this stage.");
                }

                object key = PEMDecoder.Decode(pemPrivateKey, password);

                if (key is DSAPrivateKey)
                {
                    DSAPrivateKey pk = (DSAPrivateKey)key;

                    byte[] pkeyEnc = DSASHA1Verify.EncodeSSHDSAPublicKey(pk.PublicKey);

                    TypesWriter tw = new TypesWriter();

                    byte[] h = this.tm.SessionIdentifier;

                    tw.WriteString(h, 0, h.Length);
                    tw.WriteByte(Packet.SSHMessageUserAuthorizationRequest);
                    tw.WriteString(user);
                    tw.WriteString("ssh-connection");
                    tw.WriteString("publickey");
                    tw.WriteBoolean(true);
                    tw.WriteString("ssh-dss");
                    tw.WriteString(pkeyEnc, 0, pkeyEnc.Length);

                    byte[] msg = tw.GetBytes();

                    DSASignature ds = DSASHA1Verify.GenerateSignature(msg, pk, random);

                    byte[] dsaEnc = DSASHA1Verify.EncodeSSHDSASignature(ds);

                    PacketUserauthRequestPublicKey ua = new PacketUserauthRequestPublicKey("ssh-connection", user, "ssh-dss", pkeyEnc, dsaEnc);
                    this.tm.SendMessage(ua.Payload);
                }
                else if (key is RSAPrivateKey)
                {
                    RSAPrivateKey pk = (RSAPrivateKey)key;

                    byte[] pkeyEnc = RSASHA1Verify.EncodeSSHRSAPublicKey(pk.PublicKey);

                    TypesWriter tw = new TypesWriter();
                    {
                        byte[] h = this.tm.SessionIdentifier;

                        tw.WriteString(h, 0, h.Length);
                        tw.WriteByte(Packet.SSHMessageUserAuthorizationRequest);
                        tw.WriteString(user);
                        tw.WriteString("ssh-connection");
                        tw.WriteString("publickey");
                        tw.WriteBoolean(true);
                        tw.WriteString("ssh-rsa");
                        tw.WriteString(pkeyEnc, 0, pkeyEnc.Length);
                    }

                    byte[] msg = tw.GetBytes();

                    RSASignature ds = RSASHA1Verify.GenerateSignature(msg, pk);

                    byte[] rsa_sig_enc = RSASHA1Verify.EncodeSSHRSASignature(ds);

                    PacketUserauthRequestPublicKey ua = new PacketUserauthRequestPublicKey("ssh-connection", user, "ssh-rsa", pkeyEnc, rsa_sig_enc);
                    this.tm.SendMessage(ua.Payload);
                }
                else
                {
                    throw new IOException("Unknown private key type returned by the PEM decoder.");
                }

                byte[] ar = this.GetNextMessage();

                if (ar[0] == Packet.SSHMessageUserAuthorizationSuccess)
                {
                    this.authenticated = true;
                    this.tm.RemoveMessageHandler(this, 0, 255);
                    return true;
                }

                if (ar[0] == Packet.SSHMessageUserAuthorizationFailure)
                {
                    PacketUserauthFailure puf = new PacketUserauthFailure(ar, 0, ar.Length);

                    this.remainingMethods = puf.AuthThatCanContinue;
                    this.partialSuccess = puf.PartialSuccess;

                    return false;
                }

                throw new IOException("Unexpected SSH message (type " + ar[0] + ")");
            }
            catch (IOException e)
            {
                this.tm.Close(e, false);
                throw new IOException("Publickey authentication failed.  " + e.Message);
            }
        }

        /// <summary>
        /// Tries to authenticate with 'no authentication'.
        /// </summary>
        /// <param name="user">The user to authenticate.</param>
        /// <returns>True if the authentication succeeds.</returns>
        public bool AuthenticateNone(string user)
        {
            try
            {
                this.Initialize(user);
                return this.authenticated;
            }
            catch (IOException e)
            {
                this.tm.Close(e, false);
                throw new IOException("None authentication failed.  " + e.Message);
            }
        }

        /// <summary>
        /// Tries to authenticate with a password.
        /// </summary>
        /// <param name="user">The user to authenticate.</param>
        /// <param name="password">The password to use.</param>
        /// <returns>True if the authentication succeeds.</returns>
        public bool AuthenticatePassword(string user, string password)
        {
            try
            {
                this.Initialize(user);

                if (this.MethodPossible("password") == false)
                {
                    throw new IOException("Authentication method password not supported by the server at this stage.");
                }

                PacketUserauthRequestPassword ua = new PacketUserauthRequestPassword("ssh-connection", user, password);
                this.tm.SendMessage(ua.Payload);

                byte[] ar = this.GetNextMessage();

                if (ar[0] == Packet.SSHMessageUserAuthorizationSuccess)
                {
                    this.authenticated = true;
                    this.tm.RemoveMessageHandler(this, 0, 255);
                    return true;
                }

                if (ar[0] == Packet.SSHMessageUserAuthorizationFailure)
                {
                    PacketUserauthFailure puf = new PacketUserauthFailure(ar, 0, ar.Length);

                    this.remainingMethods = puf.AuthThatCanContinue;
                    this.partialSuccess = puf.PartialSuccess;

                    return false;
                }

                throw new IOException("Unexpected SSH message (type " + ar[0] + ")");
            }
            catch (IOException e)
            {
                this.tm.Close(e, false);
                throw new IOException("Password authentication failed.  " + e.Message);
            }
        }

        /// <summary>
        /// Tries to authenticate with a password.
        /// </summary>
        /// <param name="user">The user to authenticate.</param>
        /// <param name="subMethods">String array of possible sub-methods.</param>
        /// <param name="callback">Place holder.</param>
        /// <returns>True if the authentication succeeds.</returns>
        public bool AuthenticateInteractive(string user, string[] subMethods, IInteractiveCallback callback)
        {
            try
            {
                this.Initialize(user);

                if (this.MethodPossible("keyboard-interactive") == false)
                {
                    throw new IOException("Authentication method keyboard-interactive not supported by the server at this stage.");
                }

                if (subMethods == null)
                {
                    subMethods = new string[0];
                }

                PacketUserauthRequestInteractive ua = new PacketUserauthRequestInteractive("ssh-connection", user, subMethods);

                this.tm.SendMessage(ua.Payload);

                while (true)
                {
                    byte[] ar = this.GetNextMessage();

                    if (ar[0] == Packet.SSHMessageUserAuthorizationSuccess)
                    {
                        this.authenticated = true;
                        this.tm.RemoveMessageHandler(this, 0, 255);
                        return true;
                    }

                    if (ar[0] == Packet.SSHMessageUserAuthorizationFailure)
                    {
                        PacketUserauthFailure puf = new PacketUserauthFailure(ar, 0, ar.Length);

                        this.remainingMethods = puf.AuthThatCanContinue;
                        this.partialSuccess = puf.PartialSuccess;

                        return false;
                    }

                    if (ar[0] == Packet.SSHMessageUserAuthorizationInfoRequest)
                    {
                        PacketUserauthInfoRequest pui = new PacketUserauthInfoRequest(ar, 0, ar.Length);

                        string[] responses;

                        try
                        {
                            responses = callback.ReplyToChallenge(
                                pui.Name,
                                pui.Instruction,
                                pui.NumberOfPrompts,
                                pui.Prompt,
                                pui.Echo);
                        }
                        catch (Exception e)
                        {
                            throw new IOException("Exception in callback.  " + e.Message);
                        }

                        if (responses == null)
                        {
                            throw new IOException("Your callback may not return NULL!");
                        }

                        PacketUserauthInfoResponse puir = new PacketUserauthInfoResponse(responses);
                        this.tm.SendMessage(puir.Payload);

                        continue;
                    }

                    throw new IOException("Unexpected SSH message (type " + ar[0] + ")");
                }
            }
            catch (IOException e)
            {
                this.tm.Close(e, false);
                throw new IOException("Keyboard-interactive authentication failed.  " + e.Message);
            }
        }

        /// <summary>This method looks for incoming packets and stores them in the packets array.</summary>
        /// <param name="msg">The incoming message as a byte array.</param>
        /// <param name="msgLength">The length of the message (byte count).</param>
        public override void HandleMessage(byte[] msg, int msgLength)
        {
            lock (this.packets.SyncRoot)
            {
                if (msg == null)
                {
                    this.connectionClosed = true;
                }
                else
                {
                    byte[] tmp = new byte[msgLength];
                    Array.Copy(msg, 0, tmp, 0, msgLength);
                    this.packets.Add(tmp);
                }
                
                Monitor.PulseAll(this.packets.SyncRoot);
                
                if (this.packets.Count > 5)
                {
                    this.connectionClosed = true;
                    throw new IOException("Error, peer is flooding us with authentication packets.");
                }
            }
        }

        /// <summary>
        /// Search the remaining authentication methods for methName.
        /// </summary>
        /// <param name="methName">The authentication method to search for.</param>
        /// <returns>True if methName is found.</returns>
        private bool MethodPossible(string methName)
        {
            if (this.remainingMethods == null)
            {
                return false;
            }

            for (int i = 0; i < this.remainingMethods.Length; i++)
            {
                if (this.remainingMethods[i] == methName)
                {
                    return true;
                }
            }
            
            return false;
        }
        
        /// <summary>
        /// Gets the next packet to process from the packets 'queue'.
        /// </summary>
        /// <returns>The next available packet from the packets 'queue'.</returns>
        private byte[] DeQueue()
        {
            bool wasInterrupted = false;

            try
            {
                lock (this.packets.SyncRoot)
                {
                    while (this.packets.Count == 0)
                    {
                        if (this.connectionClosed)
                        {
                            throw (IOException)new IOException("The connection is closed.");
                        }

                        try
                        {
                            Monitor.Wait(this.packets.SyncRoot);
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }
                    
                    byte[] res = (byte[])this.packets[0];
                    this.packets.RemoveAt(0);
                    return res;
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }

        /// <summary>
        /// Initialize for authentication of a user.
        /// </summary>
        /// <param name="user">The user to authenticate.</param>
        /// <returns>True if the user is authenticated.</returns>
        private bool Initialize(string user)
        {
            if (this.initDone == false)
            {
                this.tm.RegisterMessageHandler(this, 0, 255);

                PacketServiceRequest sr = new PacketServiceRequest("ssh-userauth");
                this.tm.SendMessage(sr.Payload);

                byte[] msg = this.GetNextMessage();

                // TODO: this looks very odd, but is a duplicate of the java code.  Removed!
                // new PacketServiceAccept(msg, 0, msg.Length);
                PacketUserauthRequestNone urn = new PacketUserauthRequestNone("ssh-connection", user);
                this.tm.SendMessage(urn.Payload);

                msg = this.GetNextMessage();

                this.initDone = true;

                if (msg[0] == Packet.SSHMessageUserAuthorizationSuccess)
                {
                    this.authenticated = true;
                    this.tm.RemoveMessageHandler(this, 0, 255);
                    return true;
                }

                if (msg[0] == Packet.SSHMessageUserAuthorizationFailure)
                {
                    PacketUserauthFailure puf = new PacketUserauthFailure(msg, 0, msg.Length);

                    this.remainingMethods = puf.AuthThatCanContinue;
                    this.partialSuccess = puf.PartialSuccess;
                    return false;
                }

                throw new IOException("Unexpected SSH message (type " + msg[0] + ")");
            }
            
            return this.authenticated;
        }
    }
}
