﻿// <copyright file="ChannelManager.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> ChannelManager
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Collections;
using System.IO;
using System.Text;
using System.Threading;

using Eunomia.SSH2.Log;
using Eunomia.SSH2.Packets;
using Eunomia.SSH2.Transport;

namespace Eunomia.SSH2.Channel
{
    /// <summary>
    /// This class manages the ssh channels over an ssh connection.
    /// </summary>
    /// <remarks>
    /// Please read the comments in Channel.cs.
    /// Besides the crypto part, this is the core of the library.
    /// </remarks>
    public class ChannelManager : MessageHandler
    {
        private static Logger log = Logger.GetLogger(typeof(ChannelManager));
        
        private Hashtable x11MagicCookies;
        
        private TransportManager tm;
        
        private ArrayList channels;
        private int nextLocalChannel = 100;
        private bool shutdown = false;
        private int globalSuccessCounter = 0;
        private int globalFailedCounter = 0;
        
        private Hashtable remoteForwardings;
        
        private ArrayList listenerThreads;
        
        private bool listenerThreadsAllowed = true;

        /// <summary>
        /// Initializes a new instance of the ChannelManager class.
        /// </summary>
        /// <param name="tm">The TransportManager that this ChannelManager belongs to.</param>
        public ChannelManager(TransportManager tm)
        {
            this.x11MagicCookies = new Hashtable();
            this.channels = new ArrayList(10);
            this.remoteForwardings = new Hashtable();
            this.listenerThreads = new ArrayList(10);
            this.tm = tm;
            tm.RegisterMessageHandler(this, 80, 100);
        }
        
        /// <summary>
        /// Wait until a condition.
        /// </summary>
        /// <param name="channel">The channel to use.</param>
        /// <param name="timeout">The timeout to use in milliseconds, 0 means no timeout.</param>
        /// <param name="conditionMask">The minimum event mask.</param>
        /// <returns>Returns the current condition.</returns>
        public static int WaitForCondition(SSHChannel channel, long timeout, int conditionMask)
        {
            bool wasInterrupted = false;
            
            try
            {
                long end_time = 0;
                bool end_time_set = false;
                
                lock (channel)
                {
                    while (true)
                    {
                        int current_cond = 0;
                        
                        int stdoutAvail = channel.StandardOutputWritePosition - channel.StandardOutputReadPosition;
                        int stderrAvail = channel.StandardErrorWritePosition - channel.StandardErrorReadPosition;
                        
                        if (stdoutAvail > 0)
                        {
                            current_cond = current_cond | ChannelCondition.StandardOutData;
                        }
                        
                        if (stderrAvail > 0)
                        {
                            current_cond = current_cond | ChannelCondition.StandardErrorData;
                        }
                        
                        if (channel.EOF)
                        {
                            current_cond = current_cond | ChannelCondition.EOF;
                        }
                        
                        current_cond = current_cond | ChannelCondition.ExitStatus;
                        current_cond = current_cond | ChannelCondition.ExitSignal;
                        
                        if (channel.State == SSHChannel.STATE_CLOSED)
                        {
                            return current_cond | ChannelCondition.Closed | ChannelCondition.EOF;
                        }
                        
                        if ((current_cond & conditionMask) != 0)
                        {
                            return current_cond;
                        }
                        
                        if (timeout > 0)
                        {
                            if (!end_time_set)
                            {
                                // divide ticks by 10000 to get milliseconds
                                end_time = (DateTime.Now.Ticks / 10000) + timeout;
                                end_time_set = true;
                            }
                            else
                            {
                                timeout = end_time - (DateTime.Now.Ticks / 10000);
                                
                                if (timeout <= 0)
                                {
                                    return current_cond | ChannelCondition.Timeout;
                                }
                            }
                        }
                        
                        try
                        {
                            if (timeout > 0)
                            {
                                Monitor.Wait(channel);
                            }
                            else
                            {
                                Monitor.Wait(channel);
                            }
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }
        
        /// <summary>
        /// Get the number of characters that are waiting to be processed.
        /// </summary>
        /// <param name="channel">The channel to use.</param>
        /// <param name="extended">This is true for extended data and false for regular.</param>
        /// <returns>The number of waiting characters.</returns>
        public static int GetAvailable(SSHChannel channel, bool extended)
        {
            lock (channel)
            {
                int avail;
                
                if (extended)
                {
                    avail = channel.StandardErrorWritePosition - channel.StandardErrorReadPosition;
                }
                else
                {
                    avail = channel.StandardOutputWritePosition - channel.StandardOutputReadPosition;
                }
                
                return avail;
            }
        }

        /// <summary>
        /// Register a 'fake' X11 cookie.
        /// </summary>
        /// <param name="hexFakeCookie">String that is the key for the X11 cookie.</param>
        /// <param name="data">The X11 cookie data.</param>
        public void RegisterX11Cookie(string hexFakeCookie, X11ServerData data)
        {
            lock (this.x11MagicCookies.SyncRoot)
            {
                this.x11MagicCookies[hexFakeCookie] = data;
            }
        }
        
        /// <summary>
        /// Unregister a 'fake' X11 cookie, with option to kill all associated channels.
        /// </summary>
        /// <param name="hexFakeCookie">String that is the key for the X11 cookie.</param>
        /// <param name="killChannels">True if the channels associated with the cookie should be closed.</param>
        public void UnregisterX11Cookie(string hexFakeCookie, bool killChannels)
        {
            if (hexFakeCookie == null)
            {
                throw new ArgumentException("hexFakeCookie may not be null");
            }
            
            lock (this.x11MagicCookies.SyncRoot)
            {
                this.x11MagicCookies.Remove(hexFakeCookie);
            }
            
            if (killChannels == false)
            {
                return;
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Closing all X11 channels for the given fake cookie");
            }
            
            ArrayList channel_copy;
            
            lock (this.channels.SyncRoot)
            {
                channel_copy = (ArrayList)this.channels.Clone();
            }
            
            for (int i = 0; i < channel_copy.Count; i++)
            {
                SSHChannel c = (SSHChannel)channel_copy[i];
                
                lock (c)
                {
                    if (hexFakeCookie.Equals(c.HexX11FakeCookie) == false)
                    {
                        continue;
                    }
                }
                
                try
                {
                    this.CloseChannel(c, "Closing X11 channel since the corresponding session is closing", true);
                }
                catch (IOException)
                {
                }
            }
        }
        
        /// <summary>
        /// Gets the X11 cookie data for a given X11 cookie.
        /// </summary>
        /// <param name="hexFakeCookie">String that is the key for the X11 cookie.</param>
        /// <returns>The X11 cookie data.</returns>
        public X11ServerData CheckX11Cookie(string hexFakeCookie)
        {
            lock (this.x11MagicCookies)
            {
                if (hexFakeCookie != null)
                {
                    return (X11ServerData)this.x11MagicCookies[hexFakeCookie];
                }
            }
            
            return null;
        }
        
        /// <summary>
        /// Closes all channels.
        /// </summary>
        public void CloseAllChannels()
        {
            if (log.Enabled)
            {
                log.Log(50, "Closing all channels");
            }
            
            ArrayList channel_copy;
            
            lock (this.channels.SyncRoot)
            {
                channel_copy = (ArrayList)this.channels.Clone();
            }
            
            for (int i = 0; i < channel_copy.Count; i++)
            {
                SSHChannel c = (SSHChannel)channel_copy[i];
                try
                {
                    this.CloseChannel(c, "Closing all channels", true);
                }
                catch (IOException)
                {
                }
            }
        }
        
        /// <summary>
        /// Closes a specific channel.
        /// </summary>
        /// <param name="channel">The channel that we want to close.</param>
        /// <param name="reason">The close reason.</param>
        /// <param name="force">True if we need to force the channel closed.</param>
        public void CloseChannel(SSHChannel channel, string reason, bool force)
        {
            byte[] msg = new byte[5];
            
            lock (channel)
            {
                if (force)
                {
                    channel.State = SSHChannel.STATE_CLOSED;
                    channel.EOF = true;
                }
                
                channel.ReasonClosed = reason;
                
                msg[0] = (byte)Packet.SSHMessageChannelClose;
                unchecked
                {
                    msg[1] = (byte)(channel.RemoteId >> 24);
                    msg[2] = (byte)(channel.RemoteId >> 16);
                    msg[3] = (byte)(channel.RemoteId >> 8);
                    msg[4] = (byte)channel.RemoteId;
                }
                
                Monitor.PulseAll(channel);
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent == true)
                {
                    return;
                }
                
                this.tm.SendMessage(msg);
                channel.CloseMessageSent = true;
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Sent SSHMessageCHANNEL_CLOSE (channel " + channel.LocalId + ")");
            }
        }
        
        /// <summary>
        /// Send channel EOF/closed message.
        /// </summary>
        /// <param name="channel">The channel we are closing.</param>
        public void SendEOF(SSHChannel channel)
        {
            byte[] msg = new byte[5];
            
            lock (channel)
            {
                if (channel.State != SSHChannel.STATE_OPEN)
                {
                    return;
                }
                
                msg[0] = (byte)Packet.SSHMessageChannelEOF;
                msg[1] = (byte)(channel.RemoteId >> 24);
                msg[2] = (byte)(channel.RemoteId >> 16);
                msg[3] = (byte)(channel.RemoteId >> 8);
                msg[4] = (byte)channel.RemoteId;
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent == true)
                {
                    return;
                }
                
                this.tm.SendMessage(msg);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Sent EOF (Channel " + channel.LocalId + "/" + channel.RemoteId + ")");
            }
        }
        
        /// <summary>
        /// Send an open confirmation for a channel.
        /// </summary>
        /// <param name="channel">The channel to send the confirmation.</param>
        public void SendOpenConfirmation(SSHChannel channel)
        {
            PacketChannelOpenConfirmation pcoc = null;
            
            lock (channel)
            {
                if (channel.State != SSHChannel.STATE_OPENING)
                {
                    return;
                }
                
                channel.State = SSHChannel.STATE_OPEN;
                
                pcoc = new PacketChannelOpenConfirmation(channel.RemoteId, channel.LocalId, channel.LocalWindow, channel.LocalMaxPacketSize);
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent == true)
                {
                    return;
                }
                
                this.tm.SendMessage(pcoc.Payload);
            }
        }
        
        /// <summary>
        /// Send data on a channel.
        /// </summary>
        /// <param name="channel">The channel to send the data on.</param>
        /// <param name="buffer">The buffer that contains the byte data to send.</param>
        /// <param name="pos">The beginning zero based location to start sending from.</param>
        /// <param name="len">The length of the data to send.</param>
        public void SendData(SSHChannel channel, byte[] buffer, int pos, int len)
        {
            bool wasInterrupted = false;
            
            try
            {
                while (len > 0)
                {
                    int thislen = 0;
                    byte[] msg;
                    
                    lock (channel)
                    {
                        while (true)
                        {
                            if (channel.State == SSHChannel.STATE_CLOSED)
                            {
                                throw new IOException("SSH channel is closed. (" + channel.ReasonClosed + ")");
                            }
                            
                            if (channel.State != SSHChannel.STATE_OPEN)
                            {
                                throw new IOException("SSH channel in strange state. (" + channel.State + ")");
                            }
                            
                            if (channel.RemoteWindow != 0)
                            {
                                break;
                            }
                            
                            try
                            {
                                Monitor.Wait(channel);
                            }
                            catch (ThreadInterruptedException)
                            {
                                wasInterrupted = true;
                            }
                        }
                        
                        // len > 0, no sign extension can happen when comparing
                        thislen = (channel.RemoteWindow >= len) ? len : (int)channel.RemoteWindow;
                        
                        int estimatedMaxDataLen = channel.RemoteMaxPacketSize
                            - (this.tm.PacketOverheadEstimate + 9);
                        
                        // The worst case scenario =) a true bottleneck
                        if (estimatedMaxDataLen <= 0)
                        {
                            estimatedMaxDataLen = 1;
                        }
                        
                        if (thislen > estimatedMaxDataLen)
                        {
                            thislen = estimatedMaxDataLen;
                        }
                        
                        channel.RemoteWindow -= thislen;
                        
                        msg = new byte[1 + 8 + thislen];
                        
                        msg[0] = (byte)Packet.SSHMessageChannelData;
                        unchecked
                        {
                            msg[1] = (byte)(channel.RemoteId >> 24);
                            msg[2] = (byte)(channel.RemoteId >> 16);
                            msg[3] = (byte)(channel.RemoteId >> 8);
                            msg[4] = (byte)channel.RemoteId;
                            msg[5] = (byte)(thislen >> 24);
                            msg[6] = (byte)(thislen >> 16);
                            msg[7] = (byte)(thislen >> 8);
                            msg[8] = (byte)thislen;
                        }
                        
                        Array.Copy(buffer, pos, msg, 9, thislen);
                    }
                    
                    lock (channel.ChannelSendLock)
                    {
                        if (channel.CloseMessageSent == true)
                        {
                            throw new IOException("SSH channel is closed. (" + channel.ReasonClosed + ")");
                        }
                        
                        this.tm.SendMessage(msg);
                    }
                    
                    pos += thislen;
                    len -= thislen;
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }
        
        /// <summary>
        /// Request a global forward on a channel.
        /// </summary>
        /// <param name="bindAddress">The address to forward.</param>
        /// <param name="bindPort">The port to forward.</param>
        /// <param name="targetAddress">The target address.</param>
        /// <param name="targetPort">The target port.</param>
        /// <returns>Returns the port to forward if it succeeds.</returns>
        public int RequestGlobalForward(string bindAddress, int bindPort, string targetAddress, int targetPort)
        {
            RemoteForwardingData rfd = new RemoteForwardingData();
            
            rfd.BindAddress = bindAddress;
            rfd.BindPort = bindPort;
            rfd.TargetAddress = targetAddress;
            rfd.TargetPort = targetPort;
            
            lock (this.remoteForwardings.SyncRoot)
            {
                int key = bindPort;
                
                if (this.remoteForwardings[key] != null)
                {
                    throw new IOException("There is already a forwarding for remote port " + bindPort);
                }
                
                this.remoteForwardings[key] = rfd;
            }
            
            lock (this.channels.SyncRoot)
            {
                this.globalSuccessCounter = this.globalFailedCounter = 0;
            }
            
            PacketGlobalForwardRequest pgf = new PacketGlobalForwardRequest(true, bindAddress, bindPort);
            this.tm.SendMessage(pgf.Payload);
            
            if (log.Enabled)
            {
                log.Log(50, "Requesting a remote forwarding ('" + bindAddress + "', " + bindPort + ")");
            }
            
            try
            {
                this.WaitForGlobalSuccessOrFailure();
            }
            catch (IOException e)
            {
                lock (this.remoteForwardings)
                {
                    this.remoteForwardings.Remove(rfd);
                }
                
                throw new IOException(e.Message);
            }
            
            return bindPort;
        }
        
        /// <summary>
        /// Cancel a global forward.
        /// </summary>
        /// <param name="bindPort">The bind port for the forward.</param>
        public void RequestCancelGlobalForward(int bindPort)
        {
            RemoteForwardingData rfd;
            
            lock (this.remoteForwardings)
            {
                try
                {
                    rfd = (RemoteForwardingData)this.remoteForwardings[bindPort];
                }
                catch (NullReferenceException)
                {
                    throw new IOException("Sorry, there is no known remote forwarding for remote port " + bindPort);
                }
            }
            
            lock (this.channels)
            {
                this.globalSuccessCounter = this.globalFailedCounter = 0;
            }
            
            PacketGlobalCancelForwardRequest pgcf = new PacketGlobalCancelForwardRequest(
                true,
                rfd.BindAddress,
                rfd.BindPort);
            
            this.tm.SendMessage(pgcf.Payload);
            
            if (log.Enabled)
            {
                log.Log(50, "Requesting cancelation of remote forward ('" + rfd.BindAddress + "', " + rfd.BindPort + ")");
            }
            
            this.WaitForGlobalSuccessOrFailure();
            
            /* Only now we are sure that no more forwarded connections will arrive */
            
            lock (this.remoteForwardings)
            {
                this.remoteForwardings.Remove(rfd);
            }
        }
        
        /// <summary>
        /// Register a listener thread so that we can call StopWorking on them when we close.
        /// </summary>
        /// <param name="thread">The thread to register.</param>
        public void RegisterThread(IChannelWorkerThread thread)
        {
            lock (this.listenerThreads)
            {
                if (this.listenerThreadsAllowed == false)
                {
                    throw new IOException("Too late, this connection is closed.");
                }
                
                this.listenerThreads.Add(thread);
            }
        }
        
        /// <summary>
        /// Open a direct TCP/IP channel.
        /// </summary>
        /// <param name="hostToConnect">Host to connect to.</param>
        /// <param name="portToConnect">Port to connect to.</param>
        /// <param name="originatorIPAddress">Originator IP address.</param>
        /// <param name="originatorPort">Originator port.</param>
        /// <returns>The Channel that was openned.</returns>
        public SSHChannel OpenDirectTCPIPChannel(
            string hostToConnect,
            int portToConnect,
            string originatorIPAddress,
            int originatorPort)
        {
            SSHChannel c = new SSHChannel(this);
            
            lock (c)
            {
                c.LocalId = this.AddChannel(c);
                
                // end of lock block forces writing out to main memory
            }
            
            PacketOpenDirectTCPIPChannel dtc = new PacketOpenDirectTCPIPChannel(
                c.LocalId,
                c.LocalWindow,
                c.LocalMaxPacketSize,
                hostToConnect,
                portToConnect,
                originatorIPAddress,
                originatorPort);
            
            this.tm.SendMessage(dtc.Payload);
            
            this.WaitUntilChannelOpen(c);
            
            return c;
        }
        
        /// <summary>
        /// Send a request to open a new Channel.
        /// </summary>
        /// <returns>Returns the openned Channel.</returns>
        public SSHChannel OpenSessionChannel()
        {
            SSHChannel c = new SSHChannel(this);
            
            lock (c)
            {
                c.LocalId = this.AddChannel(c);
                
                // end of lock block forces the writing out to main memory
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Sending SSHMessageCHANNEL_OPEN (Channel " + c.LocalId + ")");
            }
            
            PacketOpenSessionChannel smo = new PacketOpenSessionChannel(c.LocalId, c.LocalWindow, c.LocalMaxPacketSize);
            this.tm.SendMessage(smo.Payload);
            
            this.WaitUntilChannelOpen(c);
            
            return c;
        }
        
        /// <summary>
        /// Request a Pseudo terminal on a particular channel.
        /// </summary>
        /// <param name="channel">The channel to use.</param>
        /// <param name="term">The terminal type (dumb).</param>
        /// <param name="termWidthCharacters">The terminal width in characters.</param>
        /// <param name="termHeightCharacters">The terminal height in characters.</param>
        /// <param name="termWidthPixels">The terminal width in pixels.</param>
        /// <param name="termHeightPixels">The terminal height in pixels.</param>
        /// <param name="terminalModes">The terminal modes, (may be null) (see ssh documentation).</param>
        public void RequestPTY(
            SSHChannel channel,
            string term,
            int termWidthCharacters,
            int termHeightCharacters,
            int termWidthPixels,
            int termHeightPixels,
            byte[] terminalModes)
        {
            PacketSessionPtyRequest spr;
            
            lock (channel)
            {
                if (channel.State != SSHChannel.STATE_OPEN)
                {
                    throw new IOException("Cannot request PTY on this channel (" + channel.ReasonClosed + ")");
                }
                
                spr = new PacketSessionPtyRequest(
                    channel.RemoteId,
                    true,
                    term,
                    termWidthCharacters,
                    termHeightCharacters,
                    termWidthPixels,
                    termHeightPixels,
                    terminalModes);
                
                channel.SuccessCounter = channel.FailedCounter = 0;
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent)
                {
                    throw new IOException("Cannot request PTY on this channel (" + channel.ReasonClosed + ")");
                }
                
                this.tm.SendMessage(spr.Payload);
            }
            
            try
            {
                ChannelManager.WaitForChannelSuccessOrFailure(channel);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("PTY request failed");
            }
        }
        
        /// <summary>
        /// Request X11 forwarding.
        /// </summary>
        /// <param name="channel">The channel to use.</param>
        /// <param name="singleConnection">If true indicates that the server should forward only one connection.</param>
        /// <param name="x11AuthenticationProtocol">Always MIT-MAGIC-COOKIE-1.</param>
        /// <param name="x11AuthenticationCookie">The actual cookie to use.</param>
        /// <param name="x11ScreenNumber">Always screen 0.</param>
        public void RequestX11(
            SSHChannel channel,
            bool singleConnection,
            string x11AuthenticationProtocol,
            string x11AuthenticationCookie,
            int x11ScreenNumber)
        {
            PacketSessionX11Request psr;
            
            lock (channel)
            {
                if (channel.State != SSHChannel.STATE_OPEN)
                {
                    throw new IOException("Cannot request X11 on this channel (" + channel.ReasonClosed + ")");
                }
                
                psr = new PacketSessionX11Request(
                    channel.RemoteId,
                    true,
                    singleConnection,
                    x11AuthenticationProtocol,
                    x11AuthenticationCookie,
                    x11ScreenNumber);
                
                channel.SuccessCounter = channel.FailedCounter = 0;
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent)
                {
                    throw new IOException("Cannot request X11 on this channel (" + channel.ReasonClosed + ")");
                }
                
                this.tm.SendMessage(psr.Payload);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Requesting X11 forwarding (Channel " + channel.LocalId + "/" + channel.RemoteId + ")");
            }
            
            try
            {
                ChannelManager.WaitForChannelSuccessOrFailure(channel);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("The X11 request failed.");
            }
        }
        
        /// <summary>
        /// Request a subsystem (built-in ssh server command.
        /// </summary>
        /// <param name="channel">The channel to use.</param>
        /// <param name="subsystemName">The subsystem name.</param>
        public void RequestSubsystem(SSHChannel channel, string subsystemName)
        {
            PacketSessionSubsystemRequest ssr;
            
            lock (channel)
            {
                if (channel.State != SSHChannel.STATE_OPEN)
                {
                    throw new IOException("Cannot request subsystem on this channel (" + channel.ReasonClosed + ")");
                }
                
                ssr = new PacketSessionSubsystemRequest(channel.RemoteId, true, subsystemName);
                
                channel.SuccessCounter = channel.FailedCounter = 0;
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent)
                {
                    throw new IOException("Cannot request subsystem on this channel (" + channel.ReasonClosed + ")");
                }
                
                this.tm.SendMessage(ssr.Payload);
            }
            
            try
            {
                ChannelManager.WaitForChannelSuccessOrFailure(channel);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("The subsystem request failed.");
            }
        }
        
        /// <summary>
        /// Request to exec a command.
        /// </summary>
        /// <param name="channel">The channel to use.</param>
        /// <param name="cmd">The command to execute.</param>
        public void RequestExecCommand(SSHChannel channel, string cmd)
        {
            PacketSessionExecCommand sm;
            
            lock (channel)
            {
                if (channel.State != SSHChannel.STATE_OPEN)
                {
                    throw new IOException("Cannot execute command on this channel (" + channel.ReasonClosed + ")");
                }
                
                sm = new PacketSessionExecCommand(channel.RemoteId, true, cmd);
                
                channel.SuccessCounter = channel.FailedCounter = 0;
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent)
                {
                    throw new IOException("Cannot execute command on this channel (" + channel.ReasonClosed + ")");
                }
                
                this.tm.SendMessage(sm.Payload);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Executing command (channel " + channel.LocalId + ", '" + cmd + "')");
            }
            
            try
            {
                ChannelManager.WaitForChannelSuccessOrFailure(channel);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("The execute request failed.");
            }
        }
        
        /// <summary>
        /// Starts a shell on the specifies channel.
        /// </summary>
        /// <param name="channel">The channel to use.</param>
        public void RequestShell(SSHChannel channel)
        {
            PacketSessionStartShell sm;
            
            lock (channel)
            {
                if (channel.State != SSHChannel.STATE_OPEN)
                {
                    throw new IOException("Cannot start shell on this channel (" + channel.ReasonClosed + ")");
                }
                
                sm = new PacketSessionStartShell(channel.RemoteId, true);
                
                channel.SuccessCounter = channel.FailedCounter = 0;
            }
            
            lock (channel.ChannelSendLock)
            {
                if (channel.CloseMessageSent)
                {
                    throw new IOException("Cannot start shell on this channel (" + channel.ReasonClosed + ")");
                }
                
                this.tm.SendMessage(sm.Payload);
            }
            
            try
            {
                ChannelManager.WaitForChannelSuccessOrFailure(channel);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("The shell request failed.");
            }
        }
        
        /// <summary>
        /// Process a request to accept extended data.
        /// </summary>
        /// <param name="msg">The message that was received.</param>
        /// <param name="msgLength">The length of the message.</param>
        public void MsgChannelExtendedData(byte[] msg, int msgLength)
        {
            if (msgLength <= 13)
            {
                throw new IOException("SSHMessageCHANNEL_EXTENDED_DATA message has wrong size (" + msgLength + ")");
            }
            
            int id = ((msg[1] & 0xff) << 24) | ((msg[2] & 0xff) << 16) | ((msg[3] & 0xff) << 8) | (msg[4] & 0xff);
            int dataType = ((msg[5] & 0xff) << 24) | ((msg[6] & 0xff) << 16) | ((msg[7] & 0xff) << 8) | (msg[8] & 0xff);
            int len = ((msg[9] & 0xff) << 24) | ((msg[10] & 0xff) << 16) | ((msg[11] & 0xff) << 8) | (msg[12] & 0xff);
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_EXTENDED_DATA message for non-existent channel " + id);
            }
            
            if (dataType != Packet.SSHExtendedDataStandardError)
            {
                throw new IOException("SSHMessageCHANNEL_EXTENDED_DATA message has unknown type (" + dataType + ")");
            }
            
            if (len != (msgLength - 13))
            {
                throw new IOException("SSHMessageCHANNEL_EXTENDED_DATA message has wrong len (calculated " + (msgLength - 13)
                                      + ", got " + len + ")");
            }
            
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageCHANNEL_EXTENDED_DATA (channel " + id + ", " + len + ")");
            }
            
            lock (c)
            {
                if (c.State == SSHChannel.STATE_CLOSED)
                {
                    return; // ignore
                }
                
                if (c.State != SSHChannel.STATE_OPEN)
                {
                    throw new IOException("Got SSHMessageCHANNEL_EXTENDED_DATA, but channel is not in correct state ("
                                          + c.State + ")");
                }
                
                if (c.LocalWindow < len)
                {
                    throw new IOException("Remote sent too much data, does not fit into window.");
                }
                
                c.LocalWindow -= len;
                
                Array.Copy(msg, 13, c.StandardErrorBuffer, c.StandardErrorWritePosition, len);
                c.StandardErrorWritePosition += len;

                Monitor.PulseAll(c);
            }
        }
        
        /// <summary>
        /// Gets data from the Channel.
        /// </summary>
        /// <param name="channel">The channel we are getting data on.</param>
        /// <param name="extended">True if the data is extended (extended is normally stderr).</param>
        /// <param name="target">The byte buffer that we are going to put the data in.</param>
        /// <param name="off">The offset into the buffer where we start storing data.</param>
        /// <param name="len">The maximum length that we can copy data into the buffer.</param>
        /// <returns>The number of bytes that were retrieved.</returns>
        public int GetChannelData(SSHChannel channel, bool extended, byte[] target, int off, int len)
        {
            bool wasInterrupted = false;
            
            try
            {
                int copylen = 0;
                int increment = 0;
                int remoteID = 0;
                int localID = 0;
                
                lock (channel)
                {
                    int stdoutAvail = 0;
                    int stderrAvail = 0;
                    
                    while (true)
                    {
                        // Data available? We have to return remaining data even if the
                        // channel is already closed.
                        stdoutAvail = channel.StandardOutputWritePosition - channel.StandardOutputReadPosition;
                        stderrAvail = channel.StandardErrorWritePosition - channel.StandardErrorReadPosition;
                        
                        if ((!extended) && (stdoutAvail != 0))
                        {
                            break;
                        }
                        
                        if (extended && (stderrAvail != 0))
                        {
                            break;
                        }
                        
                        // Do not wait if more data will never arrive (EOF or CLOSED)
                        if (channel.EOF || (channel.State != SSHChannel.STATE_OPEN))
                        {
                            return -1;
                        }
                        
                        try
                        {
                            Monitor.Wait(channel);
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }
                    
                    // OK, there is some data. Return it.
                    if (!extended)
                    {
                        copylen = (stdoutAvail > len) ? len : stdoutAvail;
                        Array.Copy(channel.StandardOutputBuffer, channel.StandardOutputReadPosition, target, off, copylen);
                        channel.StandardOutputReadPosition += copylen;
                        
                        if (channel.StandardOutputReadPosition != channel.StandardOutputWritePosition)
                        {
                            Array.Copy(
                                channel.StandardOutputBuffer,
                                channel.StandardOutputReadPosition,
                                channel.StandardOutputBuffer,
                                0,
                                channel.StandardOutputWritePosition - channel.StandardOutputReadPosition);
                        }
                        
                        channel.StandardOutputWritePosition -= channel.StandardOutputReadPosition;
                        channel.StandardOutputReadPosition = 0;
                    }
                    else
                    {
                        copylen = (stderrAvail > len) ? len : stderrAvail;
                        Array.Copy(channel.StandardErrorBuffer, channel.StandardErrorReadPosition, target, off, copylen);
                        channel.StandardErrorReadPosition += copylen;
                        
                        if (channel.StandardErrorReadPosition != channel.StandardErrorWritePosition)
                        {
                            Array.Copy(
                                channel.StandardErrorBuffer,
                                channel.StandardErrorReadPosition,
                                channel.StandardErrorBuffer,
                                0,
                                channel.StandardErrorWritePosition - channel.StandardErrorReadPosition);
                        }
                        
                        channel.StandardErrorWritePosition -= channel.StandardErrorReadPosition;
                        channel.StandardErrorReadPosition = 0;
                    }
                    
                    if (channel.State != SSHChannel.STATE_OPEN)
                    {
                        return copylen;
                    }
                    
                    if (channel.LocalWindow < ((SSHChannel.CHANNEL_BUFFER_SIZE + 1) / 2))
                    {
                        int minFreeSpace = Math.Min(
                            SSHChannel.CHANNEL_BUFFER_SIZE - channel.StandardOutputWritePosition,
                            SSHChannel.CHANNEL_BUFFER_SIZE - channel.StandardErrorWritePosition);
                        
                        increment = minFreeSpace - channel.LocalWindow;
                        channel.LocalWindow = minFreeSpace;
                    }
                    
                    remoteID = channel.RemoteId; // read while holding the lock
                    localID = channel.LocalId; // read while holding the lock
                }
                
                // If a consumer reads stdout and stdin in parallel, we may end up with
                // sending two msgWindowAdjust messages. Luckily, it
                // does not matter in which order they arrive at the server.
                if (increment > 0)
                {
                    if (log.Enabled)
                    {
                        log.Log(80, "Sending SSHMessageCHANNEL_WINDOW_ADJUST (channel " + localID + ", " + increment + ")");
                    }
                    
                    lock (channel.ChannelSendLock)
                    {
                        byte[] msg = channel.MessageWindowAdjust;
                        
                        msg[0] = Packet.SSHMessageChannelWindowAdjust;
                        unchecked
                        {
                            msg[1] = (byte)(remoteID >> 24);
                            msg[2] = (byte)(remoteID >> 16);
                            msg[3] = (byte)(remoteID >> 8);
                            msg[4] = (byte)remoteID;
                            msg[5] = (byte)(increment >> 24);
                            msg[6] = (byte)(increment >> 16);
                            msg[7] = (byte)(increment >> 8);
                            msg[8] = (byte)increment;
                        }
                        
                        if (channel.CloseMessageSent == false)
                        {
                            this.tm.SendMessage(msg);
                        }
                    }
                }
                
                return copylen;
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_DATA packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelData(byte[] msg, int msgLength)
        {
            if (msgLength <= 9)
            {
                throw new IOException("SSHMessageCHANNEL_DATA message has wrong size (" + msgLength + ")");
            }
            
            int id = ((msg[1] & 0xff) << 24) |
                ((msg[2] & 0xff) << 16) |
                ((msg[3] & 0xff) << 8) |
                (msg[4] & 0xff);
            int len = ((msg[5] & 0xff) << 24) |
                ((msg[6] & 0xff) << 16) |
                ((msg[7] & 0xff) << 8) |
                (msg[8] & 0xff);
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_DATA message for non-existent channel " + id);
            }
            
            if (len != (msgLength - 9))
            {
                throw new IOException("SSHMessageCHANNEL_DATA message has wrong len (calculated " + (msgLength - 9) + ", got "
                                      + len + ")");
            }
            
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageCHANNEL_DATA (channel " + id + ", " + len + ")");
            }
            
            lock (c)
            {
                if (c.State == SSHChannel.STATE_CLOSED)
                {
                    return; // ignore
                }
                
                if (c.State != SSHChannel.STATE_OPEN)
                {
                    throw new IOException("Got SSHMessageCHANNEL_DATA, but channel is not in correct state (" + c.State + ")");
                }
                
                if (c.LocalWindow < len)
                {
                    throw new IOException("Remote sent too much data, does not fit into window.");
                }
                
                c.LocalWindow -= len;
                
                Array.Copy(msg, 9, c.StandardOutputBuffer, c.StandardOutputWritePosition, len);
                c.StandardOutputWritePosition += len;
                
                Monitor.PulseAll(c);
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_WINDOW_ADJUST packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelWindowAdjust(byte[] msg, int msgLength)
        {
            if (msgLength != 9)
            {
                throw new IOException("SSHMessageCHANNEL_WINDOW_ADJUST message has wrong size (" + msgLength + ")");
            }
            
            int id = ((msg[1] & 0xff) << 24) |
                ((msg[2] & 0xff) << 16) |
                ((msg[3] & 0xff) << 8) |
                (msg[4] & 0xff);
            int windowChange = ((msg[5] & 0xff) << 24) |
                ((msg[6] & 0xff) << 16) |
                ((msg[7] & 0xff) << 8) |
                (msg[8] & 0xff);
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_WINDOW_ADJUST message for non-existent channel " + id);
            }
            
            lock (c)
            {
                long huge = 0xFFFFffffL; /* 2^32 - 1 */
                
                c.RemoteWindow += windowChange & huge; /* avoid sign extension */
                
                // TODO - is this a good heuristic?
                if (c.RemoteWindow > huge)
                {
                    c.RemoteWindow = huge;
                }
                
                Monitor.PulseAll(c);
            }
            
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageCHANNEL_WINDOW_ADJUST (channel " + id + ", " + windowChange + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_OPEN packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelOpen(byte[] msg, int msgLength)
        {
            TypesReader tr = new TypesReader(msg, 0, msgLength);
            
            tr.ReadByte(); // skip packet type
            string channelType = tr.ReadString();
            int remoteID = tr.ReadUINT32(); // sender channel
            int remoteWindow = tr.ReadUINT32(); // initial window size
            int remoteMaxPacketSize = tr.ReadUINT32(); // maximum packet size
            PacketChannelOpenFailure pcof;
            
            if ("x11".Equals(channelType))
            {
                lock (this.x11MagicCookies)
                {
                    // If we did not request X11 forwarding, then simply ignore this bogus request.
                    if (this.x11MagicCookies.Count == 0)
                    {
                        pcof = new PacketChannelOpenFailure(
                            remoteID,
                            Packet.SSHOpenAdministrativelyProhibited,
                            "X11 forwarding not activated",
                            string.Empty);
                        
                        this.tm.SendAsynchronousMessage(pcof.Payload);
                        
                        if (log.Enabled)
                        {
                            log.Log(20, "Unexpected X11 request, denying it!");
                        }
                        
                        return;
                    }
                }
                
                string remoteOriginatorAddress = tr.ReadString();
                int remoteOriginatorPort = tr.ReadUINT32();
                
                SSHChannel c = new SSHChannel(this);
                
                lock (c)
                {
                    c.RemoteId = remoteID;
                    c.RemoteWindow = remoteWindow & 0xFFFFffffL; // properly convert UINT32 to long
                    c.RemoteMaxPacketSize = remoteMaxPacketSize;
                    c.LocalId = this.AddChannel(c);
                }
                
                // The open confirmation message will be sent from another thread
                RemoteX11AcceptThread rxat = new RemoteX11AcceptThread(c);
                rxat.IsBackground = true;
                rxat.Start();
                
                return;
            }
            
            if ("forwarded-tcpip".Equals(channelType))
            {
                string remoteConnectedAddress = tr.ReadString(); // address that was connected
                int remoteConnectedPort = tr.ReadUINT32(); // port that was connected
                string remoteOriginatorAddress = tr.ReadString(); // originator IP address
                int remoteOriginatorPort = tr.ReadUINT32(); // originator port
                
                RemoteForwardingData rfd;
                
                lock (this.remoteForwardings)
                {
                    try
                    {
                        rfd = (RemoteForwardingData)this.remoteForwardings[remoteConnectedPort];
                    }
                    catch (NullReferenceException)
                    {
                        pcof = new PacketChannelOpenFailure(
                            remoteID,
                            Packet.SSHOpenAdministrativelyProhibited,
                            "No thanks, unknown port in forwarded-tcpip request",
                            string.Empty);
                        
                        // Always try to be polite.
                        this.tm.SendAsynchronousMessage(pcof.Payload);
                        
                        if (log.Enabled)
                        {
                            log.Log(20, "Unexpected forwarded-tcpip request, denying it!");
                        }
                        
                        return;
                    }
                }

                SSHChannel c = new SSHChannel(this);
                
                lock (c)
                {
                    c.RemoteId = remoteID;
                    c.RemoteWindow = remoteWindow & 0xFFFFffffL; // convert UINT32 to long
                    c.RemoteMaxPacketSize = remoteMaxPacketSize;
                    c.LocalId = this.AddChannel(c);
                }
                
                // The open confirmation message will be sent from another thread.
                RemoteAcceptThread rat = new RemoteAcceptThread(
                    c,
                    remoteConnectedAddress,
                    remoteConnectedPort,
                    remoteOriginatorAddress,
                    remoteOriginatorPort,
                    rfd.TargetAddress,
                    rfd.TargetPort);

                rat.IsBackground = true;
                rat.Start();
                
                return;
            }
            
            // Tell the server that we have no idea what it is talking about
            pcof = new PacketChannelOpenFailure(
                remoteID,
                Packet.SSHOpenUnknownChannelType,
                "Unknown channel type",
                string.Empty);
            
            this.tm.SendAsynchronousMessage(pcof.Payload);
            
            if (log.Enabled)
            {
                log.Log(20, "The peer tried to open an unsupported channel type (" + channelType + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_REQUEST packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelRequest(byte[] msg, int msgLength)
        {
            TypesReader tr = new TypesReader(msg, 0, msgLength);
            
            tr.ReadByte(); // skip packet type
            int id = tr.ReadUINT32();
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_REQUEST message for non-existent channel " + id);
            }
            
            string type = tr.ReadString("US-ASCII");
            bool wantReply = tr.ReadBoolean();
            
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageCHANNEL_REQUEST (channel " + id + ", '" + type + "')");
            }
            
            if (type.Equals("exit-status"))
            {
                if (wantReply != false)
                {
                    throw new IOException("Badly formatted SSHMessageCHANNEL_REQUEST message, 'want reply' is true");
                }
                
                int exit_status = tr.ReadUINT32();
                
                if (tr.Remain != 0)
                {
                    throw new IOException("Badly formatted SSHMessageCHANNEL_REQUEST message");
                }
                
                lock (c)
                {
                    c.ExitStatus = exit_status;
                    Monitor.PulseAll(c);
                }
                
                if (log.Enabled)
                {
                    log.Log(50, "Got EXIT STATUS (channel " + id + ", status " + exit_status + ")");
                }
                
                return;
            }
            
            if (type.Equals("exit-signal"))
            {
                if (wantReply != false)
                {
                    throw new IOException("Badly formatted SSHMessageCHANNEL_REQUEST message, 'want reply' is true");
                }
                
                string signame = tr.ReadString("US-ASCII");
                tr.ReadBoolean();
                tr.ReadString();
                tr.ReadString();
                
                if (tr.Remain != 0)
                {
                    throw new IOException("Badly formatted SSHMessageCHANNEL_REQUEST message");
                }
                
                lock (c)
                {
                    c.ExitSignal = signame;
                    Monitor.PulseAll(c);
                }
                
                if (log.Enabled)
                {
                    log.Log(50, "Got EXIT SIGNAL (channel " + id + ", signal " + signame + ")");
                }
                
                return;
            }
            
            // We simply ignore unknown channel requests, however, if the server wants a reply,
            // then we signal that we have no idea what it is about.
            if (wantReply)
            {
                byte[] reply = new byte[5];
                
                reply[0] = Packet.SSHMessageChannelFailure;
                unchecked
                {
                    reply[1] = (byte)(c.RemoteId >> 24);
                    reply[2] = (byte)(c.RemoteId >> 16);
                    reply[3] = (byte)(c.RemoteId >> 8);
                    reply[4] = (byte)c.RemoteId;
                }
                
                this.tm.SendAsynchronousMessage(reply);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Channel request '" + type + "' is not known, ignoring it");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_EOF packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelEOF(byte[] msg, int msgLength)
        {
            if (msgLength != 5)
            {
                throw new IOException("SSHMessageCHANNEL_EOF message has wrong size (" + msgLength + ")");
            }
            
            int id = ((msg[1] & 0xff) << 24) |
                ((msg[2] & 0xff) << 16) |
                ((msg[3] & 0xff) << 8) |
                (msg[4] & 0xff);
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_EOF message for non-existent channel " + id);
            }
            
            lock (c)
            {
                c.EOF = true;
                Monitor.PulseAll(c);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Got SSHMessageCHANNEL_EOF (channel " + id + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_CLOSE packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelClose(byte[] msg, int msgLength)
        {
            if (msgLength != 5)
            {
                throw new IOException("SSHMessageCHANNEL_CLOSE message has wrong size (" + msgLength + ")");
            }
            
            int id = ((msg[1] & 0xff) << 24) | ((msg[2] & 0xff) << 16) | ((msg[3] & 0xff) << 8) | (msg[4] & 0xff);
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_CLOSE message for non-existent channel " + id);
            }
            
            lock (c)
            {
                c.EOF = true;
                c.State = SSHChannel.STATE_CLOSED;
                c.ReasonClosed = "Close requested by remote";
                c.CloseMessageReceived = true;
                
                this.RemoveChannel(c.LocalId);
                
                Monitor.PulseAll(c);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Got SSHMessageCHANNEL_CLOSE (channel " + id + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_SUCCESS packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelSuccess(byte[] msg, int msgLength)
        {
            if (msgLength != 5)
            {
                throw new IOException("SSHMessageCHANNEL_SUCCESS message has wrong size (" + msgLength + ")");
            }
            
            int id = ((msg[1] & 0xff) << 24) | ((msg[2] & 0xff) << 16) | ((msg[3] & 0xff) << 8) | (msg[4] & 0xff);
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_SUCCESS message for non-existent channel " + id);
            }
            
            lock (c)
            {
                c.SuccessCounter++;
                Monitor.PulseAll(c);
            }
            
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageCHANNEL_SUCCESS (channel " + id + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_FAILURE packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelFailure(byte[] msg, int msgLength)
        {
            if (msgLength != 5)
            {
                throw new IOException("SSHMessageCHANNEL_FAILURE message has wrong size (" + msgLength + ")");
            }
            
            int id = ((msg[1] & 0xff) << 24) | ((msg[2] & 0xff) << 16) | ((msg[3] & 0xff) << 8) | (msg[4] & 0xff);
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_FAILURE message for non-existent channel " + id);
            }
            
            lock (c)
            {
                c.FailedCounter++;
                Monitor.PulseAll(c);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Got SSHMessageCHANNEL_FAILURE (channel " + id + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_OPEN_CONFIRMATION packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelOpenConfirmation(byte[] msg, int msgLength)
        {
            PacketChannelOpenConfirmation sm = new PacketChannelOpenConfirmation(msg, 0, msgLength);
            
            SSHChannel c = this.GetChannel(sm.RecipientChannelId);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_OPEN_CONFIRMATION message for non-existent channel "
                                      + sm.RecipientChannelId);
            }
            
            lock (c)
            {
                if (c.State != SSHChannel.STATE_OPENING)
                {
                    throw new IOException("Unexpected SSHMessageCHANNEL_OPEN_CONFIRMATION message for channel "
                                          + sm.RecipientChannelId);
                }
                
                c.RemoteId = sm.SenderChannelId;
                c.RemoteWindow = sm.InitialWindowSize & 0xFFFFffffL; // convert UINT32 to long
                c.RemoteMaxPacketSize = sm.MaxPacketSize;
                c.State = SSHChannel.STATE_OPEN;
                Monitor.PulseAll(c);
            }
            
            if (log.Enabled)
            {
                log.Log(
                    50,
                    "Got SSHMessageCHANNEL_OPEN_CONFIRMATION (channel " + sm.RecipientChannelId + " / remote: " + sm.SenderChannelId + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageCHANNEL_OPEN_FAILURE packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgChannelOpenFailure(byte[] msg, int msgLength)
        {
            if (msgLength < 5)
            {
                throw new IOException("SSHMessageCHANNEL_OPEN_FAILURE message has wrong size (" + msgLength + ")");
            }
            
            TypesReader tr = new TypesReader(msg, 0, msgLength);
            
            tr.ReadByte(); // skip packet type
            int id = tr.ReadUINT32(); // sender channel
            
            SSHChannel c = this.GetChannel(id);
            
            if (c == null)
            {
                throw new IOException("Unexpected SSHMessageCHANNEL_OPEN_FAILURE message for non-existent channel " + id);
            }
            
            int reasonCode = tr.ReadUINT32();
            string description = tr.ReadString("UTF-8");
            
            string reasonCodeSymbolicName = null;
            
            switch (reasonCode)
            {
                case 1:
                    reasonCodeSymbolicName = "SSH_OPEN_ADMINISTRATIVELY_PROHIBITED";
                    break;
                case 2:
                    reasonCodeSymbolicName = "SSH_OPEN_CONNECT_FAILED";
                    break;
                case 3:
                    reasonCodeSymbolicName = "SSH_OPEN_UNKNOWN_CHANNEL_TYPE";
                    break;
                case 4:
                    reasonCodeSymbolicName = "SSH_OPEN_RESOURCE_SHORTAGE";
                    break;
                default:
                    reasonCodeSymbolicName = "UNKNOWN REASON CODE (" + reasonCode + ")";
                    break;
            }
            
            StringBuilder descriptionBuffer = new StringBuilder();
            descriptionBuffer.Append(description);
            
            for (int i = 0; i < descriptionBuffer.Length; i++)
            {
                char cc = descriptionBuffer[i];
                
                if ((cc >= 32) && (cc <= 126))
                {
                    continue;
                }
                
                descriptionBuffer[i] = '\uFFFD';
            }
            
            lock (c)
            {
                c.EOF = true;
                c.State = SSHChannel.STATE_CLOSED;
                c.ReasonClosed = "The server refused to open the channel (" + reasonCodeSymbolicName + ", '"
                    + descriptionBuffer.ToString() + "')";
                Monitor.PulseAll(c);
            }
            
            if (log.Enabled)
            {
                log.Log(50, "Got SSHMessageCHANNEL_OPEN_FAILURE (channel " + id + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageGLOBAL_REQUEST packet.
        /// </summary>
        /// <param name="msg">The packet to process.</param>
        /// <param name="msgLength">The length of the packet to process.</param>
        public void MsgGlobalRequest(byte[] msg, int msgLength)
        {
            /* Currently we do not support any kind of global request */
            
            TypesReader tr = new TypesReader(msg, 0, msgLength);
            
            tr.ReadByte(); // skip packet type
            string requestName = tr.ReadString();
            bool wantReply = tr.ReadBoolean();
            
            if (wantReply)
            {
                byte[] reply_failure = new byte[1];
                reply_failure[0] = Packet.SSHMessageRequestFailure;
                
                this.tm.SendAsynchronousMessage(reply_failure);
            }
            
            // We do not clean up the requestName string - that is OK for debug
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageGLOBAL_REQUEST (" + requestName + ")");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageREQUEST_SUCCESS packet.
        /// </summary>
        public void MsgGlobalSuccess()
        {
            lock (this.channels)
            {
                this.globalSuccessCounter++;
                Monitor.PulseAll(this.channels);
            }
            
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageREQUEST_SUCCESS");
            }
        }
        
        /// <summary>
        /// Process a SSHMessageREQUEST_FAILURE packet.
        /// </summary>
        public void MsgGlobalFailure()
        {
            lock (this.channels)
            {
                this.globalFailedCounter++;
                Monitor.PulseAll(this.channels);
            }
            
            if (log.Enabled)
            {
                log.Log(80, "Got SSHMessageREQUEST_FAILURE");
            }
        }
        
        /// <summary>
        /// Method to support the MessageHandler interface.
        /// Processes all incoming messages.
        /// </summary>
        /// <param name="msg">The message (packet) that was received.</param>
        /// <param name="msgLength">The length of the message.</param>
        public override void HandleMessage(byte[] msg, int msgLength)
        {
            if (msg == null)
            {
                if (log.Enabled)
                {
                    log.Log(50, "HandleMessage: got shutdown");
                }
                
                lock (this.listenerThreads)
                {
                    for (int i = 0; i < this.listenerThreads.Count; i++)
                    {
                        IChannelWorkerThread lat = (IChannelWorkerThread)this.listenerThreads[i];
                        lat.StopWorking();
                    }
                    
                    this.listenerThreadsAllowed = false;
                }
                
                lock (this.channels)
                {
                    this.shutdown = true;
                    
                    for (int i = 0; i < this.channels.Count; i++)
                    {
                        SSHChannel c = (SSHChannel)this.channels[i];
                        lock (c)
                        {
                            c.EOF = true;
                            c.State = SSHChannel.STATE_CLOSED;
                            c.ReasonClosed = "The connection is being shutdown";
                            c.CloseMessageReceived = true;
                            
                            // You never know, perhaps
                            // we are waiting for a
                            // pending close message
                            // from the server...
                            Monitor.PulseAll(c);
                        }
                    }
                    
                    this.channels.RemoveRange(0, this.channels.Count - 0);
                    this.channels.TrimToSize();
                    Monitor.PulseAll(this.channels); // Notify global response waiters
                    return;
                }
            }
            
            switch (msg[0])
            {
                case Packet.SSHMessageChannelOpenConfirmation:
                    this.MsgChannelOpenConfirmation(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelWindowAdjust:
                    this.MsgChannelWindowAdjust(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelData:
                    this.MsgChannelData(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelExtendedData:
                    this.MsgChannelExtendedData(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelRequest:
                    this.MsgChannelRequest(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelEOF:
                    this.MsgChannelEOF(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelOpen:
                    this.MsgChannelOpen(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelClose:
                    this.MsgChannelClose(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelSuccess:
                    this.MsgChannelSuccess(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelFailure:
                    this.MsgChannelFailure(msg, msgLength);
                    break;
                case Packet.SSHMessageChannelOpenFailure:
                    this.MsgChannelOpenFailure(msg, msgLength);
                    break;
                case Packet.SSHMessageGlobalRequest:
                    this.MsgGlobalRequest(msg, msgLength);
                    break;
                case Packet.SSHMessageRequestSuccess:
                    this.MsgGlobalSuccess();
                    break;
                case Packet.SSHMessageRequestFailure:
                    this.MsgGlobalFailure();
                    break;
                default:
                    throw new IOException("Cannot handle unknown channel message " + (msg[0] & 0xff));
            }
        }
        
        /// <summary>
        /// Wait to see if the request on the specified channel succeeds or fails.
        /// </summary>
        /// <param name="channel">The channel we need to wait for success or failure on.</param>
        private static void WaitForChannelSuccessOrFailure(SSHChannel channel)
        {
            bool wasInterrupted = false;
            
            try
            {
                lock (channel)
                {
                    while ((channel.SuccessCounter == 0) && (channel.FailedCounter == 0))
                    {
                        if (channel.State != SSHChannel.STATE_OPEN)
                        {
                            string detail = channel.ReasonClosed;
                            
                            if (detail == null)
                            {
                                detail = "state: " + channel.State;
                            }
                            
                            throw new IOException("This SSH2 channel is not open (" + detail + ")");
                        }
                        
                        try
                        {
                            Monitor.Wait(channel);
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }
                    
                    if (channel.FailedCounter != 0)
                    {
                        throw new IOException("The server denied the request.");
                    }
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }
        
        /// <summary>
        /// Gets a Channel from this ChannelManager using the Channel's id.
        /// </summary>
        /// <param name="id">This is an integer id for the new channel.</param>
        /// <returns>This is the Channel if the id exists or null if it does not.</returns>
        private SSHChannel GetChannel(int id)
        {
            lock (this.channels.SyncRoot)
            {
                for (int i = 0; i < this.channels.Count; i++)
                {
                    SSHChannel c = (SSHChannel)this.channels[i];
                    if (c.LocalId == id)
                    {
                        return c;
                    }
                }
            }
            
            return null;
        }
        
        /// <summary>
        /// Removes a Channel from this ChannelManager using the Channel's id.
        /// </summary>
        /// <param name="id">This is an integer id for the new channel.</param>
        private void RemoveChannel(int id)
        {
            lock (this.channels.SyncRoot)
            {
                for (int i = 0; i < this.channels.Count; i++)
                {
                    SSHChannel c = (SSHChannel)this.channels[i];
                    if (c.LocalId == id)
                    {
                        this.channels.RemoveAt(i);
                        break;
                    }
                }
            }
        }
        
        /// <summary>
        /// Adds a Channel to this ChannelManager.
        /// </summary>
        /// <param name="c">This is the Channel that is to be added to this ChannelManager.</param>
        /// <returns>This is the channel id of the added Channel.</returns>
        private int AddChannel(SSHChannel c)
        {
            lock (this.channels.SyncRoot)
            {
                this.channels.Add(c);
                return this.nextLocalChannel++;
            }
        }
        
        /// <summary>
        /// Waits until the channel passed to it is open.
        /// </summary>
        /// <param name="c">The channel that we are waiting till it opens.</param>
        private void WaitUntilChannelOpen(SSHChannel c)
        {
            bool wasInterrupted = false;
            
            lock (c)
            {
                while (c.State == SSHChannel.STATE_OPENING)
                {
                    try
                    {
                        Monitor.Wait(c);
                    }
                    catch (ThreadInterruptedException)
                    {
                        wasInterrupted = true;
                    }
                }
                
                if (c.State != SSHChannel.STATE_OPEN)
                {
                    this.RemoveChannel(c.LocalId);
                    
                    string detail = c.ReasonClosed;
                    
                    if (detail == null)
                    {
                        detail = "state: " + c.State;
                    }
                    
                    throw new IOException("Could not open channel (" + detail + ")");
                }
            }
            
            if (wasInterrupted)
            {
                JavaThread.Current().Interrupt();
            }
        }
        
        /// <summary>
        /// Wait to see if the global forward succeeds or fails.
        /// </summary>
        private void WaitForGlobalSuccessOrFailure()
        {
            bool wasInterrupted = false;
            
            try
            {
                lock (this.channels.SyncRoot)
                {
                    while ((this.globalSuccessCounter == 0) && (this.globalFailedCounter == 0))
                    {
                        if (this.shutdown)
                        {
                            throw new IOException("The connection is being shutdown");
                        }
                        
                        try
                        {
                            Monitor.Wait(this.channels.SyncRoot);
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }
                    
                    if (this.globalFailedCounter != 0)
                    {
                        throw new IOException("The server denied the request (did you enable port forwarding?)");
                    }
                    
                    if (this.globalSuccessCounter == 0)
                    {
                        throw new IOException("Illegal state.");
                    }
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }
    }
}
