﻿// <copyright file="RemoteAcceptThread.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> RemoteAcceptThread.
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version>  0.80, 11/26/2010
// </version>

using System;
using System.IO;
using System.Net.Sockets;
using System.Threading;

using Eunomia.SSH2.Log;

namespace Eunomia.SSH2.Channel
{
    /// <summary>
    /// Class implements a TcpClient that tries to connect to a remote system/port on it's own thread.
    /// </summary>
    public class RemoteAcceptThread : JavaThread, IDisposable
    {
        private static Logger log = Logger.GetLogger(typeof(RemoteAcceptThread));
        
        private SSHChannel c;
        
        private string targetAddress;
        private int targetPort;
        
        private TcpClient tcpClient;
        
        /// <summary>
        /// Initializes a new instance of the RemoteAcceptThread class.
        /// </summary>
        /// <param name="channel">The Channel that we are forwarding on.</param>
        /// <param name="remoteConnectedAddress">The address that is forwarded from.</param>
        /// <param name="remoteConnectedPort">The port that is forwarded from.</param>
        /// <param name="remoteOriginatorAddress">The address that the data originally came from.</param>
        /// <param name="remoteOriginatorPort">The port that the data originally came from.</param>
        /// <param name="targetAddress">The destination address of the forward.</param>
        /// <param name="targetPort">The destination port of the forward.</param>
        public RemoteAcceptThread(
            SSHChannel channel,
            string remoteConnectedAddress,
            int remoteConnectedPort,
            string remoteOriginatorAddress,
            int remoteOriginatorPort,
            string targetAddress,
            int targetPort)
        {
            this.c = channel;
            this.targetAddress = targetAddress;
            this.targetPort = targetPort;
            
            if (log.Enabled)
            {
                string msg = "RemoteAcceptThread: " + remoteConnectedAddress + "/" + remoteConnectedPort + ", R: "
                    + remoteOriginatorAddress + "/" + remoteOriginatorPort;
                log.Log(20, msg);
            }
        }
        
        /// <summary>
        /// Implement the public IDisposable interface method.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }
        
        /// <summary>
        /// Runs the thread that creates the remote port forward.
        /// </summary>
        public override void Run()
        {
            try
            {
                this.c.CM.SendOpenConfirmation(this.c);
                
                this.tcpClient = new TcpClient(this.targetAddress, this.targetPort);
                
                StreamForwarder r2l = new StreamForwarder(this.c, null, null, this.c.StandardOutputStream, this.tcpClient.GetStream(), "RemoteToLocal");
                StreamForwarder l2r = new StreamForwarder(this.c, null, null, this.tcpClient.GetStream(), this.c.StandardInputStream, "LocalToRemote");
                
                /* No need to start two threads, one can be executed in the current thread */

                r2l.IsBackground = true;
                r2l.Start();
                l2r.Run();
                
                while (r2l.IsAlive)
                {
                    try
                    {
                        r2l.Join();
                    }
                    catch (ThreadInterruptedException)
                    {
                    }
                }
                
                // If the channel is already closed, then this is a no-op
                this.c.CM.CloseChannel(this.c, "EOF on both streams reached.", true);
                this.tcpClient.Close();
            }
            catch (IOException e)
            {
                log.Log(50, "IOException in proxy code: " + e.ToString());
                
                try
                {
                    this.c.CM.CloseChannel(this.c, "IOException in proxy code (" + e.ToString() + ")", true);
                }
                catch (IOException)
                {
                }
                
                try
                {
                    if (this.tcpClient != null)
                    {
                        this.tcpClient.Close();
                    }
                }
                catch (IOException)
                {
                }
            }
        }
        
        /// <summary>
        /// Implement the protected IDisposable interface method.
        /// </summary>
        /// <param name="disposing">True if we are disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.tcpClient.Close();
            }
        }
    }
}
