﻿// <copyright file="RemoteX11AcceptThread.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> RemoteX11AcceptThread.
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version>  0.80, 11/26/2010
// </version>

using System;
using System.Globalization;
using System.IO;
using System.Net.Sockets;
using System.Text;
using System.Threading;

using Eunomia.SSH2.Log;
using Eunomia.SSH2.Util;

namespace Eunomia.SSH2.Channel
{
    /// <summary>
    /// Class implements a TcpClient that tries to connect to a remote system/port on it's own thread
    /// to forward X11 requests.
    /// </summary>
    public class RemoteX11AcceptThread : JavaThread, IDisposable
    {
        private static Logger log = Logger.GetLogger(typeof(RemoteX11AcceptThread));
        
        private SSHChannel c;
        
        private TcpClient tcpClient;

        /// <summary>
        /// Initializes a new instance of the RemoteX11AcceptThread class.
        /// </summary>
        /// <param name="channel">The Channel that we are forwarding on.</param>
        public RemoteX11AcceptThread(SSHChannel channel)
        {
            this.c = channel;
        }
        
        /// <summary>
        /// Implement the public IDisposable interface method.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }
        
        /// <summary>
        /// Runs the thread that creates the remote port forward.
        /// </summary>
        public override void Run()
        {
            try
            {
                // Send Open Confirmation
                this.c.CM.SendOpenConfirmation(this.c);
                
                // Read startup packet from client
                Stream remote_os = this.c.StandardInputStream;
                Stream remote_is = this.c.StandardOutputStream;
                
                // The following code is based on the protocol description given in:
                // Scheifler/Gettys,
                // X Windows System: Core and Extension Protocols:
                // X Version 11, Releases 6 and 6.1 ISBN 1-55558-148-X
                // (from the ETH library - after being here for almost ten
                // years one of the few books I borrowed... sad but true =)

                // Client startup:
                // 
                // 1 0X42 MSB first/0x6c lSB first - byteorder
                // 1 - unused
                // 2 card16 - protocol-major-version
                // 2 card16 - protocol-minor-version
                // 2 n - lenght of authorization-protocol-name
                // 2 d - lenght of authorization-protocol-data
                // 2 - unused
                // string8 - authorization-protocol-name
                // p - unused, p=pad(n)
                // string8 - authorization-protocol-data
                // q - unused, q=pad(d)
                // 
                // pad(X) = (4 - (X mod 4)) mod 4
                // 
                // Server response:
                // 
                // 1 (0 failed, 2 authenticate, 1 success)
                // ...
                
                // Later on we will simply forward the first 6 header bytes to the "real" X11 server
                byte[] header = new byte[6];
                
                if (remote_is.Read(header, 0, 6) != 6)
                {
                    throw new IOException("Unexpected EOF on X11 startup!");
                }
                
                // 0x42 MSB first, 0x6C LSB first
                if ((header[0] != 0x42) && (header[0] != 0x6c))
                {
                    throw new IOException("Unknown endian format in X11 message!");
                }
                
                // Yes, I came up with this myself - shall I file an application for a patent? =)
                int idxMSB = (header[0] == 0x42) ? 0 : 1;
                
                // Read authorization data header
                byte[] auth_buff = new byte[6];
                
                if (remote_is.Read(auth_buff, 0, 6) != 6)
                {
                    throw new IOException("Unexpected EOF on X11 startup!");
                }
                
                int authProtocolNameLength = ((auth_buff[idxMSB] & 0xff) << 8) | (auth_buff[1 - idxMSB] & 0xff);
                int authProtocolDataLength = ((auth_buff[2 + idxMSB] & 0xff) << 8) | (auth_buff[3 - idxMSB] & 0xff);
                
                if ((authProtocolNameLength > 256) || (authProtocolDataLength > 256))
                {
                    throw new IOException("Buggy X11 authorization data");
                }
                
                int authProtocolNamePadding = (4 - (authProtocolNameLength % 4)) % 4;
                int authProtocolDataPadding = (4 - (authProtocolDataLength % 4)) % 4;
                
                byte[] authProtocolName = new byte[authProtocolNameLength];
                byte[] authProtocolData = new byte[authProtocolDataLength];
                
                byte[] paddingBuffer = new byte[4];
                
                if (remote_is.Read(authProtocolName, 0, authProtocolNameLength) != authProtocolNameLength)
                {
                    throw new IOException("Unexpected EOF on X11 startup! (authProtocolName)");
                }
                
                if (remote_is.Read(paddingBuffer, 0, authProtocolNamePadding) != authProtocolNamePadding)
                {
                    throw new IOException("Unexpected EOF on X11 startup! (authProtocolNamePadding)");
                }
                
                if (remote_is.Read(authProtocolData, 0, authProtocolDataLength) != authProtocolDataLength)
                {
                    throw new IOException("Unexpected EOF on X11 startup! (authProtocolData)");
                }
                
                if (remote_is.Read(paddingBuffer, 0, authProtocolDataPadding) != authProtocolDataPadding)
                {
                    throw new IOException("Unexpected EOF on X11 startup! (authProtocolDataPadding)");
                }
                
                if ("MIT-MAGIC-COOKIE-1".Equals(StringEncoder.GetString(authProtocolName)) == false)
                {
                    throw new IOException("Unknown X11 authorization protocol!");
                }
                
                if (authProtocolDataLength != 16)
                {
                    throw new IOException("Wrong data length for X11 authorization data!");
                }
                
                StringBuilder tmp = new StringBuilder(32);
                for (int i = 0; i < authProtocolData.Length; i++)
                {
                    string digit2 = (authProtocolData[i] & 0xff).ToString("X2", CultureInfo.InvariantCulture);
                    tmp.Append((digit2.Length == 2) ? digit2 : "0" + digit2);
                }
                
                string hexEncodedFakeCookie = tmp.ToString();
                
                // Order is very important here - it may be that a certain x11 forwarding
                // gets disabled right in the moment when we check and register our connection
                lock (this.c)
                {
                    // Please read the comment in Channel.cs
                    this.c.HexX11FakeCookie = hexEncodedFakeCookie;
                }
                
                // Now check our fake cookie directory to see if we produced this cookie
                X11ServerData sd = this.c.CM.CheckX11Cookie(hexEncodedFakeCookie);
                
                if (sd == null)
                {
                    throw new IOException("Invalid X11 cookie received.");
                }
                
                // If the session which corresponds to this cookie is closed then we will
                // detect this: the session's close code will close all channels
                // with the session's assigned x11 fake cookie.
                this.tcpClient = new TcpClient(sd.HostName, sd.Port);
                NetworkStream x11_os = this.tcpClient.GetStream();
                NetworkStream x11_is = this.tcpClient.GetStream();
                
                // Now we are sending the startup packet to the real X11 server
                x11_os.Write(header, 0, header.Length);
                
                if (sd.X11MagicCookie == null)
                {
                    byte[] emptyAuthData = new byte[6];

                    // empty auth data, hopefully you are connecting to localhost =)
                    x11_os.Write(emptyAuthData, 0, emptyAuthData.Length);
                }
                else
                {
                    if (sd.X11MagicCookie.Length != 16)
                    {
                        throw new IOException("The real X11 cookie has an invalid length!");
                    }
                    
                    // send X11 cookie specified by client
                    x11_os.Write(auth_buff, 0, auth_buff.Length);
                    x11_os.Write(authProtocolName, 0, authProtocolName.Length); /* re-use */
                    x11_os.Write(paddingBuffer, 0, authProtocolNamePadding);
                    x11_os.Write(sd.X11MagicCookie, 0, sd.X11MagicCookie.Length);
                    x11_os.Write(paddingBuffer, 0, authProtocolDataPadding);
                }
                
                x11_os.Flush();
                
                // Start forwarding traffic
                StreamForwarder r2l = new StreamForwarder(this.c, null, null, remote_is, x11_os, "RemoteToX11");
                StreamForwarder l2r = new StreamForwarder(this.c, null, null, x11_is, remote_os, "X11ToRemote");
                
                // No need to start two threads, one can be executed in the current thread
                r2l.IsBackground = true;
                r2l.Start();
                l2r.Run();
                
                while (r2l.IsAlive)
                {
                    try
                    {
                        r2l.Join();
                    }
                    catch (ThreadInterruptedException)
                    {
                    }
                }
                
                // If the channel is already closed, then this is a no-op
                this.c.CM.CloseChannel(this.c, "EOF on both X11 streams reached.", true);
                this.tcpClient.Close();
            }
            catch (IOException e)
            {
                log.Log(50, "IOException in X11 proxy code: " + e.ToString());
                
                try
                {
                    this.c.CM.CloseChannel(this.c, "IOException in X11 proxy code (" + e.ToString() + ")", true);
                }
                catch (IOException)
                {
                }
                
                try
                {
                    if (this.tcpClient != null)
                    {
                        this.tcpClient.Close();
                    }
                }
                catch (IOException)
                {
                }
            }
        }
        
        /// <summary>
        /// Implement the protected IDisposable interface method.
        /// </summary>
        /// <param name="disposing">True if we are disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.tcpClient.Close();
            }
        }
    }
}
