﻿// <copyright file="StreamForwarder.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> StreamForwarder.
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version>  0.80, 11/26/2010
// </version>

using System;
using System.IO;
using System.Net.Sockets;
using System.Threading;

namespace Eunomia.SSH2.Channel
{
    /// <summary>
    /// A StreamForwarder forwards data between two given streams.
    /// If two StreamForwarder threads are used (one for each direction)
    /// then one can be configured to shutdown the underlying channel/socket
    /// if both threads have finished forwarding (EOF).
    /// </summary>
    public class StreamForwarder : JavaThread
    {
        private Stream ostr;
        private Stream istr;
        private byte[] buffer;
        private SSHChannel c;
        private StreamForwarder sibling;
        private Socket s;
        private string mode;

        /// <summary>
        /// Initializes a new instance of the StreamForwarder class.
        /// </summary>
        /// <param name="c">The Channel that we are forwarding on.</param>
        /// <param name="sibling">Sibling stream forwarder.</param>
        /// <param name="s">The socket that we are using.</param>
        /// <param name="istr">The input stream.</param>
        /// <param name="ostr">The output stream.</param>
        /// <param name="mode">The mode that we are in.</param>
        internal StreamForwarder(SSHChannel c, StreamForwarder sibling, Socket s, Stream istr, Stream ostr, string mode)
        {
            this.buffer = new byte[SSHChannel.CHANNEL_BUFFER_SIZE];
            this.istr = istr;
            this.ostr = ostr;
            this.mode = mode;
            this.c = c;
            this.sibling = sibling;
            this.s = s;
        }
        
        /// <summary>
        /// Runs the thread that handles the forward.
        /// </summary>
        public override void Run()
        {
            try
            {
                while (true)
                {
                    int len = this.istr.Read(this.buffer, 0, this.buffer.Length);
                    if (len <= 0)
                    {
                        break;
                    }
                    
                    this.ostr.Write(this.buffer, 0, len);
                    this.ostr.Flush();
                }
            }
            catch (IOException ignore)
            {
                try
                {
                    this.c.CM.CloseChannel(
                        this.c,
                        "Closed due to exception in StreamForwarder (" + this.mode + "): " + ignore.ToString(),
                        true);
                }
                catch (IOException)
                {
                }
            }
            finally
            {
                try
                {
                    this.ostr.Close();
                }
                catch (IOException)
                {
                }
                
                try
                {
                    this.istr.Close();
                }
                catch (IOException)
                {
                }
                
                if (this.sibling != null)
                {
                    while (this.sibling.IsAlive)
                    {
                        try
                        {
                            this.sibling.Join();
                        }
                        catch (ThreadInterruptedException)
                        {
                        }
                    }
                    
                    try
                    {
                        this.c.CM.CloseChannel(
                            this.c,
                            "StreamForwarder (" + this.mode + ") is cleaning up the connection",
                            true);
                    }
                    catch (IOException)
                    {
                    }
                    
                    try
                    {
                        if (this.s != null)
                        {
                            this.s.Close();
                        }
                    }
                    catch (IOException)
                    {
                    }
                }
            }
        }
    }
}
