// <copyright file="Connection.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> Connection
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Runtime.CompilerServices;

using Eunomia.SSH2.Auth;
using Eunomia.SSH2.Channel;
using Eunomia.SSH2.Crypto;
using Eunomia.SSH2.Crypto.Cipher;
using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Packets;
using Eunomia.SSH2.Transport;
using Eunomia.SSH2.Util;
using Org.BouncyCastle.Security;
using IPAddress = System.Net.IPEndPoint;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A Connection is used to establish an encrypted TCP/IP
    /// connection to a SSH-2 server.
    /// </summary>
    /// <remarks>
    /// Typically, one:
    /// <list type="bullet">
    /// <item>creates a <see cref="Eunomia.SSH2.Connection"></see> instance.</item>
    /// <item>calls the <see cref="Connect()"></see> method.</item>
    /// <item>calls some of the authentication methods (e.g., <see cref="Eunomia.SSH2.Auth.AuthenticationManager"></see>).</item>
    /// <item>calls one or several times the <see cref="OpenSession()"/> method.</item>
    /// <item>finally, one must close the connection and release resources with the <see cref="Close()"></see> method.</item>
    /// </list>
    /// </remarks>
    public class Connection : IDisposable
    {
        /// <summary> The identifier presented to the SSH-2 server.</summary>
        public const string Identification = "Eunomia Build 0.50";

        // Will be used to generate all random data needed for the current connection.
        // Note: SecureRandom.NextBytes() is thread safe
        // ToDo: Make sure this is true for BouncyCastle!
        private SecureRandom generator;
        private Eunomia.SSH2.Auth.AuthenticationManager am;
        private bool authenticated = false;
        private ChannelManager cm;
        private CryptoWishList cryptoWishList;
        private DHGexParameters dhgexpara;
        private string hostName;
        private int port;
        private TransportManager tm;
        private bool tcpNoDelay = false;
        private ProxyData proxyData = null;
        private ArrayList connectionMonitors;

        /// <summary>
        /// Initializes a new instance of the Connection class, which can then be used
        /// to establish a connection to the specified SSH-2 server.
        /// <para>
        /// Same as Connection(hostName, 22).
        /// </para>
        /// </summary>
        /// <param name="hostName">The hostName of the SSH-2 server that we are connecting to.
        /// </param>
        public Connection(string hostName) : this(hostName, 22)
        {
        }

        /// <summary>
        /// Initializes a new instance of the Connection class, which can then be used
        /// to establish a connection to the specified SSH-2 server.
        /// </summary>
        /// <param name="hostName">
        /// The hostName of the SSH-2 server that we are connecting to.
        /// </param>
        /// <param name="port">
        /// The port of the SSH-2 server that we are connecting to, (normally 22).
        /// </param>
        public Connection(string hostName, int port)
        {
            this.cryptoWishList = new CryptoWishList();
            this.dhgexpara = new DHGexParameters();
            this.connectionMonitors = new System.Collections.ArrayList(10);
            this.hostName = hostName;
            this.port = port;
        }

        /// <summary>
        /// Gets the hostName.
        /// </summary>
        /// <returns>The hostName.
        /// </returns>
        public string HostName
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                return this.hostName;
            }
        }

        /// <summary>
        /// Gets the port.
        /// </summary>
        /// <returns>The TCP port number.
        /// </returns>
        public int Port
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                return this.port;
            }
        }

        /// <summary> Gets a <see cref="Eunomia.SSH2.ConnectionInfo"></see> object containing the details of
        /// the connection.  Can be called as soon as the connection has been
        /// established (successfully connected).
        /// </summary>
        /// <returns> A <see cref="Eunomia.SSH2.ConnectionInfo"></see> object.
        /// </returns>
        /// <exception cref="InvalidOperationException">If the connection has not been established.</exception>
        public ConnectionInfo ConnectionInfo
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                if (this.tm == null)
                {
                    return null;
                }

                return this.tm.GetConnectionInfo(1);
            }
        }

        /// <summary>
        /// Gets a value indicating whether the authentication phase is complete. Can be called at any time.
        /// </summary>
        /// <returns>
        /// <c>true</c> if no further authentication steps are needed.
        /// </returns>
        public bool AuthenticationComplete
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                return this.authenticated;
            }
        }

        /// <summary>
        /// Gets a value indicating whether there was at least one failed authentication request and
        /// the last failed authentication request was marked with "partial success"
        /// by the server. This is only needed in the rare case of SSH-2 server setups
        /// that cannot be satisfied with a single successful authentication request
        /// (i.e., multiple authentication steps are needed.)
        /// <para>
        /// If you are interested in the details, then have a look at RFC4252.
        /// </para>
        /// </summary>
        /// <returns>
        /// True if the there was a failed authentication step and the last one
        /// was marked as a "partial success".
        /// </returns>
        public bool AuthenticationPartialSuccess
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                if (this.am == null)
                {
                    return false;
                }

                return this.am.PartialSuccess;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether TCP_NODELAY (disable/enable Nagle's algorithm)
        /// is enabled on the underlying socket.
        /// Can be called at any time. If the connection has not yet been established
        /// then the passed value will be stored and set after the socket has been set up.
        /// The default value that will be used is <c>false</c>.
        /// </summary>
        public bool TCPNoDelay
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                return this.tm.TcpNoDelay;
            }

            [MethodImpl(MethodImplOptions.Synchronized)]
            set
            {
                this.tcpNoDelay = value;
                if (this.tm != null)
                {
                    this.tm.TcpNoDelay = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the ProxyData, which determines whether the connection should be
        /// established through a proxy server.  It only makes sense to call this method before
        /// calling the <see cref="Connection"></see> method.
        /// <para>
        /// At the moment, only HTTP proxies are supported.
        /// </para>
        /// Note: This method can be called any number of times. The <see cref="Connect()"></see>
        /// method will use the value set in the last preceding invocation of this method
        /// <seealso cref="HTTPProxyData"></seealso>.
        /// </summary>
        public ProxyData ProxyData
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                return this.proxyData;
            }

            [MethodImpl(MethodImplOptions.Synchronized)]
            set
            {
                this.proxyData = value;
            }
        }

        /// <summary>
        /// Gets or sets your own instance of SecureRandom. Can be used, e.g., if you
        /// want to seed the used SecureRandom generator manually.
        /// <para>
        /// The SecureRandom instance is used during key exchanges, public key authentication,
        /// x11 cookie generation and the like.
        /// </para>
        /// </summary>
        /// <exception cref="ArgumentException">If the SecureRandom instance is null.</exception>
        public SecureRandom SecureRandom
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                return this.generator;
            }

            [MethodImpl(MethodImplOptions.Synchronized)]
            set
            {
                if (value == null)
                {
                    throw new ArgumentException("SecureRandom was passed a null argument.");
                }

                this.generator = value;
            }
        }

        /// <summary>
        /// Gets or creates and gets a <c>SecureRandom</c> instance for use with this connection.
        /// </summary>
        private SecureRandom ConnectionSecureRND
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                if (this.generator == null)
                {
                    this.generator = new SecureRandom();
                }

                return this.generator;
            }
        }

        /// <summary>
        /// Gets a string array containing the supported cipher algorithms.
        /// </summary>
        /// <remarks>
        /// Unless you know what you are doing, you will never need this.
        /// </remarks>
        /// <returns>
        /// A string array containing the supported cipher algorithms by this implementation.
        /// </returns>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public static string[] GetAvailableCiphers()
        {
            return BlockCipherFactory.GetDefaultCipherList();
        }

        /// <summary>
        /// Gets a string array containing the supported MAC algorithms.
        /// </summary>
        /// <remarks>
        /// Unless you know what you are doing, you will never need this.
        /// </remarks>
        /// <returns>
        /// A string array containing the supported MAC algorthims by this implementation.
        /// </returns>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public static string[] AvailableMACs()
        {
            return MAC.GetMacList();
        }

        /// <summary>
        /// Gets a string array containing the supported server host key algorithms.
        /// </summary>
        /// <remarks>
        /// Unless you know what you are doing, you will never need this.
        /// </remarks>
        /// <returns>
        /// A string array containing the supported server host key algorthims by this implementation.
        /// </returns>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public static string[] AvailableServerHostkeyAlgorithms()
        {
            return KexManager.DefaultServerHostkeyAlgorithmList;
        }

        /// <summary>
        /// Sets the available client to server ciphers (passed by a string array).
        /// </summary>
        /// <remarks>
        /// Unless you know what you are doing, you will never need this.
        /// </remarks>
        /// <param name="value">
        /// A string array of client to server ciphers.
        /// </param>
        /// <returns>
        /// A cleaned up string array of valid client to server ciphers.
        /// </returns>
        /// <exception cref="ArgumentException">If the string array is null or zero length.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public string[] SetClient2ServerCiphers(string[] value)
        {
            if ((value == null) || (value.Length == 0))
            {
                throw new ArgumentException("SetClient2ServerCiphers was passed a null or empty string array.");
            }

            value = Connection.RemoveDuplicates(value);
            BlockCipherFactory.CheckCipherList(value);
            this.cryptoWishList.C2SEncAlgos = value;
            return value;
        }

        /// <summary>
        /// Sets the available client to server Macs (passed by a string array).
        /// </summary>
        /// <remarks>
        /// Unless you know what you are doing, you will never need this.
        /// </remarks>
        /// <param name="value">
        /// A string array of client to server MACs.
        /// </param>
        /// <returns>
        /// A cleaned up string array of valid client to server MACs.
        /// </returns>
        /// <exception cref="ArgumentException">If the string array is null or zero length.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public string[] SetClient2ServerMACs(string[] value)
        {
            if ((value == null) || (value.Length == 0))
            {
                throw new ArgumentException("SetClient2ServerMACs was passed a null or empty string array.");
            }

            value = Connection.RemoveDuplicates(value);
            MAC.CheckMacList(value);
            this.cryptoWishList.C2SMacAlgos = value;
            return value;
        }

        /// <summary>
        /// Sets the parameters for the diffie-hellman group exchange. Unless you
        /// know what you are doing, you will never need this. Default values are
        /// defined in the <see cref="DHGexParameters"></see> class.
        /// </summary>
        /// <param name="value">The DHGexParameters instance to use.</param>
        /// <exception cref="ArgumentException">If the DHGexParameters is null.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void SetDHGexParameters(DHGexParameters value)
        {
            if (value == null)
            {
                throw new ArgumentException("SetDHGexParameters was passed a null or empty value.");
            }

            this.dhgexpara = value;
        }

        /// <summary>
        /// Sets the available server to client ciphers (passed by a string array).
        /// </summary>
        /// <remarks>
        /// Unless you know what you are doing, you will never need this.
        /// </remarks>
        /// <param name="value">
        /// A string array of server to client ciphers.
        /// </param>
        /// <returns>
        /// A cleaned up string array of valid server to client ciphers.
        /// </returns>
        /// <exception cref="ArgumentException">If the string array is null or zero length.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public string[] SetServer2ClientCiphers(string[] value)
        {
            if ((value == null) || (value.Length == 0))
            {
                throw new ArgumentException("SetServer2ClientCiphers was passed a null or empty string array.");
            }

            value = Connection.RemoveDuplicates(value);
            BlockCipherFactory.CheckCipherList(value);
            this.cryptoWishList.S2CEncAlgos = value;
            return value;
        }

        /// <summary>
        /// Sets the available server to client Macs (passed by a string array).
        /// </summary>
        /// <remarks>
        /// Unless you know what you are doing, you will never need this.
        /// </remarks>
        /// <param name="value">
        /// A string array of server to client MACs.
        /// </param>
        /// <returns>
        /// A cleaned up string array of valid server to client MACs.
        /// </returns>
        /// <exception cref="ArgumentException">If the string array is null or zero length.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public string[] SetServer2ClientMACs(string[] value)
        {
            if ((value == null) || (value.Length == 0))
            {
                throw new ArgumentException("SetServer2ClientMACs was passed a null or empty string array.");
            }

            value = Connection.RemoveDuplicates(value);
            MAC.CheckMacList(value);
            this.cryptoWishList.S2CMacAlgos = value;
            return value;
        }

        /// <summary>
        /// Sets the allowed server host key algorithms to be used for
        /// the following key exchange operations. Unless you know what you
        /// are doing, you will never need this.
        /// </summary>
        /// <remarks>
        /// An array of allowed server host key algorithms.
        /// SSH-2 defines <c>ssh-dss</c> and <c>ssh-rsa</c>.
        /// The entries of the array must be ordered after preference, i.e.,
        /// the entry at index 0 is the most preferred one. You must specify
        /// at least one entry.
        /// </remarks>
        /// <param name="value">
        /// A string array of server host key alogorithms.
        /// </param>
        /// <returns>
        /// A cleaned up string array of valid server host key alogorithms.
        /// </returns>
        /// <exception cref="ArgumentException">If the string array is null or zero length.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public string[] ServerHostkeyAlgorithms(string[] value)
        {
            if ((value == null) || (value.Length == 0))
            {
                throw new ArgumentException("ServerHostkeyAlgorithms was passed a null or empty string array.");
            }

            value = Connection.RemoveDuplicates(value);
            KexManager.CheckServerHostkeyAlgorithmsList(value);
            this.cryptoWishList.ServerHostkeyAlgorithms = value;
            return value;
        }

        /// <summary>
        /// After a successful connect, one has to authenticate oneself. This method
        /// is based on DSA (it uses DSA to sign a challenge sent by the server).
        /// <para>
        /// If the authentication phase is complete, <c>true</c> will be
        /// returned. If the server does not accept the request (or if further
        /// authentication steps are needed), <c>false</c> is returned and
        /// one can retry either by using this or any other authentication method
        /// (use the <c>GetRemainingAuthMethods</c> method to get a list of
        /// the remaining possible methods).
        /// </para>
        /// </summary>
        /// <param name="user">
        /// A <c>string</c> holding the username.
        /// </param>
        /// <param name="pem">
        /// A <c>string</c> containing the DSA private key of the
        /// user in OpenSSH key format (PEM, you can't miss the
        /// "-----BEGIN DSA PRIVATE KEY-----" tag). The string may contain
        /// linefeeds.
        /// </param>
        /// <param name="password">
        /// If the PEM string is 3DES encrypted ("DES-EDE3-CBC"), then you
        /// must specify the password. Otherwise, this argument will be
        /// ignored and can be set to <c>null</c>.
        /// </param>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <seealso cref="GetRemainingAuthMethods"></seealso>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        /// <exception cref="ArgumentException">If the user is null.</exception>
        /// <exception cref="IOException">If AuthenticationManager has a problem.</exception>
        /// <deprecated> You should use one of the {AuthenticateWithPublicKey(string, FileInfo, string) or
        /// AuthenticateWithPublicKey()} methods, this method is just a wrapper for it and will
        /// disappear in future builds.
        /// </deprecated>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool AuthenticateWithDSA(string user, string pem, string password)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Connection is not established!");
            }

            if (this.authenticated)
            {
                throw new InvalidOperationException("Connection is already authenticated!");
            }

            if (this.am == null)
            {
                this.am = new Eunomia.SSH2.Auth.AuthenticationManager(this.tm);
            }

            if (this.cm == null)
            {
                this.cm = new ChannelManager(this.tm);
            }

            if ((object)user == null)
            {
                throw new ArgumentException("user argument is null");
            }

            if ((object)pem == null)
            {
                throw new ArgumentException("pem argument is null");
            }

            this.authenticated = this.am.AuthenticatePublicKey(user, pem.ToCharArray(), password, this.ConnectionSecureRND);

            return this.authenticated;
        }

        /// <summary>
        /// A wrapper that calls AuthenticateWithKeyboardInteractive(String, String[], InteractiveCallback)
        /// a <c>null</c> submethod list.
        /// </summary>
        /// <param name="user">
        /// A string holding the username.
        /// </param>
        /// <param name="callback">
        /// An InteractiveCallback which will be used to
        /// determine the responses to the questions asked by the server.
        /// </param>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <seealso cref="GetRemainingAuthMethods"></seealso>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        /// <exception cref="ArgumentException">If the user is null.</exception>
        /// <exception cref="IOException">If AuthenticationManager has a problem.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool AuthenticateWithKeyboardInteractive(string user, IInteractiveCallback callback)
        {
            return this.AuthenticateWithKeyboardInteractive(user, null, callback);
        }

        /// <summary>
        /// After a successful connect, one has to authenticate oneself. This method
        /// is based on "keyboard-interactive", specified in
        /// draft-ietf-secsh-auth-kbdinteract-XX. Basically, you have to define a
        /// callback object which will be feeded with challenges generated by the
        /// server. Answers are then sent back to the server. It is possible that the
        /// callback will be called several times during the invocation of this
        /// method (e.g., if the server replies to the callback's answer(s) with
        /// another challenge...)
        /// <para>
        /// If the authentication phase is complete, <c>true</c> will be
        /// returned. If the server does not accept the request (or if further
        /// authentication steps are needed), <c>false</c> is returned and
        /// one can retry either by using this or any other authentication method
        /// (use the <c>getRemainingAuthMethods</c> method to get a list of
        /// the remaining possible methods).
        /// </para>
        /// <para>
        /// Note: some SSH servers advertise "keyboard-interactive", however, any
        /// interactive request will be denied (without having sent any challenge to
        /// the client).
        /// </para>
        /// </summary>
        /// <param name="user">
        /// A string holding the username.
        /// </param>
        /// <param name="subMethods">
        /// An array of submethod names, see
        /// draft-ietf-secsh-auth-kbdinteract-XX. May be null
        /// to indicate an empty list.
        /// </param>
        /// <param name="callback">
        /// An InteractiveCallback which will be used to
        /// determine the responses to the questions asked by the server.
        /// </param>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <seealso cref="GetRemainingAuthMethods"></seealso>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        /// <exception cref="ArgumentException">If the user is null.</exception>
        /// <exception cref="IOException">If AuthenticationManager has a problem.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool AuthenticateWithKeyboardInteractive(
            string user,
            string[] subMethods,
            IInteractiveCallback callback)
        {
            if (callback == null)
            {
                throw new ArgumentException("Callback may not ne NULL!");
            }

            if (this.tm == null)
            {
                throw new InvalidOperationException("Connection is not established!");
            }

            if (this.authenticated)
            {
                throw new InvalidOperationException("Connection is already authenticated!");
            }

            if (this.am == null)
            {
                this.am = new Eunomia.SSH2.Auth.AuthenticationManager(this.tm);
            }

            if (this.cm == null)
            {
                this.cm = new ChannelManager(this.tm);
            }

            if ((object)user == null)
            {
                throw new ArgumentException("user argument is null");
            }

            this.authenticated = this.am.AuthenticateInteractive(user, subMethods, callback);

            return this.authenticated;
        }

        /// <summary>
        /// After a successful connect, one has to authenticate oneself. This method
        /// sends username and password to the server.
        /// <para>
        /// If the authentication phase is complete, <c>true</c> will be
        /// returned. If the server does not accept the request (or if further
        /// authentication steps are needed), <c>false</c> is returned and
        /// one can retry either by using this or any other authentication method
        /// (use the <c>getRemainingAuthMethods</c> method to get a list of
        /// the remaining possible methods).
        /// </para>
        /// <para>
        /// Note: if this method fails, then please double-check that it is actually
        /// offered by the server (use {@link #getRemainingAuthMethods(String) getRemainingAuthMethods()}.
        /// </para>
        /// <para>
        /// Often, password authentication is disabled, but users are not aware of it.
        /// Many servers only offer "publickey" and "keyboard-interactive". However,
        /// even though "keyboard-interactive" *feels* like password authentication
        /// (e.g., when using the putty or openssh clients) it is *not* the same mechanism.
        /// </para>
        /// </summary>
        /// <param name="user">The user to authenticate.
        /// </param>
        /// <param name="password">The password to use.
        /// </param>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <seealso cref="GetRemainingAuthMethods"></seealso>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        /// <exception cref="ArgumentException">If the user is null.</exception>
        /// <exception cref="IOException">If AuthenticationManager has a problem.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool AuthenticateWithPassword(string user, string password)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Connection is not established!");
            }

            if (this.authenticated)
            {
                throw new InvalidOperationException("Connection is already authenticated!");
            }

            if (this.am == null)
            {
                this.am = new Eunomia.SSH2.Auth.AuthenticationManager(this.tm);
            }

            if (this.cm == null)
            {
                this.cm = new ChannelManager(this.tm);
            }

            if ((object)user == null)
            {
                throw new ArgumentException("user argument is null");
            }

            if ((object)password == null)
            {
                throw new ArgumentException("password argument is null");
            }

            this.authenticated = this.am.AuthenticatePassword(user, password);

            return this.authenticated;
        }

        /// <summary>
        /// After a successful connect, one has to authenticate oneself.
        /// This method can be used to explicitly use the special "none"
        /// authentication method (where only a username has to be specified).
        /// <para>
        /// Note 1: The "none" method may always be tried by clients, however as by
        /// the specs, the server will not explicitly announce it. In other words,
        /// the "none" token will never show up in the list returned by
        /// <c>AuthenticationManager.GetRemainingMethods</c>
        /// </para>
        /// <para>
        /// Note 2: no matter which one of the AuthenticateWithXXX() methods
        /// you call, the library will always issue exactly one initial "none"
        /// authentication request to retrieve the initially allowed list of
        /// authentication methods by the server. Please read RFC 4252 for the
        /// details.
        /// </para>
        /// <para>
        /// If the authentication phase is complete, <c>true</c> will be
        /// returned. If further authentication steps are needed, <c>false</c>
        /// is returned and one can retry by any other authentication method
        /// (use the <c>GetRemainingAuthMethods</c> method to get a list of
        /// the remaining possible methods).
        /// </para>
        /// </summary>
        /// <see cref="GetRemainingAuthMethods"></see>
        /// <param name="user">The user name to authenticate.
        /// </param>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <seealso cref="GetRemainingAuthMethods"></seealso>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        /// <exception cref="ArgumentException">If the user is null.</exception>
        /// <exception cref="IOException">If AuthenticationManager has a problem.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool AuthenticateWithNone(string user)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Connection is not established!");
            }

            if (this.authenticated)
            {
                throw new InvalidOperationException("Connection is already authenticated!");
            }

            if (this.am == null)
            {
                this.am = new Eunomia.SSH2.Auth.AuthenticationManager(this.tm);
            }

            if (this.cm == null)
            {
                this.cm = new ChannelManager(this.tm);
            }

            if ((object)user == null)
            {
                throw new ArgumentException("user argument is null");
            }

            // Trigger the sending of the PacketUserauthRequestNone packet
            // (if not already done)
            this.authenticated = this.am.AuthenticateNone(user);

            return this.authenticated;
        }

        /// <summary>
        /// After a successful connect, one has to authenticate oneself.
        /// The authentication method "publickey" works by signing a challenge
        /// sent by the server. The signature is either DSA or RSA based - it
        /// just depends on the type of private key you specify, either a DSA
        /// or RSA private key in PEM format. And yes, this is may seem to be a
        /// little confusing, the method is called "publickey" in the SSH-2 protocol
        /// specification, however since we need to generate a signature, you
        /// actually have to supply a private key =).
        /// <para>
        /// The private key contained in the PEM file may also be encrypted ("Proc-Type: 4,ENCRYPTED").
        /// The library supports DES-CBC and DES-EDE3-CBC encryption, as well
        /// as the more exotic PEM encrpytions AES-128-CBC, AES-192-CBC and AES-256-CBC.
        /// </para>
        /// <para>
        /// If the authentication phase is complete, <c>true</c> will be
        /// returned. If the server does not accept the request (or if further
        /// authentication steps are needed), <c>false</c> is returned and
        /// one can retry either by using this or any other authentication method
        /// (use the <c>GetRemainingAuthMethods</c> method to get a list of
        /// the remaining possible methods).
        /// </para>
        /// <para>
        /// NOTE PUTTY USERS: Even though your key file may start with "-----BEGIN..."
        /// it is not in the expected format. You have to convert it to the OpenSSH
        /// key format by using the "puttygen" tool (can be downloaded from the Putty
        /// website). Simply load your key and then use the "Conversions/Export OpenSSH key"
        /// functionality to get a proper PEM file.
        /// </para>
        /// </summary>
        /// <param name="user">
        /// A <c>string</c> holding the username.
        /// </param>
        /// <param name="pemPrivateKey">
        /// A <c>char[]</c> containing a DSA or RSA private key of the
        /// user in OpenSSH key format (PEM, you can't miss the
        /// "-----BEGIN DSA PRIVATE KEY-----" or "-----BEGIN RSA PRIVATE KEY-----"
        /// tag). The char array may contain linebreaks/linefeeds.
        /// </param>
        /// <param name="password">
        /// If the PEM structure is encrypted ("Proc-Type: 4,ENCRYPTED") then
        /// you must specify a password. Otherwise, this argument will be ignored
        /// and can be set to <c>null</c>.
        /// </param>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <seealso cref="GetRemainingAuthMethods"></seealso>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        /// <exception cref="ArgumentException">If the user is null.</exception>
        /// <exception cref="IOException">If AuthenticationManager has a problem.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool AuthenticateWithPublicKey(string user, char[] pemPrivateKey, string password)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Connection is not established!");
            }

            if (this.authenticated)
            {
                throw new InvalidOperationException("Connection is already authenticated!");
            }

            if (this.am == null)
            {
                this.am = new Eunomia.SSH2.Auth.AuthenticationManager(this.tm);
            }

            if (this.cm == null)
            {
                this.cm = new ChannelManager(this.tm);
            }

            if ((object)user == null)
            {
                throw new ArgumentException("user argument is null");
            }

            if (pemPrivateKey == null)
            {
                throw new ArgumentException("pemPrivateKey argument is null");
            }

            this.authenticated = this.am.AuthenticatePublicKey(user, pemPrivateKey, password, this.ConnectionSecureRND);

            return this.authenticated;
        }

        /// <summary>
        /// A convenience wrapper function which reads in a private key (PEM format, either DSA or RSA)
        /// and then calls <c>AuthenticateWithPublicKey(string, char[], string)</c>.
        /// <para>
        /// NOTE PUTTY USERS: Event though your key file may start with "-----BEGIN..."
        /// it is not in the expected format. You have to convert it to the OpenSSH
        /// key format by using the "puttygen" tool (can be downloaded from the Putty
        /// website). Simply load your key and then use the "Conversions/Export OpenSSH key"
        /// functionality to get a proper PEM file.
        /// </para>
        /// </summary>
        /// <param name="user">
        /// A <c>string</c> holding the username.
        /// </param>
        /// <param name="pemFile">
        /// A <c>FileInfo</c> object pointing to a file containing a DSA or RSA
        /// private key of the user in OpenSSH key format (PEM, you can't miss the
        /// "-----BEGIN DSA PRIVATE KEY-----" or "-----BEGIN RSA PRIVATE KEY-----"
        /// tag).
        /// </param>
        /// <param name="password">
        /// If the PEM file is encrypted then you must specify the password.
        /// Otherwise, this argument will be ignored and can be set to <c>null</c>.
        /// </param>
        /// <returns>True if the connection is now authenticated.
        /// </returns>
        /// <seealso cref="GetRemainingAuthMethods"></seealso>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        /// <exception cref="ArgumentException">If the user is null.</exception>
        /// <exception cref="IOException">If AuthenticationManager has a problem.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool AuthenticateWithPublicKey(string user, FileSystemInfo pemFile, string password)
        {
            if (pemFile == null)
            {
                throw new ArgumentException("pemFile argument is null");
            }

            char[] buff = new char[256];

            StreamWriter cw = new StreamWriter(new MemoryStream());
            StreamReader fr = new StreamReader(pemFile.FullName);

            while (true)
            {
                int len = fr.Read((char[])buff, 0, buff.Length);
                if (len < 0)
                {
                    break;
                }
                
                cw.Write(buff, 0, len);
            }

            fr.Close();

            return this.AuthenticateWithPublicKey(user, cw.ToString().ToCharArray(), password);
        }

        /// <summary>
        /// Add a <c>IConnectionMonitor</c> to this connection. Can be invoked at any time,
        /// but it is best to add connection monitors before invoking
        /// <c>Connect()</c> to avoid glitches (e.g., you add a connection monitor after
        /// a successful Connect(), but the connection has died in the mean time. Then,
        /// your connection monitor won't be notified.)
        /// <para>
        /// You can add as many monitors as you like.
        /// </para>
        /// </summary>
        /// <seealso cref="IConnectionMonitor"></seealso>
        /// <param name="connMonitor">An object implementing the <c>IConnectionMonitor</c> interface.
        /// </param>
        /// <exception cref="ArgumentException">If the passed connection monitor is null.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void AddConnectionMonitor(IConnectionMonitor connMonitor)
        {
            if (connMonitor == null)
            {
                throw new ArgumentException("connMonitor argument is null");
            }

            this.connectionMonitors.Add(connMonitor);

            if (this.tm != null)
            {
                this.tm.ReplaceConnectionMonitors(this.connectionMonitors);
            }
        }

        /// <summary>
        /// Close the connection to the SSH-2 server. All assigned sessions will be
        /// closed, too. Can be called at any time. Don't forget to call this once
        /// you don't need a connection anymore - otherwise the receiver thread may
        /// run forever.
        /// </summary>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void Close()
        {
            Exception t = new IOException("Closed due to user request.");
            this.Close(t, false);
        }

        /// <summary>
        /// Same as <c>Connect(null, 0, 0)}</c>.
        /// </summary>
        /// <returns>See comments for the Connect(ServerHostKeyVerifier, int, int)} method.
        /// </returns>
        /// <seealso cref="Connect(IServerHostkeyVerifier, int, int)"></seealso>
        /// <exception cref="IOException">If we are already connected, or cannot conect to the host and port.</exception>
        /// <exception cref="SocketException">If there is an exception raised by the socket.</exception>
        /// <exception cref="IOException">If the key exchange times out.</exception>
        /// <exception cref="HttpProxyException">If the http proxy raises an exception.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public ConnectionInfo Connect()
        {
            return this.Connect(null, 0, 0);
        }

        /// <summary>
        /// Same as <c>Connect(verifier, 0, 0)}.</c>
        /// </summary>
        /// <param name="verifier">An instance of a IServerHostKeyVerifier.</param>
        /// <returns>See comments for the Connect(ServerHostKeyVerifier, int, int)} method.
        /// </returns>
        /// <seealso cref="Connect(IServerHostkeyVerifier, int, int)"></seealso>
        /// <exception cref="IOException">If we are already connected, or cannot conect to the host and port.</exception>
        /// <exception cref="SocketException">If there is an exception raised by the socket.</exception>
        /// <exception cref="IOException">If the key exchange times out.</exception>
        /// <exception cref="HttpProxyException">If the http proxy raises an exception.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public ConnectionInfo Connect(IServerHostkeyVerifier verifier)
        {
            return this.Connect(verifier, 0, 0);
        }

        /// <summary>
        /// Connect to the SSH-2 server and, as soon as the server has presented its
        /// host key, use the IServerHostKeyVerifier.VerifyServerHostkey(string,
        /// int, string, byte[]) method of the <c>verifier</c> to ask for permission to proceed.
        /// If <c>verifier</c> is <c>null</c>, then any host key will be
        /// accepted - this is NOT recommended, since it makes man-in-the-middle attackes
        /// VERY easy (somebody could put a proxy SSH server between you and the real server).
        /// <para>
        /// Note: The verifier will be called before doing any crypto calculations
        /// (i.e., diffie-hellman). Therefore, if you don't like the presented host key then
        /// no CPU cycles are wasted (and the evil server has less information about us).
        /// </para>
        /// <para>
        /// However, it is still possible that the server presented a fake host key: the server
        /// cheated (typically a sign for a man-in-the-middle attack) and is not able to generate
        /// a signature that matches its host key. Don't worry, the library will detect such
        /// a scenario later when checking the signature (the signature cannot be checked before
        /// having completed the diffie-hellman exchange).
        /// </para>
        /// <para>
        /// Note 2: The IServerHostKeyVerifier.VerifyServerHostkey(string,
        /// int, string, byte[]) method will *NOT* be called from the current thread, the call is being made
        /// from a background thread (there is a background dispatcher thread for every established connection).
        /// </para>
        /// <para>
        /// Note 3: This method will block as long as the key exchange of the underlying connection
        /// has not been completed (and you have not specified any timeouts).
        /// </para>
        /// <para>
        /// Note 4: If you want to re-use a connection object that was successfully connected,
        /// then you must call the <c>Close</c> method before invoking <c>Connect()</c> again.
        /// </para>
        /// </summary>
        /// <param name="verifier">
        /// An object that implements the <c>IServerHostKeyVerifier</c> interface. Pass <c>null</c>
        /// to accept any server host key - NOT recommended.
        /// </param>
        /// <param name="connectTimeout">
        /// Connect the underlying TCP socket to the server with the given timeout
        /// value (non-negative, in milliseconds). Zero means no timeout. If a proxy is being
        /// used <c>SetProxyData(ProxyData)</c>, then this timeout is used for the
        /// connection establishment to the proxy.
        /// </param>
        /// <param name="kexTimeout">
        /// Timeout for complete connection establishment (non-negative,
        /// in milliseconds). Zero means no timeout. The timeout counts from the
        /// moment you invoke the connect() method and is cancelled as soon as the
        /// first key-exchange round has finished. It is possible that
        /// the timeout event will be fired during the invocation of the
        /// <c>verifier</c> callback, but it will only have an effect after
        /// the <c>verifier</c> returns.
        /// </param>
        /// <returns> A <c>ConnectionInfo</c> object containing the details of
        /// the established connection.  If any problem occurs, e.g., the server's host key is not
        /// accepted by the <c>verifier</c> or there is problem during
        /// the initial crypto setup (e.g., the signature sent by the server is wrong).
        /// </returns>
        /// <exception cref="ArgumentException">
        /// If either of the specified timeouts is negative.
        /// </exception>
        /// <exception cref="SocketException">
        /// In case of a timeout (either connectTimeout or kexTimeout)
        /// a SocketException is thrown.
        /// </exception>
        /// <exception cref="IOException">
        /// An IOException may also be thrown if the connection was already successfully
        /// connected (no matter if the connection broke in the mean time) and you invoke
        /// <c>Connect()</c> again without having called <c>Close</c> first.
        /// </exception>
        /// <exception cref="HttpProxyException">
        /// If a HTTP proxy is being used and the proxy refuses the connection,
        /// then a <c>HTTPProxyException</c> may be thrown, which
        /// contains the details returned by the proxy. If the proxy is buggy and does
        /// not return a proper HTTP response, then a normal IOException is thrown instead.
        /// </exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public ConnectionInfo Connect(IServerHostkeyVerifier verifier, int connectTimeout, int kexTimeout)
        {
            if (this.tm != null)
            {
                throw new IOException("Connection to " + this.hostName + " is already in connected state!");
            }

            if (connectTimeout < 0)
            {
                throw new ArgumentException("connectTimeout must be non-negative!");
            }

            if (kexTimeout < 0)
            {
                throw new ArgumentException("kexTimeout must be non-negative!");
            }

            TimeoutState state = new TimeoutState();
            this.tm = new TransportManager(this.hostName, this.port);
            this.tm.ReplaceConnectionMonitors(this.connectionMonitors);

            // Make sure that the runnable below will observe the new value of "tm"
            // and "state" (the runnable will be executed in a different thread, which
            // may be already running, that is why we need a memory barrier here).
            // See also the comment in Channel.java if you
            // are interested in the details.
            //
            // OKOK, this is paranoid since adding the runnable to the todo list
            // of the TimeoutService will ensure that all writes have been flushed
            // before the Runnable reads anything
            // (there is a synchronized block in TimeoutService.addTimeoutHandler).
            // Not sure if this is neccessary...so for now it is removed (unlike ganymed).
            // lock (this.tm)
            // {
            // We could actually synchronize on anything.
            // }
            try
            {
                TimeoutToken token = null;

                if (kexTimeout > 0)
                {
                    IThreadRunnable timeoutHandler = new TimeoutHandler(state, this);

                    // divide ticks by 10000 to get milliseconds
                    long timeoutHorizont = (DateTime.Now.Ticks / 10000) + kexTimeout;
                    token = TimeoutService.AddTimeoutHandler(timeoutHorizont, timeoutHandler);
                }

                try
                {
                    this.tm.Initialize(this.cryptoWishList, verifier, this.dhgexpara, connectTimeout, this.ConnectionSecureRND, this.proxyData);
                }
                catch (SocketException se)
                {
                    throw new SocketException(se.ErrorCode);
                }

                this.tm.TcpNoDelay = this.tcpNoDelay;

                // Wait until first KEX has finished
                ConnectionInfo ci = this.tm.GetConnectionInfo(1);

                // Now try to cancel the timeout, if needed
                if (token != null)
                {
                    TimeoutService.CancelTimeoutHandler(token);

                    // Were we too late?
                    lock (state)
                    {
                        if (state.TimeoutSocketClosed)
                        {
                            throw new IOException("This exception will be replaced by the one below =)");
                        }

                        // Just in case the "cancelTimeoutHandler" invocation came just a little bit
                        // too late but the handler did not enter the semaphore yet - we can
                        // still stop it.
                        state.IsCancelled = true;
                    }
                }

                return ci;
            }
            catch (SocketException)
            {
                throw;
            }
            catch (IOException e1)
            {
                // This will also invoke any registered connection monitors
                this.Close(new IOException("There was a problem during connect."), false);

                lock (state)
                {
                    // Show a clean exception, not something like "the socket is closed!?!"
                    if (state.TimeoutSocketClosed)
                    {
                        throw new IOException("The kexTimeout (" + kexTimeout + " ms) expired.");
                    }
                }

                // Do not wrap a HTTPProxyException
                if (e1 is HttpProxyException)
                {
                    throw;
                }

                throw (IOException)new IOException("There was a problem while connecting to " + this.hostName + ":" + this.port);
            }
        }

        /// <summary>
        /// Creates a new <c>LocalPortForwarder</c>.
        /// A <c>LocalPortForwarder</c> forwards TCP/IP connections that arrive at a local
        /// port via the secure tunnel to another host (which may or may not be
        /// identical to the remote SSH-2 server).
        /// <para>
        /// This method must only be called after one has passed successfully the authentication step.
        /// There is no limit on the number of concurrent forwardings.
        /// </para>
        /// </summary>
        /// <param name="localPort">
        /// The local port the LocalPortForwarder shall bind to.
        /// </param>
        /// <param name="hostToConnect">
        /// Target address (IP or hostName).
        /// </param>
        /// <param name="portToConnect">
        /// Target port.
        /// </param>
        /// <returns> A LocalPortForwarder object.
        /// </returns>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or not authenticated.
        /// </exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public LocalPortForwarder CreateLocalPortForwarder(
            int localPort,
            string hostToConnect,
            int portToConnect)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Cannot forward ports, you need to establish a connection first.");
            }

            if (!this.authenticated)
            {
                throw new InvalidOperationException("Cannot forward ports, connection is not authenticated.");
            }

            return new LocalPortForwarder(this.cm, localPort, hostToConnect, portToConnect);
        }

        /// <summary>
        /// Creates a new <c>LocalPortForwarder</c>.
        /// A <c>LocalPortForwarder</c> forwards TCP/IP connections that arrive at a local
        /// port via the secure tunnel to another host (which may or may not be
        /// identical to the remote SSH-2 server).
        /// <para>
        /// This method must only be called after one has passed successfully the authentication step.
        /// There is no limit on the number of concurrent forwardings.
        /// </para>
        /// </summary>
        /// <param name="address">Specifies the IPAddress where the local socket shall be bound to.
        /// </param>
        /// <param name="hostToConnect">Target address (IP or hostName).
        /// </param>
        /// <param name="portToConnect">Target port.
        /// </param>
        /// <returns>A <c>LocalPortForwarder</c> object.
        /// </returns>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or not authenticated.
        /// </exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public LocalPortForwarder CreateLocalPortForwarder(
            IPAddress address,
            string hostToConnect,
            int portToConnect)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Cannot forward ports, you need to establish a connection first.");
            }

            if (!this.authenticated)
            {
                throw new InvalidOperationException("Cannot forward ports, connection is not authenticated.");
            }

            return new LocalPortForwarder(this.cm, address, hostToConnect, portToConnect);
        }

        /// <summary>
        /// Creates a new <c>LocalStreamForwarder</c>.
        /// A <c>LocalStreamForwarder</c> manages an Input/Outputstream pair
        /// that is being forwarded via the secure tunnel into a TCP/IP connection to another host
        /// (which may or may not be identical to the remote SSH-2 server).
        /// </summary>
        /// <param name="hostToConnect">The host to connect to.
        /// </param>
        /// <param name="portToConnect">The port to connect to.
        /// </param>
        /// <returns> A <c>LocalStreamForwarder</c> object.
        /// </returns>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or not authenticated.
        /// </exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public LocalStreamForwarder CreateLocalStreamForwarder(string hostToConnect, int portToConnect)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Cannot forward, you need to establish a connection first.");
            }

            if (!this.authenticated)
            {
                throw new InvalidOperationException("Cannot forward, connection is not authenticated.");
            }

            return new LocalStreamForwarder(this.cm, hostToConnect, portToConnect);
        }

        /// <summary> Create a very basic <c>SCPClient</c> that can be used to copy
        /// files from/to the SSH-2 server.
        /// <para>
        /// Works only after one has passed successfully the authentication step.
        /// There is no limit on the number of concurrent SCP clients.
        /// </para><para>
        /// Note: This factory method will probably disappear in the future.
        /// </para>
        /// </summary>
        /// <returns> A <c>SCPClient</c> object.
        /// </returns>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or not authenticated.
        /// </exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public SCPClient CreateSCPClient()
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Cannot create SCP client, you need to establish a connection first.");
            }

            if (!this.authenticated)
            {
                throw new InvalidOperationException("Cannot create SCP client, connection is not authenticated.");
            }

            return new SCPClient(this);
        }

        /// <summary>
        /// Force an asynchronous key re-exchange (the call does not block). The
        /// latest values set for MAC, Cipher and DH group exchange parameters will
        /// be used. If a key exchange is currently in progress, then this method has
        /// the only effect that the so far specified parameters will be used for the
        /// next (server driven) key exchange.
        /// <para>
        /// Note: This implementation will never start a key exchange (other than the initial one)
        /// unless you or the SSH-2 server ask for it.
        /// </para>
        /// </summary>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established.
        /// </exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void ForceKeyExchange()
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("You need to establish a connection first.");
            }

            this.tm.ForceKeyExchange(this.cryptoWishList, this.dhgexpara);
        }

        /// <summary>
        /// After a successful connect, one has to authenticate oneself. This method
        /// can be used to tell which authentication methods are supported by the
        /// server at a certain stage of the authentication process (for the given
        /// username).
        /// <para>
        /// Note 1: the username will only be used if no authentication step was done
        /// so far (it will be used to ask the server for a list of possible
        /// authentication methods by sending the initial "none" request). Otherwise,
        /// this method ignores the user name and returns a cached method list
        /// (which is based on the information contained in the last negative server response).
        /// </para><para>
        /// Note 2: the server may return method names that are not supported by this
        /// implementation.
        /// </para><para>
        /// After a successful authentication, this method must not be called
        /// anymore.
        /// </para>
        /// </summary>
        /// <param name="user">
        /// A <c>string</c> holding the username.
        /// </param>
        /// <returns>A (possibly emtpy) array holding authentication method names.
        /// </returns>
        /// <exception cref="ArgumentException">If the user name is null.</exception>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is already authenticated.
        /// </exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public string[] GetRemainingAuthMethods(string user)
        {
            if ((object)user == null)
            {
                throw new ArgumentException("user argument may not be NULL!");
            }

            if (this.tm == null)
            {
                throw new InvalidOperationException("Connection is not established!");
            }

            if (this.authenticated)
            {
                throw new InvalidOperationException("Connection is already authenticated!");
            }

            if (this.am == null)
            {
                this.am = new Eunomia.SSH2.Auth.AuthenticationManager(this.tm);
            }

            if (this.cm == null)
            {
                this.cm = new ChannelManager(this.tm);
            }

            return this.am.GetRemainingMethods(user);
        }

        /// <summary>
        /// Checks if a specified authentication method is available. This method is
        /// actually just a wrapper for <c>GetRemainingAuthMethods(string)</c>.
        /// </summary>
        /// <param name="user">
        /// A <c>string</c> holding the username.
        /// </param>
        /// <param name="method">
        /// An authentication method name (e.g., "publickey", "password",
        /// "keyboard-interactive") as specified by the SSH-2 standard.
        /// </param>
        /// <returns>True if the specified authentication method is currently available.
        /// </returns>
        /// <exception cref="ArgumentException">If the user name is null.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public bool IsAuthMethodAvailable(string user, string method)
        {
            if ((object)method == null)
            {
                throw new ArgumentException("method argument may not be NULL!");
            }

            string[] methods = this.GetRemainingAuthMethods(user);

            for (int i = 0; i < methods.Length; i++)
            {
                if (string.Compare(methods[i], method, StringComparison.Ordinal) == 0)
                {
                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Open a new <c>Session</c> on this connection. Works only after one has passed
        /// successfully the authentication step. There is no limit on the number of
        /// concurrent sessions.
        /// </summary>
        /// <returns> A <c>Session</c> object.
        /// </returns>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or not authenticated.
        /// </exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public Session OpenSession()
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("Cannot open session, you need to establish a connection first.");
            }

            if (!this.authenticated)
            {
                throw new InvalidOperationException("Cannot open session, connection is not authenticated.");
            }

            return new Session(this.cm, this.ConnectionSecureRND);
        }

        /// <summary>
        /// Send an SSHMessageIGNORE packet. This method will generate a random data attribute
        /// (length between 0 (invlusive) and 16 (exclusive) bytes, contents are random bytes).
        /// <para>
        /// This method must only be called once the connection is established.
        /// </para>
        /// </summary>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        /// <exception cref="InvalidOperationException">If the connection has not been established.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void SendIgnorePacket()
        {
            SecureRandom rnd = this.ConnectionSecureRND;

            byte[] data = new byte[rnd.Next(16)];

            this.SendIgnorePacket(data);
        }

        /// <summary>
        /// Send an SSHMessageIGNORE packet with the given data attribute.
        /// <para>
        /// This method must only be called once the connection is established.
        /// </para>
        /// </summary>
        /// <exception cref="ArgumentException">If the data passed is null</exception>
        /// <exception cref="InvalidOperationException">If the connection has not been established.</exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        /// <param name="data">The data to send in the ignore packet (normally 16 random bytes).</param>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void SendIgnorePacket(byte[] data)
        {
            if (data == null)
            {
                throw new ArgumentException("data argument must not be null.");
            }

            if (this.tm == null)
            {
                throw new InvalidOperationException("Cannot send SSHMessageIGNORE packet, you need to establish a connection first.");
            }

            PacketIgnore pi = new PacketIgnore();
            pi.Data = data;

            this.tm.SendMessage(pi.Payload);
        }

        /// <summary>
        /// Request a remote port forwarding.
        /// If successful, then forwarded connections will be redirected to the given target address.
        /// You can cancel a requested remote port forwarding by calling
        /// <c>CancelRemotePortForwarding(int)</c>.
        /// <para>
        /// A call of this method will block until the peer either agreed or disagreed to your request
        /// Note 1: this method typically fails if you
        /// <list type="bullet">
        /// <item>pass a port number for which the used remote user has not enough permissions (i.e., port
        /// &lt; 1024)</item>
        /// <item>or pass a port number that is already in use on the remote server</item>
        /// <item>or if remote port forwarding is disabled on the server.</item>
        /// </list>
        /// </para>
        /// <para>
        /// Note 2: (from the openssh man page): By default, the listening socket on the server will be
        /// bound to the loopback interface only. This may be overriden by specifying a bind address.
        /// Specifying a remote bind address will only succeed if the server's <c>GatewayPorts</c> option
        /// is enabled (see sshd_config(5)).
        /// </para>
        /// </summary>
        /// <param name="bindAddress">Address to bind to on the server:
        /// <list type="bullet">
        /// <item>""means that connections are to be accepted on all protocol families
        /// supported by the SSH implementation</item>
        /// <item>"0.0.0.0" means to listen on all IPv4 addresses</item>
        /// <item>"::" means to listen on all IPv6 addresses</item>
        /// <item>"localhost" means to listen on all protocol families supported by the SSH
        /// implementation on loopback addresses only, [RFC3330] and RFC3513]</item>
        /// <item>"127.0.0.1" and "::1" indicate listening on the loopback interfaces for
        /// IPv4 and IPv6 respectively.</item>
        /// </list>
        /// </param>
        /// <param name="bindPort">Port number to bind on the server (must be &gt; 0).
        /// </param>
        /// <param name="targetAddress">The target address (IP or hostName).
        /// </param>
        /// <param name="targetPort">The target port.
        /// </param>
        /// <exception cref="ArgumentException">If the bndAddress passed is null</exception>
        /// <exception cref="InvalidOperationException">
        /// If the connection has not been established or authenticated.</exception>
        /// <exception cref="IOException">In case of any failure behind the scenes.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void RequestRemotePortForwarding(string bindAddress, int bindPort, string targetAddress, int targetPort)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("You need to establish a connection first.");
            }

            if (!this.authenticated)
            {
                throw new InvalidOperationException("The connection is not authenticated.");
            }

            if (((object)bindAddress == null) || ((object)targetAddress == null) || (bindPort <= 0) || (targetPort <= 0))
            {
                throw new ArgumentException("RequestRemotePortForwarding was passed a null or empty address.");
            }

            this.cm.RequestGlobalForward(bindAddress, bindPort, targetAddress, targetPort);
        }

        /// <summary>
        /// Cancel an earlier requested remote port forwarding.
        /// Currently active forwardings will not be affected (e.g., disrupted).
        /// Note that further connection forwarding requests may be received until
        /// this method has returned.
        /// </summary>
        /// <param name="bindPort">The allocated port number on the server.
        /// </param>
        /// <exception cref="IOException">
        /// If the remote side refuses the cancel request or another low
        /// level error occurs (e.g., the underlying connection is closed)
        /// </exception>
        /// <exception cref="InvalidOperationException">
        /// If the connection is not established or is not authenticated.
        /// </exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void CancelRemotePortForwarding(int bindPort)
        {
            if (this.tm == null)
            {
                throw new InvalidOperationException("You need to establish a connection first.");
            }

            if (!this.authenticated)
            {
                throw new InvalidOperationException("The connection is not authenticated.");
            }

            this.cm.RequestCancelGlobalForward(bindPort);
        }
        
        /// <summary>
        /// Implement the public IDisposable interface method.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }
        
        /// <summary>
        /// Implement the protected IDisposable interface method.
        /// </summary>
        /// <param name="disposing">True if we are disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.Close();
            }
        }

        /// <summary>
        /// Removes duplicates from a string array, keeps only first occurence
        /// of each element. Does not destroy order of elements; can handle nulls.
        /// Uses a very efficient O(N^2) algorithm =) .
        /// </summary>
        /// <param name="list">A string array.
        /// </param>
        /// <returns>A cleaned string array (with duplicates eliminated).
        /// </returns>
        private static string[] RemoveDuplicates(string[] list)
        {
            if ((list == null) || (list.Length < 2))
            {
                return list;
            }

            string[] list2 = new string[list.Length];

            int count = 0;

            for (int i = 0; i < list.Length; i++)
            {
                bool duplicate = false;

                string element = list[i];

                for (int j = 0; j < count; j++)
                {
                    if (((object)element == null && (object)list2[j] == null)
                        || ((object)element != null && element.Equals(list2[j])))
                    {
                        duplicate = true;
                        break;
                    }
                }

                if (duplicate)
                {
                    continue;
                }

                list2[count++] = list[i];
            }

            if (count == list2.Length)
            {
                return list2;
            }

            string[] tmp = new string[count];
            Array.Copy((System.Array)list2, 0, (System.Array)tmp, 0, count);

            return tmp;
        }

        /// <summary>
        /// Close the connection to the SSH-2 server. All assigned sessions will be
        /// closed, too. Can be called at any time. Don't forget to call this once
        /// you don't need a connection anymore - otherwise the receiver thread may
        /// run forever.
        /// </summary>
        /// <param name="t">The exception that is causing us to close.</param>
        /// <param name="hard">True if we want to force the close.</param>
        private void Close(Exception t, bool hard)
        {
            if (this.cm != null)
            {
                this.cm.CloseAllChannels();
            }

            if (this.tm != null)
            {
                this.tm.Close(t, hard == false);
                this.tm = null;
            }

            this.am = null;
            this.cm = null;
            this.authenticated = false;
        }

        /// <summary>
        /// Class that is used to track the state of a Connection.
        /// It must be a class because we 'lock' on it.
        /// </summary>
        private class TimeoutState
        {
            private bool isCancelled = false;
            private bool timeoutSocketClosed = false;

            /// <summary>
            /// Initializes a new instance of the TimeoutState class.
            /// </summary>
            public TimeoutState()
            {
            }

            /// <summary>
            /// Gets or sets a value indicating whether the connection is cancelled.
            /// </summary>
            public bool IsCancelled
            {
                get { return this.isCancelled; }
                set { this.isCancelled = value; }
            }

            /// <summary>
            /// Gets or sets a value indicating whether the underlying socket timed out.
            /// </summary>
            public bool TimeoutSocketClosed
            {
                get { return this.timeoutSocketClosed; }
                set { this.timeoutSocketClosed = value; }
            }
        }

        /// <summary>
        /// Nested class that helps implements a timeout along with the TimeoutService class (in util).
        /// </summary>
        private class TimeoutHandler : IThreadRunnable
        {
            private TimeoutState state;
            private Connection conn;

            /// <summary>
            /// Initializes a new instance of the TimeoutHandler class.
            /// </summary>
            /// <param name="state">See the preceeding class.</param>
            /// <param name="conn">The underlying connection.</param>
            public TimeoutHandler(TimeoutState state, Connection conn)
            {
                this.state = state;
                this.conn = conn;
            }

            /// <summary>
            /// Run method returns when cancelled or the socket is closed.
            /// </summary>
            public void Run()
            {
                lock (this.state)
                {
                    if (this.state.IsCancelled)
                    {
                        return;
                    }

                    this.state.TimeoutSocketClosed = true;
                    if (this.conn.tm != null)
                    {
                        this.conn.tm.Close(new IOException("The connect timeout expired"), false);
                    }
                }
            }
        }
    }
}
