﻿// <copyright file="BlockCipherFactory.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> BlockCipherFactory
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Collections;
using System.Text;

namespace Eunomia.SSH2.Crypto.Cipher
{
    /// <summary>
    /// Static class that contains all the ciphers that we support, as well as
    /// routines to generate the ciphers.
    /// </summary>
    public static class BlockCipherFactory
    {
        private static ArrayList ciphers = InitializeCiphers();

        /// <summary>
        /// Gets a string array with all the supported cipher names.
        /// </summary>
        /// <returns>A string array with all the supported cipher names.</returns>
        public static string[] GetDefaultCipherList()
        {
            string[] list = new string[ciphers.Count];
            for (int i = 0; i < ciphers.Count; i++)
            {
                CipherEntry ce = (CipherEntry)ciphers[i];
                list[i] = new StringBuilder(ce.Type).ToString();
            }
            
            return list;
        }

        /// <summary>
        /// Checks a string array of cipher names to make sure that we support them.
        /// </summary>
        /// <param name="cipherCandidates">A String array of possible cipher names.</param>
        /// <exception cref="ArgumentException">If the cipher name is not supported.</exception>
        public static void CheckCipherList(string[] cipherCandidates)
        {
            for (int i = 0; i < cipherCandidates.Length; i++)
            {
                GetEntry(cipherCandidates[i]);
            }
        }

        /// <summary>
        /// Creates and returns a cipher based on the information passed to it.
        /// </summary>
        /// <param name="type">The cipher name.</param>
        /// <param name="encrypt">True for encryption, false for decryption.</param>
        /// <param name="key">The key to use.</param>
        /// <param name="iv">The IV to use.</param>
        /// <returns>The block cipher.</returns>
        /// <exception cref="ArgumentException">If the cipher is not supported.</exception>
        public static IBlockCipher CreateCipher(string type, bool encrypt, byte[] key, byte[] iv)
        {
            try
            {
                CipherEntry ce = GetEntry(type);
                Type cc = Type.GetType(ce.CipherClass);
                IBlockCipher bc = (IBlockCipher)Activator.CreateInstance(cc);

                if (type.EndsWith("-cbc", StringComparison.Ordinal))
                {
                    bc.Init(encrypt, key);
                    return new CBCMode(bc, iv, encrypt);
                }
                else if (type.EndsWith("-ctr", StringComparison.Ordinal))
                {
                    bc.Init(true, key);
                    return new CTRMode(bc, iv);
                }
                
                throw new ArgumentException("Cannot instantiate " + type);
            }
            catch (Exception)
            {
                throw new ArgumentException("Cannot instantiate " + type);
            }
        }

        /// <summary>
        /// Gets the block size for the supplied cipher type.
        /// </summary>
        /// <param name="type">The name of the cipher.</param>
        /// <returns>The block size of the cipher.</returns>
        /// <exception cref="ArgumentException">If the cipher name is not supported.</exception>
        public static int GetBlockSize(string type)
        {
            CipherEntry ce = GetEntry(type);
            return ce.Blocksize;
        }

        /// <summary>
        /// Gets the key size for the supplied cipher type.
        /// </summary>
        /// <param name="type">The name of the cipher.</param>
        /// <returns>The key size of the cipher.</returns>
        /// <exception cref="ArgumentException">If the cipher name is not supported.</exception>
        public static int GetKeySize(string type)
        {
            CipherEntry ce = GetEntry(type);
            return ce.Keysize;
        }

        /// <summary>
        /// Checks through the supported ciphers and returns the CipherEntry if it exists.
        /// If the desired cipher is not in the list an ArgumentException is thrown.
        /// </summary>
        /// <param name="type">The name of the cipher.</param>
        /// <returns>The correct CipherEntry, if found.</returns>
        /// <exception cref="ArgumentException">If the cipher name is not supported.</exception>
        private static CipherEntry GetEntry(string type)
        {
            for (int i = 0; i < ciphers.Count; i++)
            {
                CipherEntry ce = (CipherEntry)ciphers[i];
                if (ce.Type.Equals(type))
                {
                    return ce;
                }
            }

            throw new ArgumentException("Unknown algorithm " + type);
        }

        /// <summary>
        /// Initializes the ArrayList of ciphers.
        /// </summary>
        /// <returns>An ArrayList with the supported ciphers in it.</returns>
        private static ArrayList InitializeCiphers()
        {
            ciphers = new ArrayList(10);
            {
                // Higher Priority First
                ciphers.Add(new CipherEntry("aes256-ctr", 16, 32, "Eunomia.SSH2.Crypto.Cipher.AES"));
                ciphers.Add(new CipherEntry("aes192-ctr", 16, 24, "Eunomia.SSH2.Crypto.Cipher.AES"));
                ciphers.Add(new CipherEntry("aes128-ctr", 16, 16, "Eunomia.SSH2.Crypto.Cipher.AES"));
                ciphers.Add(new CipherEntry("blowfish-ctr", 8, 16, "Eunomia.SSH2.Crypto.Cipher.BlowFish"));

                ciphers.Add(new CipherEntry("aes256-cbc", 16, 32, "Eunomia.SSH2.Crypto.Cipher.AES"));
                ciphers.Add(new CipherEntry("aes192-cbc", 16, 24, "Eunomia.SSH2.Crypto.Cipher.AES"));
                ciphers.Add(new CipherEntry("aes128-cbc", 16, 16, "Eunomia.SSH2.Crypto.Cipher.AES"));
                ciphers.Add(new CipherEntry("blowfish-cbc", 8, 16, "Eunomia.SSH2.Crypto.Cipher.BlowFish"));

                ciphers.Add(new CipherEntry("3des-ctr", 8, 24, "Eunomia.SSH2.Crypto.Cipher.DESede"));
                ciphers.Add(new CipherEntry("3des-cbc", 8, 24, "Eunomia.SSH2.Crypto.Cipher.DESede"));
            }

            return ciphers;
        }

        /// <summary>
        /// A class that holds information about a single supported cipher.
        /// </summary>
        internal class CipherEntry
        {
            private string type;
            private int blocksize;
            private int keysize;
            private string cipherClass;
            
            /// <summary>
            /// Initializes a new instance of the CipherEntry class.
            /// </summary>
            /// <param name="type">The cipher type.</param>
            /// <param name="blocksize">The block size.</param>
            /// <param name="keysize">The key size.</param>
            /// <param name="cipherClass">The actual cipher class.</param>
            public CipherEntry(string type, int blocksize, int keysize, string cipherClass)
            {
                this.type = type;
                this.blocksize = blocksize;
                this.keysize = keysize;
                this.cipherClass = cipherClass;
            }
            
            /// <summary>
            /// Gets the cipher type.
            /// </summary>
            public string Type
            {
                get { return this.type; }
            }
            
            /// <summary>
            /// Gets the blocksize.
            /// </summary>
            public int Blocksize
            {
                get { return this.blocksize; }
            }
            
            /// <summary>
            /// Gets the key size.
            /// </summary>
            public int Keysize
            {
                get { return this.keysize; }
            }
            
            /// <summary>
            /// Gets the underlying cipher class.
            /// </summary>
            public string CipherClass
            {
                get { return this.cipherClass; }
            }
        }
    }
}
