// <copyright file="CipherInputStream.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> CipherInputStream
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Crypto.Cipher
{
    /// <summary>
    /// A class that implements an encrypted Stream (input only).
    /// </summary>
    public class CipherInputStream : Stream
    {
        private const int BUFF_SIZE = 2048;

        private IBlockCipher currentCipher;
        private Stream inStream;
        private byte[] buffer;
        private byte[] enc;
        private int blockSize;
        private int pos;

        private byte[] input_buffer;
        private int input_buffer_pos = 0;
        private int input_buffer_size = 0;

        /// <summary>
        /// Initializes a new instance of the CipherInputStream class.
        /// </summary>
        /// <param name="blockCipher">The block cipher to use for encryption.</param>
        /// <param name="inStream">The underlying Stream to use.</param>
        public CipherInputStream(IBlockCipher blockCipher, Stream inStream)
        {
            this.input_buffer = new byte[BUFF_SIZE];
            this.inStream = inStream;
            this.ChangeCipher(blockCipher);
        }

        /// <summary>
        /// Gets a bool indicating that we can never write to this stream.
        /// </summary>
        public override bool CanWrite
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets a bool indicating that we can read from this stream.
        /// </summary>
        public override bool CanRead
        {
            get { return true; }
        }
        
        /// <summary>
        /// Gets a bool indicating that we cannot seek on this stream.
        /// </summary>
        public override bool CanSeek
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets a bool indicating that we cannot timeout on this stream.
        /// </summary>
        public override bool CanTimeout
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets the length of the stream (just returns the buffer length).
        /// </summary>
        public override long Length
        {
            get
            {
                return BUFF_SIZE;
            }
        }
        
        /// <summary>
        /// Gets or sets...nothing (not implemented).
        /// </summary>
        public override long Position
        {
            get { return 0L; }
            set { }
        }

        /// <summary>
        /// Seeks on the stream (not implemented).
        /// </summary>
        /// <param name="offset">The offset we want to seek to.</param>
        /// <param name="origin">The origin that we will seek from.</param>
        /// <returns>The new position in the stream.</returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// Flush the stream (does nothing).
        /// </summary>
        public override void Flush()
        {
        }
        
        /// <summary>
        /// Set the length of the stream (not implemented).
        /// </summary>
        /// <param name="value">The length we want.</param>
        public override void SetLength(long value)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// Write to the stream (not implemented).
        /// </summary>
        /// <param name="buffer">The buffer to write.</param>
        /// <param name="offset">The offset into the buffer to write from.</param>
        /// <param name="count">The number of bytes to write.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// This allows us to change the underlying cipher.
        /// </summary>
        /// <param name="bc">The block cipher that we now want to use.</param>
        public void ChangeCipher(IBlockCipher bc)
        {
            this.currentCipher = bc;
            this.blockSize = bc.BlockSize;
            this.buffer = new byte[this.blockSize];
            this.enc = new byte[this.blockSize];
            this.pos = this.blockSize;
        }

        /// <summary>
        /// Read from the stream into an existing byte array.
        /// </summary>
        /// <param name="dst">The byte array to read into.</param>
        /// <returns>The number of bytes that were read.</returns>
        public int Read(byte[] dst)
        {
            return this.Read(dst, 0, dst.Length);
        }

        /// <summary>
        /// Read from the stream into an existing byte array.
        /// </summary>
        /// <param name="buffer">The byte array to read into.</param>
        /// <param name="offset">The offset in the byte array to begin reading into.</param>
        /// <param name="count">The number of bytes to get from the stream.</param>
        /// <returns>The number of bytes that were read.</returns>
        public override int Read(byte[] buffer, int offset, int count)
        {
            int cnt = 0;

            while (count > 0)
            {
                if (this.pos >= this.blockSize)
                {
                    this.GetBlock();
                }

                int avail = this.blockSize - this.pos;
                int copy = Math.Min(avail, count);
                Array.Copy(this.buffer, this.pos, buffer, offset, copy);
                this.pos += copy;
                offset += copy;
                count -= copy;
                cnt += copy;
            }

            return cnt;
        }

        /// <summary>
        /// Reads a single byte and returns it as an integer.
        /// </summary>
        /// <returns>The value of the single byte that was read.</returns>
        public int Read()
        {
            if (this.pos >= this.blockSize)
            {
                this.GetBlock();
            }

            return this.buffer[this.pos++] & 0xff;
        }

        /// <summary>
        /// Read bytes from the buffer without encryption.
        /// </summary>
        /// <param name="buffer">The buffer to read to.</param>
        /// <param name="off">The offset in the buffer where we begin placing bytes.</param>
        /// <param name="len">The number of bytes that we want to read.</param>
        /// <returns>The number of bytes that are read.</returns>
        public int ReadPlain(byte[] buffer, int off, int len)
        {
            if (this.pos != this.blockSize)
            {
                throw new IOException("Cannot read plain since crypto buffer is not aligned.");
            }
            
            int n = 0;
            while (n < len)
            {
                int cnt = this.InternalRead(buffer, off + n, len - n);
                if (cnt == 0)
                {
                    throw new IOException("Cannot fill buffer, EOF reached.");
                }

                n += cnt;
            }
            
            return n;
        }

        /// <summary>
        /// Gets a block of data (bytes) with encryption.
        /// </summary>
        private void GetBlock()
        {
            int n = 0;
            while (n < this.blockSize)
            {
                int len = this.InternalRead(this.enc, n, this.blockSize - n);
                if (len == 0)
                {
                    throw new IOException("Cannot read full block, EOF reached.");
                }

                n += len;
            }

            try
            {
                this.currentCipher.TransformBlock(this.enc, 0, this.buffer, 0);
            }
            catch (Exception)
            {
                throw new IOException("Error while decrypting block.");
            }

            this.pos = 0;
        }

        /// <summary>
        /// Read from our internal buffer.
        /// </summary>
        /// <param name="b">The byte array to read into.</param>
        /// <param name="off">The offset where we begin placing bytes.</param>
        /// <param name="len">The number of bytes to read.</param>
        /// <returns>The number of bytes actually read.</returns>
        private int InternalRead(byte[] b, int off, int len)
        {
            if (this.input_buffer_size < 0)
            {
                return 0;
            }

            if (this.input_buffer_pos >= this.input_buffer_size)
            {
                if (this.FillBuffer() <= 0)
                {
                    return 0;
                }
            }

            int avail = this.input_buffer_size - this.input_buffer_pos;
            int thiscopy = (len > avail) ? avail : len;

            Array.Copy(this.input_buffer, this.input_buffer_pos, b, off, thiscopy);
            this.input_buffer_pos += thiscopy;

            return thiscopy;
        }

        /// <summary>
        /// Fill our internal buffer from the underlying stream.
        /// </summary>
        /// <returns>The number of bytes read from the underlying stream.</returns>
        private int FillBuffer()
        {
            this.input_buffer_pos = 0;
            this.input_buffer_size = this.inStream.Read(this.input_buffer, 0, BUFF_SIZE);
            return this.input_buffer_size;
        }
    }
}
