// <copyright file="DhExchange.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> DhExchange
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;

using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Log;
using Eunomia.SSH2.Util;

using Org.BouncyCastle.Math;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.DH
{
    /// <summary>
    /// DHExchange class is all about Diffie-Hellman related data and methods.
    /// </summary>
    public class DHExchange
    {
        private const string P1String = "17976931348623159077083915679378745319786029604875"
            + "60117064444236841971802161585193689478337958649255415021805654859805036464"
            + "40548199239100050792877003355816639229553136239076508735759914822574862575"
            + "00742530207744771258955095793777842444242661733472762929938766870920560605"
            + "0270810842907692932019128194467627007";

        private const string P14String = "FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129"
            + "024E088A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B302B0"
            + "A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9A637ED6B0BFF5CB"
            + "6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3DC2007CB8A"
            + "163BF0598DA48361C55D39A69163FA8FD24CF5F83655D23DCA3AD961C62F356208"
            + "552BB9ED529077096966D670C354E4ABC9804F1746C08CA18217C32905E462E36C"
            + "E3BE39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF69558171"
            + "83995497CEA956AE515D2261898FA051015728E5A8AACAA68FFFFFFFFFFFFFFFF";

        private static Logger log = Logger.GetLogger(typeof(DHExchange));

        // Given by the standard
        private BigInteger p;

        // Client public and private
        private BigInteger e;
        private BigInteger x;

        // Server public
        private BigInteger f;

        // Shared secret
        private BigInteger k;

        private BigInteger p1;
        private BigInteger p14;
        private BigInteger g;

        /// <summary>
        /// Initializes a new instance of the DHExchange class.
        /// </summary>
        public DHExchange()
        {
            this.p1 = new BigInteger(P1String, 16);
            this.p14 = new BigInteger(P14String, 16);
            this.g = new BigInteger("2");
        }

        /// <summary>
        /// Gets the 'public' version of our randomly generated key.
        /// </summary>
        /// <exception cref="InvalidOperationException">If still null.</exception>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "E", Justification = "Consistancy to ganymed project.")]
        public BigInteger E
        {
            get
            {
                if (this.e == null)
                {
                    throw new InvalidOperationException("DhDsaExchange not initialized!");
                }

                return this.e;
            }
        }

        /// <summary>
        /// Gets the shared secret if it exists.
        /// </summary>
        /// <exception cref="InvalidOperationException">If still null.</exception>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "K", Justification = "Consistancy to ganymed project.")]
        public BigInteger K
        {
            get
            {
                if (this.k == null)
                {
                    throw new InvalidOperationException("Shared secret not yet known, need f first!");
                }

                return this.k;
            }
        }
        
        /// <summary>
        /// Gets or sets the server 'public' key and calculates the shared secret (k).
        /// </summary>
        /// <exception cref="InvalidOperationException">If our 'public key is not set yet.</exception>
        /// <exception cref="ArgumentException">If the passed value is negative.</exception>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "F", Justification = "Consistancy to ganymed project.")]
        public BigInteger F
        {
            get
            {
                return this.f;
            }
            
            set
            {
                if (this.e == null)
                {
                    throw new InvalidOperationException("DhDsaExchange not initialized!");
                }

                BigInteger zero = BigInteger.Zero;

                if (zero.CompareTo(value) >= 0 || this.p.CompareTo(value) <= 0)
                {
                    throw new ArgumentException("Invalid f specified!");
                }

                this.f = value;
                this.k = value.ModPow(this.x, this.p);
            }
        }

        /// <summary>
        /// Initializes the DhExchange data.
        /// </summary>
        /// <param name="group">Which DH group to use (1 or 14).</param>
        /// <param name="random">A SecureRandom instance.</param>
        [SuppressMessage("Microsoft.Design", "CA1011:ConsiderPassingBaseTypesAsParameters", Justification = "SecureRandom is required.")]
        public void Init(int group, SecureRandom random)
        {
            this.k = null;

            if (group == 1)
            {
                this.p = this.p1;
            }
            else if (group == 14)
            {
                this.p = this.p14;
            }
            else
            {
                throw new ArgumentException("Unknown DH group " + group);
            }

            this.x = new BigInteger(this.p.BitLength - 1, random);

            this.e = this.g.ModPow(this.x, this.p);
        }

        /// <summary>
        /// Calulates the hash used for verification of the DH data.
        /// </summary>
        /// <param name="clientVersion">The client version in a byte 'string'.</param>
        /// <param name="serverVersion">The server version in a byte 'string'.</param>
        /// <param name="clientKexPayload">The client key exchange payload.</param>
        /// <param name="serverKexPayload">The server key exchange payload.</param>
        /// <param name="hostkey">The host key.</param>
        /// <returns>The calculated hash in a byte array.</returns>
        public byte[] CalculateH(
            byte[] clientVersion,
            byte[] serverVersion,
            byte[] clientKexPayload,
            byte[] serverKexPayload,
            byte[] hostkey)
        {
            HashForSSH2Types hash = new HashForSSH2Types("SHA1");

            if (log.Enabled)
            {
                log.Log(90, "Client: '" + StringEncoder.GetString(clientVersion) + "'");
                log.Log(90, "Server: '" + StringEncoder.GetString(serverVersion) + "'");
            }

            hash.UpdateByteString(clientVersion);
            hash.UpdateByteString(serverVersion);
            hash.UpdateByteString(clientKexPayload);
            hash.UpdateByteString(serverKexPayload);
            hash.UpdateByteString(hostkey);
            hash.UpdateBigInt(this.e);
            hash.UpdateBigInt(this.f);
            hash.UpdateBigInt(this.k);

            return hash.GetDigest();
        }
    }
}
