// <copyright file="HMAC.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> HMAC
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;

namespace Eunomia.SSH2.Crypto.Digest
{
    /// <summary>
    /// HMAC class.
    /// </summary>
    public sealed class HMAC : IDigest
    {
        private IDigest md;
        private byte[] xorIpad;
        private byte[] xorOpad;

        private byte[] tmp;

        private int size;

        /// <summary>
        /// Initializes a new instance of the HMAC class.
        /// </summary>
        /// <param name="md">The underlying digest.</param>
        /// <param name="key">The HMAC key.</param>
        /// <param name="size">The HMAC digest size.</param>
        public HMAC(IDigest md, byte[] key, int size)
        {
            this.md = md;
            this.size = size;

            this.tmp = new byte[md.DigestLength];

            const int BLOCKSIZE = 64;

            this.xorIpad = new byte[BLOCKSIZE];
            this.xorOpad = new byte[BLOCKSIZE];

            if (key.Length > BLOCKSIZE)
            {
                md.Reset();
                md.Update(key);
                md.Digest(this.tmp);
                key = this.tmp;
            }

            Array.Copy(key, 0, this.xorIpad, 0, key.Length);
            Array.Copy(key, 0, this.xorOpad, 0, key.Length);

            for (int i = 0; i < BLOCKSIZE; i++)
            {
                this.xorIpad[i] ^= 0x36;
                this.xorOpad[i] ^= 0x5C;
            }

            md.Update(this.xorIpad);
        }

        /// <summary>
        /// Gets the digest length in bytes.
        /// </summary>
        public int DigestLength
        {
            get { return this.size; }
        }

        /// <summary>
        /// Update a single byte.
        /// </summary>
        /// <param name="value">The byte to update.</param>
        public void Update(byte value)
        {
            this.md.Update(value);
        }

        /// <summary>
        /// Update a byte array.
        /// </summary>
        /// <param name="buffer">The byte array.</param>
        public void Update(byte[] buffer)
        {
            this.md.Update(buffer);
        }

        /// <summary>
        /// Update a byte array with a starting offset and length.
        /// </summary>
        /// <param name="buffer">The  byte array.</param>
        /// <param name="offset">The starting byte in the array.</param>
        /// <param name="length">The number of bytes to process.</param>
        public void Update(byte[] buffer, int offset, int length)
        {
            this.md.Update(buffer, offset, length);
        }

        /// <summary>
        /// Resets the HMAC 'digest'.
        /// </summary>
        public void Reset()
        {
            this.md.Reset();
            this.md.Update(this.xorIpad);
        }

        /// <summary>
        /// Copies the resulting digest into a byte array.
        /// </summary>
        /// <param name="buffer">The byte array to copy the digest to.</param>
        public void Digest(byte[] buffer)
        {
            this.md.Digest(buffer, 0);
        }

        /// <summary>
        /// Copies the resulting digest into a byte array with a starting position.
        /// </summary>
        /// <param name="buffer">The byte array to copy the digest to.</param>
        /// <param name="offset">The starting position of the byte array to copy to.</param>
        public void Digest(byte[] buffer, int offset)
        {
            this.md.Digest(this.tmp);

            this.md.Update(this.xorOpad);
            this.md.Update(this.tmp);

            this.md.Digest(this.tmp);

            Array.Copy(this.tmp, 0, buffer, offset, this.size);

            this.md.Update(this.xorIpad);
        }
    }
}
