// <copyright file="HashForSSH2Types.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> HashForSSH2Types
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;

using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Crypto.Digest
{
    /// <summary>
    /// Convenience class to deal with digest/hash selection.
    /// </summary>
    public class HashForSSH2Types
    {
        private IDigest md;

        /// <summary>
        /// Initializes a new instance of the HashForSSH2Types class.
        /// </summary>
        /// <param name="md">The digest to use.</param>
        public HashForSSH2Types(IDigest md)
        {
            this.md = md;
        }

        /// <summary>
        /// Initializes a new instance of the HashForSSH2Types class.
        /// </summary>
        /// <param name="type">The name of the digest to create an instance of.</param>
        public HashForSSH2Types(string type)
        {
            if (type == "SHA1")
            {
                this.md = new SHA1();
            }
            else if (type == "MD5")
            {
                this.md = new MD5();
            }
            else
            {
                throw new ArgumentException("Unknown algorithm " + type);
            }
        }

        /// <summary>
        /// Gets the underlying digest length in bytes.
        /// </summary>
        public int DigestLength
        {
            get { return this.md.DigestLength; }
        }

        /// <summary>
        /// Updates a single byte.
        /// </summary>
        /// <param name="value">The byte to update.</param>
        public void UpdateByte(byte value)
        {
            byte[] tmp = new byte[1];
            tmp[0] = value;
            this.md.Update(tmp);
        }

        /// <summary>
        /// Update a byte array.
        /// </summary>
        /// <param name="buffer">The byte array to update.</param>
        public void UpdateBytes(byte[] buffer)
        {
            this.md.Update(buffer);
        }

        /// <summary>
        /// Update an integer.
        /// </summary>
        /// <remarks>
        /// Despite the name, the integer is signed!</remarks>
        /// <param name="value">The integer to update.</param>
        public void UpdateUINT32(int value)
        {
            this.md.Update(unchecked((byte)(value >> 24)));
            this.md.Update(unchecked((byte)(value >> 16)));
            this.md.Update(unchecked((byte)(value >> 8)));
            this.md.Update(unchecked((byte)value));
        }

        /// <summary>
        /// Update a 'byte string'.  We update with the Length of the byte array and then the contents of the array.
        /// </summary>
        /// <param name="byteString">The byte array to update.</param>
        public void UpdateByteString(byte[] byteString)
        {
            this.UpdateUINT32(byteString.Length);
            this.UpdateBytes(byteString);
        }

        /// <summary>
        /// Update a BigInteger.
        /// </summary>
        /// <param name="value">The BigInteger to update.</param>
        public void UpdateBigInt(BigInteger value)
        {
            this.UpdateByteString(value.ToByteArray());
        }

        /// <summary>
        /// Reset the underlying digest/hash.
        /// </summary>
        public void Reset()
        {
            this.md.Reset();
        }

        /// <summary>
        /// Get the digest, return it in a byte array.
        /// </summary>
        /// <returns>A byte array with the digest in it.</returns>
        public byte[] GetDigest()
        {
            byte[] buffer = new byte[this.DigestLength];
            this.GetDigest(buffer, 0);
            return buffer;
        }

        /// <summary>
        /// Get the digest, return it in a supplied byte array.
        /// </summary>
        /// <param name="buffer">The byte array that the digest is returned in.</param>
        /// <param name="offset">The offset into the byte array where we begin storing the digest.</param>
        public void GetDigest(byte[] buffer, int offset)
        {
            this.md.Digest(buffer, offset);
        }
    }
}
