﻿// <copyright file="KeyMaterial.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> KeyMaterial
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;
using Eunomia.SSH2.Crypto.Digest;
using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Crypto
{
    /// <summary>
    /// Class that establishes key material for iv/key/mac (both directions).
    /// </summary>
    public class KeyMaterial
    {
        private byte[] initialIVClientToServer;
        private byte[] initialIVServerToClient;
        private byte[] encKeyClientToServer;
        private byte[] encKeyServerToClient;
        private byte[] integrityKeyClientToServer;
        private byte[] integrityKeyServerToClient;
        
        /// <summary>
        /// Gets the initial IV from client to server.
        /// </summary>
        public byte[] InitialIVClientToServer
        {
            get { return this.initialIVClientToServer; }
        }
        
        /// <summary>
        /// Gets the initial IV from server to client.
        /// </summary>
        public byte[] InitialIVServerToClient
        {
            get { return this.initialIVServerToClient; }
        }
        
        /// <summary>
        /// Gets the encryption key from client to server.
        /// </summary>
        public byte[] EncKeyClientToServer
        {
            get { return this.encKeyClientToServer; }
        }
        
        /// <summary>
        /// Gets the encryption ley from server to client.
        /// </summary>
        public byte[] EncKeyServerToClient
        {
            get { return this.encKeyServerToClient; }
        }
        
        /// <summary>
        /// Gets the integrity key from client to server.
        /// </summary>
        public byte[] IntegrityKeyClientToServer
        {
            get { return this.integrityKeyClientToServer; }
        }
        
        /// <summary>
        /// Gets the integrity key from server to client.
        /// </summary>
        public byte[] IntegrityKeyServerToClient
        {
            get { return this.integrityKeyServerToClient; }
        }

        /// <summary>
        /// Static method to create all the above data.
        /// </summary>
        /// <param name="hashType">Hash/digest to use.</param>
        /// <param name="h">H parameter to use.</param>
        /// <param name="k">K parameter to use.</param>
        /// <param name="sessionId">The session id.</param>
        /// <param name="keyLengthCS">The client to server key length.</param>
        /// <param name="blockSizeCS">The client to server block size.</param>
        /// <param name="macLengthCS">The client to server MAC length.</param>
        /// <param name="keyLengthSC">The server to client key length.</param>
        /// <param name="blockSizeSC">The server to client block size.</param>
        /// <param name="macLengthSC">The server to client MAC length.</param>
        /// <returns>A new instance of the KeyMaterial class.</returns>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "h", Justification = "Consistancy to ganymed project.")]
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "k", Justification = "Consistancy to ganymed project.")]
        public static KeyMaterial Create(
            string hashType,
            byte[] h,
            BigInteger k,
            byte[] sessionId,
            int keyLengthCS,
            int blockSizeCS,
            int macLengthCS,
            int keyLengthSC,
            int blockSizeSC,
            int macLengthSC)
        {
            KeyMaterial km = new KeyMaterial();

            HashForSSH2Types sh = new HashForSSH2Types(hashType);

            km.initialIVClientToServer = CalculateKey(sh, k, h, (byte)'A', sessionId, blockSizeCS);

            km.initialIVServerToClient = CalculateKey(sh, k, h, (byte)'B', sessionId, blockSizeSC);

            km.encKeyClientToServer = CalculateKey(sh, k, h, (byte)'C', sessionId, keyLengthCS);

            km.encKeyServerToClient = CalculateKey(sh, k, h, (byte)'D', sessionId, keyLengthSC);

            km.integrityKeyClientToServer = CalculateKey(sh, k, h, (byte)'E', sessionId, macLengthCS);

            km.integrityKeyServerToClient = CalculateKey(sh, k, h, (byte)'F', sessionId, macLengthSC);

            return km;
        }

        /// <summary>
        /// Internal calculation routine for the KeyMaterial elements.
        /// </summary>
        /// <param name="sh">The hash function to use.</param>
        /// <param name="k">The k parameter.</param>
        /// <param name="h">The h parameter.</param>
        /// <param name="type">The calculation type.</param>
        /// <param name="sessionID">The session id.</param>
        /// <param name="keyLength">The key length.</param>
        /// <returns>The calculated key in a byte array.</returns>
        private static byte[] CalculateKey(
            HashForSSH2Types sh,
            BigInteger k,
            byte[] h,
            byte type,
            byte[] sessionID,
            int keyLength)
        {
            byte[] res = new byte[keyLength];

            int dglen = sh.DigestLength;
            int numRounds = (keyLength + dglen - 1) / dglen;

            byte[][] tmp = new byte[numRounds][];

            sh.Reset();
            sh.UpdateBigInt(k);
            sh.UpdateBytes(h);
            sh.UpdateByte(type);
            sh.UpdateBytes(sessionID);

            tmp[0] = sh.GetDigest();

            int off = 0;
            int produced = Math.Min(dglen, keyLength);

            Array.Copy(tmp[0], 0, res, off, produced);

            keyLength -= produced;
            off += produced;

            for (int i = 1; i < numRounds; i++)
            {
                sh.UpdateBigInt(k);
                sh.UpdateBytes(h);

                for (int j = 0; j < i; j++)
                {
                    sh.UpdateBytes(tmp[j]);
                }

                tmp[i] = sh.GetDigest();

                produced = Math.Min(dglen, keyLength);
                Array.Copy(tmp[i], 0, res, off, produced);
                keyLength -= produced;
                off += produced;
            }

            return res;
        }
    }
}
