// <copyright file="SimpleDERReader.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> SimpleDERReader
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;

using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Crypto
{
    /// <summary>
    /// Use to read a X.509 certificate ("Distiguished Encoding Rules").
    /// </summary>
    public class SimpleDERReader
    {
        private byte[] buffer;
        private int pos;
        private int count;

        /// <summary>
        /// Initializes a new instance of the SimpleDERReader class.
        /// </summary>
        /// <param name="buffer">The byte array with the data to read in it.</param>
        public SimpleDERReader(byte[] buffer)
        {
            this.ResetInput(buffer);
        }

        /// <summary>
        /// Initializes a new instance of the SimpleDERReader class.
        /// </summary>
        /// <param name="buffer">The byte array with the data to read in it.</param>
        /// <param name="offset">The offset into the byte array to start at.</param>
        /// <param name="length">The number of bytes to process.</param>
        public SimpleDERReader(byte[] buffer, int offset, int length)
        {
            this.ResetInput(buffer, offset, length);
        }

        /// <summary>
        /// Sets the private data up before we decode.
        /// </summary>
        /// <param name="buffer">The byte array with the data to read in it.</param>
        public void ResetInput(byte[] buffer)
        {
            this.ResetInput(buffer, 0, buffer.Length);
        }

        /// <summary>
        /// Sets the private data up before we decode.
        /// </summary>
        /// <param name="buffer">The byte array with the data to read in it.</param>
        /// <param name="offset">The offset into the byte array to start at.</param>
        /// <param name="length">The number of bytes to process.</param>
        public void ResetInput(byte[] buffer, int offset, int length)
        {
            this.buffer = buffer;
            this.pos = offset;
            this.count = length;
        }

        /// <summary>
        /// Gets the number of bytes to be processed.
        /// </summary>
        /// <returns>The number of bytes to be processed.</returns>
        public int Available()
        {
            // TODO: This should be a property?
            return this.count;
        }

        /// <summary>
        /// Skip over a piece of data.  Reads the type and length and then the data (which is discarded).
        /// </summary>
        /// <returns>The type of the read data as an integer.</returns>
        public int IgnoreNextObject()
        {
            int type = this.ReadByte() & 0xff;

            int len = this.ReadLength();

            if ((len < 0) || len > this.Available())
            {
                throw new IOException("Illegal len in DER object (" + len  + ")");
            }

            this.ReadBytes(len);

            return type;
        }

        /// <summary>
        /// Reads a DER BigInteger.
        /// </summary>
        /// <returns>The BigInteger that was read.</returns>
        public BigInteger ReadInt()
        {
            int type = this.ReadByte() & 0xff;

            if (type != 0x02)
            {
                throw new IOException("Expected DER Integer, but found type " + type);
            }

            int len = this.ReadLength();

            if ((len < 0) || len > this.Available())
            {
                throw new IOException("Illegal len in DER object (" + len  + ")");
            }

            byte[] b = this.ReadBytes(len);

            BigInteger bi = new BigInteger(b);

            return bi;
        }

        /// <summary>
        /// Reads a DER sequence as a byte array.
        /// </summary>
        /// <returns>The byte array with the sequence in it.</returns>
        public byte[] ReadSequenceAsByteArray()
        {
            int type = this.ReadByte() & 0xff;

            if (type != 0x30)
            {
                throw new IOException("Expected DER Sequence, but found type " + type);
            }

            int len = this.ReadLength();

            if ((len < 0) || len > this.Available())
            {
                throw new IOException("Illegal len in DER object (" + len  + ")");
            }

            byte[] b = this.ReadBytes(len);

            return b;
        }

        /// <summary>
        /// Read an Octel string and return as a byte array.
        /// </summary>
        /// <returns>The byte array with the octet string in it.</returns>
        public byte[] ReadOctetString()
        {
            int type = this.ReadByte() & 0xff;

            if (type != 0x04)
            {
                throw new IOException("Expected DER Octetstring, but found type " + type);
            }

            int len = this.ReadLength();

            if ((len < 0) || len > this.Available())
            {
                throw new IOException("Illegal len in DER object (" + len  + ")");
            }

            byte[] b = this.ReadBytes(len);

            return b;
        }

        /// <summary>
        /// Read a single byte from the buffer.
        /// </summary>
        /// <returns>The byte that was read.</returns>
        private byte ReadByte()
        {
            if (this.count <= 0)
            {
                throw new IOException("DER byte array: out of data");
            }

            this.count--;
            return this.buffer[this.pos++];
        }

        /// <summary>
        /// Reads a byte array from the buffer.
        /// </summary>
        /// <param name="len">The number of bytes to read.</param>
        /// <returns>The bytes that were retrieved.</returns>
        private byte[] ReadBytes(int len)
        {
            if (len > this.count)
            {
                throw new IOException("DER byte array: out of data");
            }

            byte[] b = new byte[len];

            Array.Copy(this.buffer, this.pos, b, 0, len);

            this.pos += len;
            this.count -= len;

            return b;
        }

        /// <summary>
        /// Read the length of the next piece of data.
        /// </summary>
        /// <returns>An integer with the length in it.</returns>
        private int ReadLength()
        {
            int len = this.ReadByte() & 0xff;

            if ((len & 0x80) == 0)
            {
                return len;
            }

            int remain = len & 0x7F;

            if (remain == 0)
            {
                return -1;
            }

            len = 0;

            while (remain > 0)
            {
                len = len << 8;
                len = len | (this.ReadByte() & 0xff);
                remain--;
            }

            return len;
        }
    }
}
