// <copyright file="DHGexParameters.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> DHGexParameters
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A <c>DHGexParameters</c> object can be used to specify parameters for
    /// the diffie-hellman group exchange.
    /// <para>
    /// Depending on which constructor is used, either the use of a
    /// <c>SSHMessageKEX_DH_GEX_REQUEST</c> or <c>SSHMessageKEX_DH_GEX_REQUEST_OLD</c>
    /// can be forced.
    /// </para>
    /// </summary>
    /// <seealso cref="Connection.SetDHGexParameters"></seealso>
    public class DHGexParameters
    {
        private const int MIN_ALLOWED = 1024;
        private const int MAX_ALLOWED = 8192;

        private int minGroupLen;
        private int prefGroupLen;
        private int maxGroupLen;

        /// <summary>
        /// Initializes a new instance of the DHGexParameters class.
        /// Same as calling DHGexParameters(1024, 1024, 4096)}.
        /// This is also the default used by the Connection class.
        /// </summary>
        public DHGexParameters() : this(1024, 1024, 4096)
        {
        }

        /// <summary>
        /// Initializes a new instance of the DHGexParameters class.
        /// This constructor can be used to force the sending of a
        /// <c>SSHMessageKEX_DH_GEX_REQUEST_OLD</c> request.
        /// Internally, the minimum and maximum group lengths will
        /// be set to zero.
        /// </summary>
        /// <param name="preferredGroupLen">
        /// Has to be >= 1024 and &lt;= 8192.
        /// </param>
        public DHGexParameters(int preferredGroupLen)
        {
            if ((preferredGroupLen < MIN_ALLOWED) || (preferredGroupLen > MAX_ALLOWED))
            {
                throw new InvalidOperationException("prefGroupLen out of range!");
            }

            this.prefGroupLen = preferredGroupLen;
            this.minGroupLen = 0;
            this.maxGroupLen = 0;
        }

        /// <summary>
        /// Initializes a new instance of the DHGexParameters class.
        /// This constructor can be used to force the sending of a
        /// <c>SSHMessageKEX_DH_GEX_REQUEST</c> request.
        /// <para>
        /// Note: older OpenSSH servers don't understand this request, in which
        /// case you should use the {@link #DHGexParameters(int)} constructor.
        /// </para>
        /// <para>
        /// All values have to be >= 1024 and &lt;= 8192. Furthermore,
        /// minGroupLen &lt;= prefGroupLen &lt;= maxGroupLen.
        /// </para>
        /// </summary>
        /// <param name="minGroupLen">The minimum group length.</param>
        /// <param name="preferredGroupLen">The preferred group length.</param>
        /// <param name="maxGroupLen">The maximum group length.</param>
        public DHGexParameters(int minGroupLen, int preferredGroupLen, int maxGroupLen)
        {
            if ((minGroupLen < MIN_ALLOWED) || (minGroupLen > MAX_ALLOWED))
            {
                throw new ArgumentException("minGroupLen out of range!");
            }

            if ((preferredGroupLen < MIN_ALLOWED) || (preferredGroupLen > MAX_ALLOWED))
            {
                throw new ArgumentException("prefGroupLen out of range!");
            }

            if ((maxGroupLen < MIN_ALLOWED) || (maxGroupLen > MAX_ALLOWED))
            {
                throw new ArgumentException("maxGroupLen out of range!");
            }

            if ((preferredGroupLen < minGroupLen) || (preferredGroupLen > maxGroupLen))
            {
                throw new ArgumentException("prefGroupLen is incompatible with min and max!");
            }

            if (maxGroupLen < minGroupLen)
            {
                throw new ArgumentException("maxGroupLen must not be smaller than minGroupLen!");
            }

            this.minGroupLen = minGroupLen;
            this.prefGroupLen = preferredGroupLen;
            this.maxGroupLen = maxGroupLen;
        }

        /// <summary>
        /// Gets the maximum group length.  May be zero if
        /// SSHMessageKEX_DH_GEX_REQUEST_OLD should be requested.
        /// </summary>
        public int MaxGroupLength
        {
            get { return this.maxGroupLen; }
        }

        /// <summary>
        /// Gets the minimum group length.  May be zero if
        /// SSHMessageKEX_DH_GEX_REQUEST_OLD should be requested.
        /// </summary>
        public int MinGroupLength
        {
            get { return this.minGroupLen; }
        }

        /// <summary>
        /// Gets the preferred group length.  May be zero if
        /// SSHMessageKEX_DH_GEX_REQUEST_OLD should be requested.
        /// </summary>
        public int PreferredGroupLength
        {
            get { return this.prefGroupLen; }
        }
    }
}
