// <copyright file="JavaThread.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> JavaThread
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Threading;

namespace Eunomia.SSH2
{
    /// <summary>
    /// This interface is inherited by JavaThread and TimeoutHandler (in Connection.cs).
    /// </summary>
    public interface IThreadRunnable
    {
        /// <summary>
        /// The one and only required method - Run.
        /// </summary>
        void Run();
    }

    /// <summary>
    /// This class saved a huge amount of refactoring by providing most of the
    /// functionality of the Java Thread class.
    /// </summary>
    public class JavaThread : IThreadRunnable
    {
        private Thread threadField;

        /// <summary>
        /// Initializes a new instance of the JavaThread class.
        /// </summary>
        public JavaThread()
        {
            this.threadField = new Thread(new ThreadStart(this.Run));
        }

        /// <summary>
        ///  Initializes a new instance of the JavaThread class.
        /// </summary>
        /// <param name="start">A ThreadStart instance.</param>
        public JavaThread(ThreadStart start)
        {
            this.threadField = new Thread(start);
        }

        /// <summary>
        /// Gets or sets the underlying .Net Thread.
        /// </summary>
        public Thread Instance
        {
            get { return this.threadField; }
            set { this.threadField = value; }
        }

        /// <summary>
        /// Gets or sets the Name of the thread.
        /// </summary>
        public string Name
        {
            get
            {
                return this.threadField.Name;
            }
            
            set
            {
                if (this.threadField.Name == null)
                {
                    this.threadField.Name = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the priority of the thread.
        /// </summary>
        public ThreadPriority Priority
        {
            get { return this.threadField.Priority; }
            set { this.threadField.Priority = value; }
        }

        /// <summary>
        /// Gets a value indicating whether the thread is 'alive' or not.
        /// </summary>
        public bool IsAlive
        {
            get { return this.threadField.IsAlive; }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the thread is a bcakground thread or not.
        /// </summary>
        public bool IsBackground
        {
            get { return this.threadField.IsBackground; }
            set { this.threadField.IsBackground = value; }
        }

        /// <summary>
        /// Static method that retriees the current thread's JavaThread instance.
        /// </summary>
        /// <returns>The JavaThread instance.</returns>
        public static JavaThread Current()
        {
            JavaThread currentThread = new JavaThread();
            currentThread.Instance = Thread.CurrentThread;
            return currentThread;
        }
        
        /// <summary>
        /// When overidden, this is the method that will be run in a new thread. 
        /// </summary>
        public virtual void Run()
        {
            throw new NotImplementedException("You must override the Run method when you inherit from this class");
        }

        /// <summary>
        /// Starts the thread.
        /// </summary>
        public void Start()
        {
            this.threadField.Start();
        }

        /// <summary>
        /// Interrupts the thread.  See .Net documentation on how this works.
        /// </summary>
        public void Interrupt()
        {
            this.threadField.Interrupt();
        }

        /// <summary>
        /// Blocks the current thread until the referenced thread terminates.
        /// </summary>
        public void Join()
        {
            this.threadField.Join();
        }

        /// <summary>
        /// Blocks the calling thread until a thread terminates or the specified time elapses.
        /// </summary>
        /// <param name="milliseconds">The timeout in milliseconds.</param>
        public void Join(long milliseconds)
        {
            lock (this)
            {
                this.threadField.Join(new System.TimeSpan(milliseconds * 10000));
            }
        }

        /// <summary>
        /// Blocks the calling thread until a thread terminates or the specified time elapses.
        /// </summary>
        /// <param name="milliseconds">The timeout in milliseconds.</param>
        /// <param name="nanoseconds">Additional timeout added in nano seconds.</param>
        public void Join(long milliseconds, int nanoseconds)
        {
            lock (this)
            {
                this.threadField.Join(new System.TimeSpan((milliseconds * 10000) + (nanoseconds * 100)));
            }
        }

        /// <summary>
        /// Aborts a thread.  This is more abrupt that Interrupt and is considered
        /// a bit unsafe...but we are using it.
        /// </summary>
        public void Abort()
        {
            this.threadField.Abort();
        }

        /// <summary>
        /// Aborts a thread.  This is more abrupt that Interrupt and is considered
        /// a bit unsafe...but we are using it.
        /// </summary>
        /// <param name="stateInfo">Provides exception information about the thread termination.</param>
        public void Abort(object stateInfo)
        {
            lock (this)
            {
                this.threadField.Abort(stateInfo);
            }
        }

        /// <summary>
        /// Creates a more informative string about the thread.
        /// </summary>
        /// <returns>The string to use.</returns>
        public override string ToString()
        {
            return "Thread[" + this.Name + "," + this.Priority.ToString() + "," + string.Empty + "]";
        }
    }
}
