﻿// <copyright file="PacketChannelWindowAdjust.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketChannelWindowAdjust
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketChannelWindowAdjust packet.
    /// </summary>
    public class PacketChannelWindowAdjust
    {
        private byte[] payload;

        private int recipientChannelId;
        private int windowChange;

        /// <summary>
        /// Initializes a new instance of the PacketChannelWindowAdjust class.
        /// </summary>
        /// <param name="recipientChannelId">The receipient channel id.</param>
        /// <param name="windowChange">The new size for the window.</param>
        public PacketChannelWindowAdjust(int recipientChannelId, int windowChange)
        {
            this.recipientChannelId = recipientChannelId;
            this.windowChange = windowChange;
        }

        /// <summary>
        /// Initializes a new instance of the PacketChannelWindowAdjust class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketChannelWindowAdjust(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageChannelWindowAdjust)
            {
                throw new IOException(
                    "This is not a SSHMessageCHANNEL_WINDOW_ADJUST! ("
                    + packet_type + ")");
            }
            
            this.recipientChannelId = tr.ReadUINT32();
            this.windowChange = tr.ReadUINT32();
            
            if (tr.Remain != 0)
            {
                throw new IOException("Padding in SSHMessageCHANNEL_WINDOW_ADJUST packet!");
            }
        }
        
        /// <summary>Gets or sets the recipient channel id.</summary>
        public int RecipientChannelId
        {
            get { return this.recipientChannelId; }
            set { this.recipientChannelId = value; }
        }
        
        /// <summary>Gets or sets the window change value.</summary>
        public int WindowChange
        {
            get { return this.windowChange; }
            set { this.windowChange = value; }
        }

        /// <summary>
        /// Gets the constructed payload for the packet.
        /// </summary>
        public byte[] Payload
        {
            get
            {
                if (this.payload == null)
                {
                    TypesWriter tw = new TypesWriter();
                    tw.WriteByte(Packet.SSHMessageChannelWindowAdjust);
                    tw.WriteUINT32((uint)this.recipientChannelId);
                    tw.WriteUINT32((uint)this.windowChange);
                    this.payload = tw.GetBytes();
                }
                
                return this.payload;
            }
        }
    }
}
