﻿// <copyright file="PacketDisconnect.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketDisconnect
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketDisconnect packet.
    /// </summary>
    public class PacketDisconnect
    {
        private byte[] payload;

        private int reason;
        private string desc;
        private string lang;
        
        /// <summary>
        /// Initializes a new instance of the PacketDisconnect class.
        /// </summary>
        /// <param name="reason">The reason for the disconnect.</param>
        /// <param name="description">The description for the disconnect.</param>
        /// <param name="lang">The lang parameter.</param>
        public PacketDisconnect(int reason, string description, string lang)
        {
            this.reason = reason;
            this.desc = description;
            this.lang = lang;
        }

        /// <summary>
        /// Initializes a new instance of the PacketDisconnect class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketDisconnect(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageDisconnect)
            {
                throw new IOException("This is not a Disconnect Packet! ("
                                      + packet_type + ")");
            }
            
            this.reason = tr.ReadUINT32();
            this.desc = tr.ReadString();
            this.lang = tr.ReadString();
        }
        
        /// <summary>Gets or sets the reason.</summary>
        public int Reason
        {
            get { return this.reason; }
            set { this.reason = value; }
        }
        
        /// <summary>Gets or sets the desc.</summary>
        public string Description
        {
            get { return this.desc; }
            set { this.desc = value; }
        }
        
        /// <summary>Gets or sets the language.</summary>
        public string Lang
        {
            get { return this.lang; }
            set { this.lang = value; }
        }
        
        /// <summary>
        /// Gets the constructed payload for the packet.
        /// </summary>
        public byte[] Payload
        {
            get
            {
                if (this.payload == null)
                {
                    TypesWriter tw = new TypesWriter();
                    tw.WriteByte(Packet.SSHMessageDisconnect);
                    tw.WriteUINT32((uint)this.reason);
                    tw.WriteString(this.desc);
                    tw.WriteString(this.lang);
                    this.payload = tw.GetBytes();
                }
                
                return this.payload;
            }
        }
    }
}
