// <copyright file="PacketKexDHReply.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketKexDHReply
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;
using System.IO;

using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketKexDhReply packet.
    /// </summary>
    public class PacketKexDHReply
    {
        private byte[] payload;

        private byte[] hostKey;
        private BigInteger f;
        private byte[] signature;

        /// <summary>
        /// Initializes a new instance of the PacketKexDHReply class.
        /// This is step 2 in an older style Diffie-Hellman exchange described in RFC 4253.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketKexDHReply(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageKEXDHReply)
            { 
                throw new IOException("This is not a SSHMessageKEXDH_REPLY! ("
                                      + packet_type + ")");
            }

            this.hostKey = tr.ReadByteString();
            this.f = tr.ReadMPINT();
            this.signature = tr.ReadByteString();

            if (tr.Remain != 0)
            {
                throw new IOException("PADDING IN SSHMessageKEXDH_REPLY!");
            }
        }

        /// <summary>
        /// Gets the server's 'public' key as a BigInteger.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "F", Justification = "Consistancy to ganymed project.")]
        public BigInteger F
        {
            get { return this.f; }
        }

        /// <summary>
        /// Gets the server's public host key.
        /// </summary>
        public byte[] Hostkey
        {
            get { return this.hostKey; }
        }

        /// <summary>
        /// Gets the signature (see RFC4419 for explanation).  Used to authenticate the key exchange.
        /// </summary>
        public byte[] Signature
        {
            get { return this.signature; }
        }
    }
}
