// <copyright file="PacketKexDhGexGroup.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketKexDhGexGroup
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;
using System.IO;

using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketKexDHGexGroup packet.
    /// This is step 2 in the newer style Diffie-Hellman exchange as described in RFC 4419.
    /// </summary>
    public class PacketKexDHGexGroup
    {
        private byte[] payload;

        private BigInteger p;
        private BigInteger g;

        /// <summary>
        /// Initializes a new instance of the PacketKexDHGexGroup class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketKexDHGexGroup(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageKEXDHGEXGroup)
            {
                throw new ArgumentException(
                    "This is not a SSHMessageKEX_DH_GEX_GROUP! (" + packet_type
                    + ")");
            }

            this.p = tr.ReadMPINT();
            this.g = tr.ReadMPINT();

            if (tr.Remain != 0)
            {
                throw new IOException("PADDING IN SSHMessageKEX_DH_GEX_GROUP!");
            }
        }

        /// <summary>
        /// Gets the 'generator' used in the Diffie-Hellman key exchange.  
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "G", Justification = "Consistancy to ganymed project.")]
        public BigInteger G
        {
            get { return this.g; }
        }

        /// <summary>
        /// Gets the large safe prime number used in the Diffie-Hellman key exchange.
        /// </summary>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "P", Justification = "Consistancy to ganymed project.")]
        public BigInteger P
        {
            get { return this.p; }
        }
    }
}
