// <copyright file="PacketOpenSessionChannel.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketOpenSessionChannel
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketOpenSessionChannel packet.
    /// </summary>
    public class PacketOpenSessionChannel
    {
        private byte[] payload;

        private int channelID;
        private int initialWindowSize;
        private int maxPacketSize;

        /// <summary>
        /// Initializes a new instance of the PacketOpenSessionChannel class.
        /// </summary>
        /// <param name="channelId">The channel id.</param>
        /// <param name="initialWindowSize">The initial window size.</param>
        /// <param name="maxPacketSize">The maximum packet size.</param>
        public PacketOpenSessionChannel(
            int channelId,
            int initialWindowSize,
            int maxPacketSize)
        {
            this.channelID = channelId;
            this.initialWindowSize = initialWindowSize;
            this.maxPacketSize = maxPacketSize;
        }

        /// <summary>
        /// Initializes a new instance of the PacketOpenSessionChannel class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketOpenSessionChannel(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageChannelOpen)
            {
                throw new IOException("This is not a SSHMessageCHANNEL_OPEN! ("
                                      + packet_type + ")");
            }
            
            this.channelID = tr.ReadUINT32();
            this.initialWindowSize = tr.ReadUINT32();
            this.maxPacketSize = tr.ReadUINT32();

            if (tr.Remain != 0)
            {
                throw new IOException("Padding in SSHMessageCHANNEL_OPEN packet!");
            }
        }

        /// <summary>
        /// Gets the payload of the packet (creates if required).
        /// </summary>
        public byte[] Payload
        {
            get
            {
                if (this.payload == null)
                {
                    TypesWriter tw = new TypesWriter();
                    tw.WriteByte(Packet.SSHMessageChannelOpen);
                    tw.WriteString("session");
                    tw.WriteUINT32((uint)this.channelID);
                    tw.WriteUINT32((uint)this.initialWindowSize);
                    tw.WriteUINT32((uint)this.maxPacketSize);
                    this.payload = tw.GetBytes();
                }

                return this.payload;
            }
        }
    }
}
