// <copyright file="SCPClient.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> SCPClient
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Globalization;
using System.IO;
using System.Text;

using Eunomia.SSH2.Util;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A very basic <c>SCPClient</c> that can be used to copy files from/to
    /// the SSH-2 server. On the server side, the "scp" program must be in the PATH.
    /// <para>
    /// This scp client is thread safe - you can download (and upload) different sets
    /// of files concurrently without any troubles. The <c>SCPClient</c> is
    /// actually mapping every request to a distinct <c>Session</c>.
    /// </para>
    /// </summary>
    public class SCPClient
    {
        private Connection conn;

        /// <summary>
        /// Initializes a new instance of the SCPClient class.
        /// </summary>
        /// <param name="conn">The underlying connection.</param>
        /// <exception cref="ArgumentException">If the connection is null.</exception>
        public SCPClient(Connection conn)
        {
            if (conn == null)
            {
                throw new ArgumentException("Cannot accept null argument!");
            }
            
            this.conn = conn;
        }

        /// <summary>
        /// Copy a local file to a remote directory, uses mode 0600 when creating
        /// the file on the remote side.
        /// </summary>
        /// <param name="localFile">Path and name of local file.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <exception cref="IOException"></exception>
        public void Put(string localFile, string remoteTargetDirectory)
        {
            this.Put(new string[] { localFile }, remoteTargetDirectory, "0600");
        }

        /// <summary>
        /// Copy an array of local files to a remote directory, uses mode 0600 when creating
        /// the file on the remote side.
        /// </summary>
        /// <param name="localFiles">Array of path and names of local files.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Put(string[] localFiles, string remoteTargetDirectory)
        {
            this.Put(localFiles, remoteTargetDirectory, "0600");
        }

        /// <summary>
        /// Copy a local file to a remote directory, uses the specified mode when creating
        /// the file on the remote side.
        /// </summary>
        /// <param name="localFile">Path and name of local file.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <param name="mode">A four digit string (e.g., 0644, see "man chmod", "man open").</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Put(string localFile, string remoteTargetDirectory, string mode)
        {
            this.Put(new string[] { localFile }, remoteTargetDirectory, mode);
        }
        
        /// <summary>
        /// Copy a local file to a remote directory, uses the specified mode and remote filename
        /// when creating the file on the remote side.
        /// </summary>
        /// <param name="localFile">Path and name of local file.</param>
        /// <param name="remoteFileName">The name of the file which will be created in the remote target directory.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <param name="mode">A four digit string (e.g., 0644, see "man chmod", "man open").</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Put(string localFile, string remoteFileName, string remoteTargetDirectory, string mode)
        {
            this.Put(new string[] { localFile }, new string[] { remoteFileName }, remoteTargetDirectory, mode);
        }

        /// <summary>
        /// Create a remote file and copy the contents of the passed byte array into it.
        /// Uses mode 0600 for creating the remote file.
        /// </summary>
        /// <param name="data">The data to be copied into the remote file.</param>
        /// <param name="remoteFileName">The name of the file which will be created in the remote target directory.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Put(byte[] data, string remoteFileName, string remoteTargetDirectory)
        {
            this.Put(data, remoteFileName, remoteTargetDirectory, "0600");
        }
        
        /// <summary>
        /// Create a remote file and copy the contents of the passed byte array into it.
        /// The method use the specified mode when creating the file on the remote side.
        /// </summary>
        /// <param name="data">The data to be copied into the remote file.</param>
        /// <param name="remoteFileName">The name of the file which will be created in the remote target directory.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <param name="mode">A four digit string (e.g., 0644, see "man chmod", "man open").</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Put(byte[] data, string remoteFileName, string remoteTargetDirectory, string mode)
        {
            Session sess = null;

            if ((remoteFileName == null) || (remoteTargetDirectory == null) || (mode == null))
            {
                throw new ArgumentException("Null argument.");
            }

            if (mode.Length != 4)
            {
                throw new ArgumentException("Invalid mode.");
            }

            for (int i = 0; i < mode.Length; i++)
            {
                if (Char.IsDigit(mode[i]) == false)
                {
                    throw new ArgumentException("Invalid mode.");
                }
            }

            remoteTargetDirectory = remoteTargetDirectory.Trim();
            remoteTargetDirectory = (remoteTargetDirectory.Length > 0) ? remoteTargetDirectory : ".";

            string cmd = "scp -t -d " + remoteTargetDirectory;

            try
            {
                sess = this.conn.OpenSession();
                sess.ExecCommand(cmd);
                SCPClient.SendBytes(sess, data, remoteFileName, mode);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("Error during SCP transfer.");
            }
            finally
            {
                if (sess != null)
                {
                    sess.Close();
                }
            }
        }
        
        /// <summary>
        /// Copy a set of local files to a remote directory, uses the specified mode
        /// when creating the files on the remote side.
        /// </summary>
        /// <param name="localFiles">Paths and names of the local files.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <param name="mode">A four digit string (e.g., 0644, see "man chmod", "man open").</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Put(string[] localFiles, string remoteTargetDirectory, string mode)
        {
            this.Put(localFiles, null, remoteTargetDirectory, mode);
        }

        /// <summary>
        /// Copy a set of local files to a set of remote files, uses the specified mode
        /// when creating the files on the remote side.
        /// </summary>
        /// <param name="localFiles">Paths and names of the local files.</param>
        /// <param name="remoteFiles">Names of the remote files.</param>
        /// <param name="remoteTargetDirectory">Remote target directory. Use an empty string to specify the default directory.</param>
        /// <param name="mode">A four digit string (e.g., 0644, see "man chmod", "man open").</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Put(string[] localFiles, string[] remoteFiles, string remoteTargetDirectory, string mode)
        {
            Session sess = null;

            // remoteFiles may be null, indicating that the local filenames shall be used
            if ((localFiles == null) || (remoteTargetDirectory == null) || (mode == null))
            {
                throw new ArgumentException("Null argument.");
            }

            if (mode.Length != 4)
            {
                throw new ArgumentException("Invalid mode.");
            }

            for (int i = 0; i < mode.Length; i++)
            {
                if (Char.IsDigit(mode[i]) == false)
                {
                    throw new ArgumentException("Invalid mode.");
                }
            }

            if (localFiles.Length == 0)
            {
                return;
            }

            remoteTargetDirectory = remoteTargetDirectory.Trim();
            remoteTargetDirectory = (remoteTargetDirectory.Length > 0) ? remoteTargetDirectory : ".";

            string cmd = "scp -t -d " + remoteTargetDirectory;

            for (int i = 0; i < localFiles.Length; i++)
            {
                if (localFiles[i] == null)
                {
                    throw new ArgumentException("Cannot accept null filename.");
                }
            }

            try
            {
                sess = this.conn.OpenSession();
                sess.ExecCommand(cmd);
                SCPClient.SendFiles(sess, localFiles, remoteFiles, mode);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("Error during SCP transfer.");
            }
            finally
            {
                if (sess != null)
                {
                    sess.Close();
                }
            }
        }

        /// <summary>
        /// Download a file from the remote server to a local directory.
        /// </summary>
        /// <param name="remoteFile">Path and name of the remote file.</param>
        /// <param name="localTargetDirectory">Local directory to put the downloaded file.</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Get(string remoteFile, string localTargetDirectory)
        {
            this.Get(new string[] { remoteFile }, localTargetDirectory);
        }

        /// <summary>
        /// Download a file from the remote server and pipe its contents into an <c>OutputStream</c>.
        /// Please note that, to enable flexible usage of this method, the <c>OutputStream</c> will not
        /// be closed nor flushed.
        /// </summary>
        /// <param name="remoteFile">Path and name of the remote file.</param>
        /// <param name="target">OutputStream where the contents of the file will be sent to.</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Get(string remoteFile, Stream target)
        {
            this.Get(new string[] { remoteFile }, new Stream[] { target });
        }

        /// <summary>
        /// Download a set of files from the remote server to a local directory.
        /// </summary>
        /// <param name="remoteFiles">Paths and names of the remote files.</param>
        /// <param name="localTargetDirectory">Local directory to put the downloaded files.</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="ArgumentException">If any argument is invalid.</exception>
        public void Get(string[] remoteFiles, string localTargetDirectory)
        {
            Session sess = null;

            if ((remoteFiles == null) || (localTargetDirectory == null))
            {
                throw new ArgumentException("Null argument.");
            }

            if (remoteFiles.Length == 0)
            {
                return;
            }

            string cmd = "scp -f";

            for (int i = 0; i < remoteFiles.Length; i++)
            {
                if (remoteFiles[i] == null)
                {
                    throw new ArgumentException("Cannot accept null filename.");
                }

                string tmp = remoteFiles[i].Trim();

                if (tmp.Length == 0)
                {
                    throw new ArgumentException("Cannot accept empty filename.");
                }

                cmd += " " + tmp;
            }

            try
            {
                sess = this.conn.OpenSession();
                sess.ExecCommand(cmd);
                SCPClient.ReceiveFiles(sess, remoteFiles, localTargetDirectory);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("Error during SCP transfer.");
            }
            finally
            {
                if (sess != null)
                {
                    sess.Close();
                }
            }
        }

        /// <summary>
        /// Receives a line/string one byte at a time.
        /// </summary>
        /// <param name="istr">The underlying input stream.</param>
        /// <returns>The string that was read.</returns>
        /// <exception cref="IOException"></exception>
        private static string ReceiveLine(Stream istr)
        {
            StringBuilder sb = new StringBuilder(30);

            while (true)
            {
                // This is a random limit - if your path names are longer, then adjust it
                if (sb.Length > 8192)
                {
                    throw new IOException("Remote scp sent a too long line");
                }

                int c = istr.ReadByte();

                if (c < 0)
                {
                    throw new IOException("Remote scp terminated unexpectedly.");
                }

                if (c == '\n')
                {
                    break;
                }

                sb.Append((char)c);
            }
            
            return sb.ToString();
        }

        /// <summary>
        /// Parses/reads a length/name pair.
        /// </summary>
        /// <param name="line">The line/string to process.</param>
        /// <returns>A LenNamePair.</returns>
        /// <exception cref="IOException"></exception>
        private static LenNamePair ParseCLine(string line)
        {
            /* Minimum line: "xxxx y z" ---> 8 chars */

            long len;

            if (line.Length < 8)
            {
                throw new IOException("Malformed C line sent by remote SCP binary, line too short.");
            }

            if ((line[4] != ' ') || (line[5] == ' '))
            {
                throw new IOException("Malformed C line sent by remote SCP binary.");
            }

            int length_name_sep = line.IndexOf(' ', 5);

            if (length_name_sep == -1)
            {
                throw new IOException("Malformed C line sent by remote SCP binary.");
            }

            string length_substring = line.Substring(5, length_name_sep - 5);
            string name_substring = line.Substring(length_name_sep + 1);

            if ((length_substring.Length <= 0) || (name_substring.Length <= 0))
            {
                throw new IOException("Malformed C line sent by remote SCP binary.");
            }

            if ((6 + length_substring.Length + name_substring.Length) != line.Length)
            {
                throw new IOException("Malformed C line sent by remote SCP binary.");
            }

            try
            {
                len = Int64.Parse(length_substring, CultureInfo.InvariantCulture);
            }
            catch (FormatException)
            {
                throw new IOException("Malformed C line sent by remote SCP binary, cannot parse file length.");
            }

            if (len < 0)
            {
                throw new IOException("Malformed C line sent by remote SCP binary, illegal file length.");
            }

            LenNamePair lnp = new LenNamePair();
            lnp.Length = len;
            lnp.Filename = name_substring;

            return lnp;
        }

        /// <summary>
        /// Read a response from the server (a single byte).
        /// </summary>
        /// <param name="istr">The underlying input stream.</param>
        /// <exception cref="IOException"></exception>
        private static void ReadResponse(Stream istr)
        {
            int c = istr.ReadByte();

            if (c == 0)
            {
                return;
            }

            if (c == -1)
            {
                throw new IOException("Remote scp terminated unexpectedly.");
            }

            if ((c != 1) && (c != 2))
            {
                throw new IOException("Remote scp sent illegal error code.");
            }

            if (c == 2)
            {
                throw new IOException("Remote scp terminated with error.");
            }

            string err = SCPClient.ReceiveLine(istr);
            throw new IOException("Remote scp terminated with error (" + err + ").");
        }

        /// <summary>
        /// Sends a byte array to the server.
        /// </summary>
        /// <param name="sess">The underlying session.</param>
        /// <param name="data">The byte array to send.</param>
        /// <param name="fileName">The file name we are sending to.</param>
        /// <param name="mode">A four digit string (e.g., 0644, see "man chmod", "man open").</param>
        private static void SendBytes(Session sess, byte[] data, string fileName, string mode)
        {
            Stream os = sess.StandardInput;
            Stream istr = new BufferedStream(sess.StandardOutput, 512);

            SCPClient.ReadResponse(istr);

            string cline = "C" + mode + " " + data.Length + " " + fileName + "\n";
            byte[] bcline = StringEncoder.GetBytes(cline);
            os.Write(bcline, 0, bcline.Length);
            os.Flush();

            SCPClient.ReadResponse(istr);

            os.Write(data, 0, data.Length);
            os.WriteByte(0);
            os.Flush();

            SCPClient.ReadResponse(istr);

            byte[] temp = StringEncoder.GetBytes("E\n");
            os.Write(temp, 0, temp.Length);
            os.Flush();
        }

        /// <summary>
        /// Send a list of local files to a list of remote files created with a specified mode.
        /// </summary>
        /// <param name="sess">The underlying session.</param>
        /// <param name="files">An array of local files to send.</param>
        /// <param name="remoteFiles">An array of remote files to send to.</param>
        /// <param name="mode">A four digit string (e.g., 0644, see "man chmod", "man open").</param>
        private static void SendFiles(Session sess, string[] files, string[] remoteFiles, string mode)
        {
            byte[] buffer = new byte[8192];
            byte[] temp;

            Stream os = new BufferedStream(sess.StandardInput, 40000);
            Stream istr = new BufferedStream(sess.StandardOutput, 512);

            SCPClient.ReadResponse(istr);

            for (int i = 0; i < files.Length; i++)
            {
                FileInfo f = new FileInfo(files[i]);
                long remain = f.Length;

                string remoteName;

                if ((remoteFiles != null) && (remoteFiles.Length > i) && (remoteFiles[i] != null))
                {
                    remoteName = remoteFiles[i];
                }
                else
                {
                    remoteName = f.Name;
                }

                string cline = "C" + mode + " " + remain + " " + remoteName + "\n";

                temp = StringEncoder.GetBytes(cline);
                os.Write(temp, 0, temp.Length);
                os.Flush();

                SCPClient.ReadResponse(istr);

                FileStream fis = null;

                try
                {
                    fis = new FileStream(f.FullName, System.IO.FileMode.Open, System.IO.FileAccess.Read);

                    while (remain > 0)
                    {
                        int trans;
                        if (remain > buffer.Length)
                        {
                            trans = buffer.Length;
                        }
                        else
                        {
                            trans = (int)remain;
                        }

                        if (fis.Read(buffer, 0, trans) != trans)
                        {
                            throw new IOException("Cannot read enough from local file " + files[i]);
                        }

                        os.Write(buffer, 0, trans);

                        remain -= trans;
                    }
                }
                finally
                {
                    if (fis != null)
                    {
                        fis.Close();
                    }
                }

                os.WriteByte(0);
                os.Flush();

                SCPClient.ReadResponse(istr);
            }

            temp = StringEncoder.GetBytes("E\n");
            os.Write(temp, 0, temp.Length);
            os.Flush();
        }

        /// <summary>
        /// Receive files to an array of Streams.
        /// </summary>
        /// <param name="sess">The underlying session.</param>
        /// <param name="targets">
        /// Array of OutputStreams where the contents of the file will be sent to.
        /// Normally called with just a single stream in the array.
        /// </param>
        /// <exception cref="IOException"></exception>
        private static void ReceiveFiles(Session sess, Stream[] targets)
        {
            byte[] buffer = new byte[8192];

            Stream os = new BufferedStream(sess.StandardInput, 512);
            Stream istr = new BufferedStream(sess.StandardOutput, 40000);

            os.WriteByte(0x0);
            os.Flush();

            for (int i = 0; i < targets.Length; i++)
            {
                LenNamePair lnp;

                while (true)
                {
                    int c = istr.ReadByte();
                    if (c < 0)
                    {
                        throw new IOException("Remote scp terminated unexpectedly.");
                    }

                    string line = SCPClient.ReceiveLine(istr);

                    if (c == 'T')
                    {
                        // Ignore modification times
                        continue;
                    }

                    if ((c == 1) || (c == 2))
                    {
                        throw new IOException("Remote SCP error: " + line);
                    }

                    if (c == 'C')
                    {
                        lnp = SCPClient.ParseCLine(line);
                        break;
                    }
                    
                    throw new IOException("Remote SCP error: " + ((char)c) + line);
                }

                os.WriteByte(0x0);
                os.Flush();

                long remain = lnp.Length;

                while (remain > 0)
                {
                    int trans;
                    if (remain > buffer.Length)
                    {
                        trans = buffer.Length;
                    }
                    else
                    {
                        trans = (int)remain;
                    }

                    int this_time_received = istr.Read(buffer, 0, trans);

                    if (this_time_received < 0)
                    {
                        throw new IOException("Remote scp terminated connection unexpectedly");
                    }

                    targets[i].Write(buffer, 0, this_time_received);

                    remain -= this_time_received;
                }

                SCPClient.ReadResponse(istr);

                os.WriteByte(0x0);
                os.Flush();
            }
        }

        /// <summary>
        /// Receive files from the remote server.
        /// </summary>
        /// <param name="sess">The underlying session.</param>
        /// <param name="files">An array of file names to receive.</param>
        /// <param name="target">The local target directory.</param>
        /// <exception cref="IOException"></exception>
        private static void ReceiveFiles(Session sess, string[] files, string target)
        {
            byte[] buffer = new byte[8192];

            Stream os = new BufferedStream(sess.StandardInput, 512);
            Stream istr = new BufferedStream(sess.StandardOutput, 40000);

            os.WriteByte(0x0);
            os.Flush();

            for (int i = 0; i < files.Length; i++)
            {
                LenNamePair lnp;

                while (true)
                {
                    int c = istr.ReadByte();
                    if (c < 0)
                    {
                        throw new IOException("Remote scp terminated unexpectedly.");
                    }

                    string line = SCPClient.ReceiveLine(istr);

                    if (c == 'T')
                    {
                        // Ignore modification times
                        continue;
                    }

                    if ((c == 1) || (c == 2))
                    {
                        throw new IOException("Remote SCP error: " + line);
                    }

                    if (c == 'C')
                    {
                        lnp = SCPClient.ParseCLine(line);
                        break;
                    }

                    throw new IOException("Remote SCP error: " + ((char)c) + line);
                }

                os.WriteByte(0x0);
                os.Flush();

                FileInfo f = new FileInfo(target + Path.DirectorySeparatorChar + lnp.Filename);
                FileStream fop = null;

                try
                {
                    fop = new FileStream(f.FullName, FileMode.Create);

                    long remain = lnp.Length;

                    while (remain > 0)
                    {
                        int trans;
                        if (remain > buffer.Length)
                        {
                            trans = buffer.Length;
                        }
                        else
                        {
                            trans = (int)remain;
                        }

                        int this_time_received = istr.Read(buffer, 0, trans);

                        if (this_time_received < 0)
                        {
                            throw new IOException("Remote scp terminated connection unexpectedly");
                        }

                        fop.Write(buffer, 0, this_time_received);

                        remain -= this_time_received;
                    }
                }
                finally
                {
                    if (fop != null)
                    {
                        fop.Flush();
                        fop.Close();
                    }
                }

                SCPClient.ReadResponse(istr);

                os.WriteByte(0x0);
                os.Flush();
            }
        }

        /// <summary>
        /// Downloads an array of files from a remote server to an array of streams.
        /// </summary>
        /// <param name="remoteFiles">Array of paths and names of remote files.</param>
        /// <param name="targets">Array of streams to receive the files to.</param>
        /// <exception cref="ArgumentException">If there is a problem with the arguments.</exception>
        /// <exception cref="IOException">If there is a problem during the SCP transfer.</exception>
        private void Get(string[] remoteFiles, Stream[] targets)
        {
            Session sess = null;

            if ((remoteFiles == null) || (targets == null))
            {
                throw new ArgumentException("Null argument.");
            }

            if (remoteFiles.Length != targets.Length)
            {
                throw new ArgumentException("Length of arguments does not match.");
            }

            if (remoteFiles.Length == 0)
            {
                return;
            }

            string cmd = "scp -f";

            for (int i = 0; i < remoteFiles.Length; i++)
            {
                if (remoteFiles[i] == null)
                {
                    throw new ArgumentException("Cannot accept null filename.");
                }

                string tmp = remoteFiles[i].Trim();

                if (tmp.Length == 0)
                {
                    throw new ArgumentException("Cannot accept empty filename.");
                }

                cmd += " " + tmp;
            }

            try
            {
                sess = this.conn.OpenSession();
                sess.ExecCommand(cmd);
                SCPClient.ReceiveFiles(sess, targets);
            }
            catch (IOException)
            {
                throw (IOException)new IOException("Error during SCP transfer.");
            }
            finally
            {
                if (sess != null)
                {
                    sess.Close();
                }
            }
        }
        
        /// <summary>
        /// Class to store a length and file name.
        /// </summary>
        private class LenNamePair
        {
            private long length;
            private string filename;
            
            /// <summary>
            /// Gets or sets the length of the file.
            /// </summary>
            public long Length
            {
                get { return this.length; }
                set { this.length = value; }
            }
            
            /// <summary>
            /// Gets or sets the file name.
            /// </summary>
            public string Filename
            {
                get { return this.filename; }
                set { this.filename = value; }
            }
        }
    }
}
