﻿// <copyright file="AttribFlags.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> AttribFlags
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;

namespace Eunomia.SSH2.SFTP
{
    /// <summary>
    /// Attribute Flags. The 'valid-attribute-flags' field in
    /// the SFTP ATTRS data type specifies which of the fields are actually present.
    /// </summary>
    [SuppressMessage("Microsoft.Performance", "CA1815:OverrideEqualsAndOperatorEqualsOnValueTypes",
                     Justification = "We will never compare or sort instances of this struct.")]
    [SuppressMessage("Microsoft.Naming", "CA1726:UsePreferredTerms", MessageId = "Flags",
                     Justification = "Class name need to match (except for case) the ganymed Java library.")]
    public struct AttribFlags
    {
        /// <summary>
        /// Indicates that the 'allocation-size' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeSize = 0x00000001;

        /// <summary>
        /// Protocol version 6:
        /// 0x00000002 was used in a previous version of this protocol.
        /// It is now a reserved value and MUST NOT appear in the mask.
        /// Some future version of this protocol may reuse this value.
        /// </summary>
        public const int SSHFileTransferAttributeV3UIDGID = 0x00000002;

        /// <summary>
        /// Indicates that the 'permissions' field is present.
        /// </summary>
        public const int SSHFileTransferAttributePermissions = 0x00000004;

        /// <summary>
        /// Indicates that the 'atime' and 'mtime' field are present
        /// (protocol v3).
        /// </summary>
        public const int SSHFileTransferAttributeV3ACModtime = 0x00000008;

        /// <summary>
        /// Indicates that the 'atime' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeAccessTime = 0x00000008;

        /// <summary>
        /// Indicates that the 'createtime' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeCreateTime = 0x00000010;

        /// <summary>
        /// Indicates that the 'mtime' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeModifyTime = 0x00000020;

        /// <summary>
        /// Indicates that the 'acl' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeACL = 0x00000040;

        /// <summary>
        /// Indicates that the 'owner' and 'group' fields are present.
        /// </summary>
        public const int SSHFileTransferAttributeOwnerGroup = 0x00000080;

        /// <summary>
        /// Indicates that additionally to the 'atime', 'createtime',
        /// 'mtime' and 'ctime' fields (if present), there is also
        /// 'atime-nseconds', 'createtime-nseconds',  'mtime-nseconds'
        /// and 'ctime-nseconds'.
        /// </summary>
        public const int SSHFileTransferAttributeSubSecondTimes = 0x00000100;

        /// <summary>
        /// Indicates that the 'attrib-bits' and 'attrib-bits-valid'
        /// fields are present.
        /// </summary>
        public const int SSHFileTransferAttributeBits = 0x00000200;

        /// <summary>
        /// Indicates that the 'allocation-size' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeAllocationSize = 0x00000400;

        /// <summary>
        /// Indicates that the 'text-hint' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeTextHint = 0x00000800;

        /// <summary>
        /// Indicates that the 'mime-type' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeMimeType = 0x00001000;

        /// <summary>
        /// Indicates that the 'link-count' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeLinkCount = 0x00002000;

        /// <summary>
        /// Indicates that the 'untranslated-name' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeUntranslatedName = 0x00004000;

        /// <summary>
        /// Indicates that the 'ctime' field is present.
        /// </summary>
        public const int SSHFileTransferAttributeCTime = 0x00008000;

        /// <summary>
        /// Indicates that the 'extended-count' field (and probablby some
        /// 'extensions') is present.
        /// </summary>
        public const int SSHFileTransferAttributeExtended = unchecked((int)0x80000000);
    }
}