﻿// <copyright file="RSASHA1Verify.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> RSASHA1Verify
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;
using Eunomia.SSH2.Crypto;
using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Log;
using Eunomia.SSH2.Packets;
using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Signature
{
    /// <summary>
    /// Class full of static methods to decode and encode public keys and
    /// signatures, and verify and generate signatures.
    /// </summary>
    public static class RSASHA1Verify
    {
        private static Logger log = Logger.GetLogger(typeof(RSASHA1Verify));

        /// <summary>
        /// Decode a RSA public key.
        /// </summary>
        /// <param name="key">The key to decode.</param>
        /// <returns>The RSAPublicKey.</returns>
        public static RSAPublicKey DecodeSSHRSAPublicKey(byte[] key)
        {
            TypesReader tr = new TypesReader(key);

            string key_format = tr.ReadString();

            if (key_format.Equals("ssh-rsa") == false)
            {
                throw new ArgumentException("This is not a ssh-rsa public key");
            }

            BigInteger e = tr.ReadMPINT();
            BigInteger n = tr.ReadMPINT();

            if (tr.Remain != 0)
            {
                throw new IOException("Padding in RSA public key!");
            }

            return new RSAPublicKey(e, n);
        }

        /// <summary>
        /// Encode a RSA public key.
        /// </summary>
        /// <param name="publicKey">The RSAPublicKey to encode.</param>
        /// <returns>The encoded key as a byte array.</returns>
        public static byte[] EncodeSSHRSAPublicKey(RSAPublicKey publicKey)
        {
            TypesWriter tw = new TypesWriter();

            tw.WriteString("ssh-rsa");
            tw.WriteMPInt(publicKey.E);
            tw.WriteMPInt(publicKey.N);

            return tw.GetBytes();
        }

        /// <summary>
        /// Decode a RSA signature.
        /// </summary>
        /// <param name="signature">The signature to decoode.</param>
        /// <returns>The RSA signature.</returns>
        public static RSASignature DecodeSSHRSASignature(byte[] signature)
        {
            TypesReader tr = new TypesReader(signature);

            string sig_format = tr.ReadString();

            if (sig_format.Equals("ssh-rsa") == false)
            {
                throw new IOException("Peer sent wrong signature format");
            }

            // S is NOT an MPINT. "The value for 'rsa_signature_blob' is encoded as a string
            // containing s (which is an integer, without lengths or padding, unsigned and in
            // network byte order)." See also below.
            byte[] s = tr.ReadByteString();

            if (s.Length == 0)
            {
                throw new IOException("Error in RSA signature, S is empty.");
            }

            if (log.Enabled)
            {
                log.Log(80, "Decoding ssh-rsa signature string (length: " + s.Length + ")");
            }

            if (tr.Remain != 0)
            {
                throw new IOException("Padding in RSA signature!");
            }

            return new RSASignature(new BigInteger(1, s));
        }

        /// <summary>
        /// Encode a RSA signature.
        /// </summary>
        /// <param name="signature">The RSASignature to encode.</param>
        /// <returns>The encoded signature as a byte array.</returns>
        public static byte[] EncodeSSHRSASignature(RSASignature signature)
        {
            TypesWriter tw = new TypesWriter();

            tw.WriteString("ssh-rsa");

            // S is NOT an MPINT. "The value for 'rsa_signature_blob' is encoded as a string
            // containing s (which is an integer, without lengths or padding, unsigned and in
            // network byte order)."
            byte[] s = signature.Signature.ToByteArray();

            // Remove first zero sign byte, if present
            if ((s.Length > 1) && (s[0] == 0x00))
            {
                tw.WriteString(s, 1, s.Length - 1);
            }
            else
            {
                tw.WriteString(s, 0, s.Length);
            }

            return tw.GetBytes();
        }

        /// <summary>
        /// Generate a signature for a message and private key.
        /// </summary>
        /// <param name="message">A byte array containing the message.</param>
        /// <param name="privateKey">The private key.</param>
        /// <returns>The RSA signature.</returns>
        public static RSASignature GenerateSignature(byte[] message, RSAPrivateKey privateKey)
        {
            SHA1 md = new SHA1();
            md.Update(message);
            byte[] sha_message = new byte[md.DigestLength];
            md.Digest(sha_message);

            byte[] der_header = new byte[] { 0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x1a, 0x05, 0x00, 0x04, 0x14 };

            int rsa_block_len = (privateKey.N.BitLength + 7) / 8;

            int num_pad = rsa_block_len - (2 + der_header.Length + sha_message.Length) - 1;

            if (num_pad < 8)
            {
                throw new IOException("Cannot sign with RSA, message too long");
            }

            byte[] sig = new byte[der_header.Length + sha_message.Length + 2 + num_pad];

            sig[0] = 0x01;

            for (int i = 0; i < num_pad; i++)
            {
                sig[i + 1] = (byte)0xff;
            }

            sig[num_pad + 1] = 0x00;

            Array.Copy(der_header, 0, sig, 2 + num_pad, der_header.Length);
            Array.Copy(sha_message, 0, sig, 2 + num_pad + der_header.Length, sha_message.Length);

            BigInteger m = new BigInteger(1, sig);

            BigInteger s = m.ModPow(privateKey.D, privateKey.N);

            return new RSASignature(s);
        }

        /// <summary>
        /// Verify the RSASignature.
        /// </summary>
        /// <param name="message">The message as a byte array.</param>
        /// <param name="ds">The signature.</param>
        /// <param name="publicKey">The public key.</param>
        /// <returns>True if the message verifies.</returns>
        public static bool VerifySignature(byte[] message, RSASignature ds, RSAPublicKey publicKey)
        {
            SHA1 md = new SHA1();
            md.Update(message);
            byte[] sha_message = new byte[md.DigestLength];
            md.Digest(sha_message);

            BigInteger n = publicKey.N;
            BigInteger e = publicKey.E;
            BigInteger s = ds.Signature;

            if (n.CompareTo(s) <= 0)
            {
                log.Log(20, "ssh-rsa signature: n.compareTo(s) <= 0");
                return false;
            }

            int rsa_block_len = (n.BitLength + 7) / 8;

            // And now the show begins
            if (rsa_block_len < 1)
            {
                log.Log(20, "ssh-rsa signature: rsa_block_len < 1");
                return false;
            }

            byte[] v = s.ModPow(e, n).ToByteArray();

            int startpos = 0;

            if ((v.Length > 0) && (v[0] == 0x00))
            {
                startpos++;
            }

            if ((v.Length - startpos) != (rsa_block_len - 1))
            {
                log.Log(20, "ssh-rsa signature: (v.length - startpos) != (rsa_block_len - 1)");
                return false;
            }

            if (v[startpos] != 0x01)
            {
                log.Log(20, "ssh-rsa signature: v[startpos] != 0x01");
                return false;
            }

            int pos = startpos + 1;

            while (true)
            {
                if (pos >= v.Length)
                {
                    log.Log(20, "ssh-rsa signature: pos >= v.length");
                    return false;
                }

                if (v[pos] == 0x00)
                {
                    break;
                }

                if (v[pos] != (byte)0xff)
                {
                    log.Log(20, "ssh-rsa signature: v[pos] != (byte) 0xff");
                    return false;
                }

                pos++;
            }

            int num_pad = pos - (startpos + 1);

            if (num_pad < 8)
            {
                log.Log(20, "ssh-rsa signature: num_pad < 8");
                return false;
            }

            pos++;

            if (pos >= v.Length)
            {
                log.Log(20, "ssh-rsa signature: pos >= v.length");
                return false;
            }

            SimpleDERReader dr = new SimpleDERReader(v, pos, v.Length - pos);

            byte[] seq = dr.ReadSequenceAsByteArray();

            if (dr.Available() != 0)
            {
                log.Log(20, "ssh-rsa signature: dr.available() != 0");
                return false;
            }

            dr.ResetInput(seq);

            // Read digestAlgorithm
            byte[] digestAlgorithm = dr.ReadSequenceAsByteArray();

            // Inspired by RFC 3347, however, ignoring the comment regarding old BER based implementations
            if ((digestAlgorithm.Length < 8) || (digestAlgorithm.Length > 9))
            {
                log.Log(20, "ssh-rsa signature: (digestAlgorithm.length < 8) || (digestAlgorithm.length > 9)");
                return false;
            }

            byte[] digestAlgorithm_sha1 = new byte[] { 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x1a, 0x05, 0x00 };

            for (int i = 0; i < digestAlgorithm.Length; i++)
            {
                if (digestAlgorithm[i] != digestAlgorithm_sha1[i])
                {
                    log.Log(20, "ssh-rsa signature: digestAlgorithm[i] != digestAlgorithm_sha1[i]");
                    return false;
                }
            }

            byte[] digest = dr.ReadOctetString();

            if (dr.Available() != 0)
            {
                log.Log(20, "ssh-rsa signature: dr.available() != 0 (II)");
                return false;
            }
            
            if (digest.Length != sha_message.Length)
            {
                log.Log(20, "ssh-rsa signature: digest.length != sha_message.length");
                return false;
            }

            for (int i = 0; i < sha_message.Length; i++)
            {
                if (sha_message[i] != digest[i])
                {
                    log.Log(20, "ssh-rsa signature: sha_message[i] != digest[i]");
                    return false;
                }
            }

            return true;
        }
    }
}
