// <copyright file="StreamGobbler.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> StreamGobbler
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Threading;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A <c>StreamGobbler</c> is an InputStream that uses an internal worker
    /// thread to constantly consume input from another InputStream. It uses a buffer
    /// to store the consumed data. The buffer size is automatically adjusted, if needed.
    /// <para>
    /// This class is sometimes very convenient - if you wrap a session's STDOUT and STDERR
    /// InputStreams with instances of this class, then you don't have to bother about
    /// the shared window of STDOUT and STDERR in the low level SSH-2 protocol,
    /// since all arriving data will be immediatelly consumed by the worker threads.
    /// Also, as a side effect, the streams will be buffered (e.g., single byte
    /// read() operations are faster).
    /// </para>
    /// <para>
    /// Other SSH for Java libraries include this functionality by default in
    /// their STDOUT and STDERR InputStream implementations, however, please be aware
    /// that this approach has also a downside:
    /// </para>
    /// <para>
    /// If you do not call the StreamGobbler's <c>read()</c> method often enough
    /// and the peer is constantly sending huge amounts of data, then you will sooner or later
    /// encounter a low memory situation due to the aggregated data (well, it also depends on the Java heap size).
    /// Joe Average will like this class anyway - a paranoid programmer would never use such an approach.
    /// </para>
    /// <para>
    /// The term "StreamGobbler" was taken from an article called "When Runtime.exec() won't",
    /// see http://www.javaworld.com/javaworld/jw-12-2000/jw-1229-traps.html.
    /// </para>
    /// </summary>
    [SuppressMessage("Microsoft.Naming", "CA1710:IdentifiersShouldHaveCorrectSuffix",
                    Justification = "Use StreamGobbler so that we are compatible to ganymed.")]
    public class StreamGobbler : Stream
    {
        private Stream istr;
        private GobblerThread t;

        private object synchronizer;

        private bool isEOF = false;
        private bool isClosed = false;
        private IOException exception = null;

        private byte[] buffer;
        private int read_pos = 0;
        private int write_pos = 0;

        /// <summary>
        /// Initializes a new instance of the StreamGobbler class.
        /// </summary>
        /// <param name="inStream">The input stream.</param>
        public StreamGobbler(Stream inStream)
        {
            this.buffer = new byte[2048];
            this.synchronizer = new object();
            this.istr = inStream;
            this.t = new GobblerThread(this);
            this.t.IsBackground = true;
            this.t.Start();
        }

        /// <summary>
        /// Gets a value indicating whether this stream can be read from, (always true).
        /// </summary>
        public override bool CanRead
        {
            get { return true; }
        }

        /// <summary>
        /// Gets a value indicating whether this stream can be seeked on, (always false).
        /// </summary>
        public override bool CanSeek
        {
            get { return false; }
        }

        /// <summary>
        /// Gets a value indicating whether this stream can be timed out, (always false).
        /// </summary>
        public override bool CanTimeout
        {
            get { return false; }
        }

        /// <summary>
        /// Gets a value indicating whether this stream can be written to, (always false).
        /// </summary>
        public override bool CanWrite
        {
            get { return false; }
        }

        /// <summary>
        /// Gets the length of the stream (just returns the buffer length).
        /// </summary>
        public override long Length
        {
            get
            {
                lock (this.synchronizer)
                {
                    return this.write_pos - this.read_pos;
                }
            }
        }

        /// <summary>
        /// Gets the postion in the stream...not implemented.
        /// </summary>
        public override long Position
        {
            get { return 0L; }
            set { }
        }

        /// <summary>
        /// Read a single byte (returned as an int).
        /// </summary>
        /// <returns>The byte that was read (as an integer).</returns>
        public override int ReadByte()
        {
            bool wasInterrupted = false;

            try
            {
                lock (this.synchronizer)
                {
                    if (this.isClosed)
                    {
                        throw new IOException("This StreamGobbler is closed.");
                    }

                    while (this.read_pos == this.write_pos)
                    {
                        if (this.exception != null)
                        {
                            throw this.exception;
                        }

                        if (this.isEOF)
                        {
                            return 0;
                        }

                        try
                        {
                            Monitor.Wait(this.synchronizer);
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }

                    int b = this.buffer[this.read_pos++] & 0xff;

                    return b;
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }

        /// <summary>
        /// Read a byte array.
        /// </summary>
        /// <param name="buffer">The byte array to write to.</param>
        /// <returns>The number of bytes.</returns>
        public int Read(byte[] buffer)
        {
            return this.Read(buffer, 0, buffer.Length);
        }

        /// <summary>
        /// Closes the StreamGobbler (and the underlying stream).
        /// </summary>
        public override void Close()
        {
            lock (this.synchronizer)
            {
                if (this.isClosed)
                {
                    return;
                }
                
                this.isClosed = true;
                this.isEOF = true;
                Monitor.PulseAll(this.synchronizer);
                this.istr.Close();
            }
        }

        /// <summary>
        /// Read a byte array.
        /// </summary>
        /// <param name="buffer">The byte array to read into.</param>
        /// <param name="offset">The offset in the byte array where we start reading into.</param>
        /// <param name="count">The number of bytes to read into the byte array.</param>
        /// <returns>The number of bytes actually read.</returns>
        /// <exception cref="NullReferenceException">If the byte array is null.</exception>
        /// <exception cref="IndexOutOfRangeException">
        /// If we write outside the bounds of the byte array.
        /// </exception>
        public override int Read(byte[] buffer, int offset, int count)
        {
            if (buffer == null)
            {
                throw new System.ArgumentException("buffer is null");
            }

            if ((offset < 0) || (count < 0) || ((offset + count) > buffer.Length) || ((offset + count) < 0) || (offset > buffer.Length))
            {
                throw new System.ArgumentException("offset or count values are invalid");
            }

            if (count == 0)
            {
                return 0;
            }

            bool wasInterrupted = false;

            try
            {
                lock (this.synchronizer)
                {
                    if (this.isClosed)
                    {
                        throw new IOException("This StreamGobbler is closed.");
                    }

                    while (this.read_pos == this.write_pos)
                    {
                        if (this.exception != null)
                        {
                            throw this.exception;
                        }

                        if (this.isEOF)
                        {
                            return 0;
                        }

                        try
                        {
                            Monitor.Wait(this.synchronizer);
                        }
                        catch (System.Threading.ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }

                    int avail = this.write_pos - this.read_pos;

                    avail = (avail > count) ? count : avail;

                    Array.Copy(this.buffer, this.read_pos, buffer, offset, avail);

                    this.read_pos += avail;

                    return avail;
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }

        /// <summary>
        /// Sets the length of the stream, (not implemented).
        /// </summary>
        /// <param name="value">The length we want the stream to be.</param>
        public override void SetLength(long value)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Write a byte array to the stream, (not implemented).
        /// </summary>
        /// <param name="buffer">The byte array to write.</param>
        /// <param name="offset">The offset into the byte array to begin writing from.</param>
        /// <param name="count">The number of bytes to write.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Flush the stream, not implemented.
        /// </summary>
        public override void Flush()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Seek on the stream, (not implemented).
        /// </summary>
        /// <param name="offset">The offset to seek to.</param>
        /// <param name="origin">The origin to offset from.</param>
        /// <returns>The new position in the stream.</returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// The Gobbler thread runs the StreamGobbler in it's own thread.
        /// </summary>
        internal class GobblerThread : JavaThread
        {
            private StreamGobbler sg;

            /// <summary>
            /// Initializes a new instance of the GobblerThread class.
            /// </summary>
            /// <param name="sg">The StreamGobbler class instance.</param>
            public GobblerThread(StreamGobbler sg)
            {
                this.sg = sg;
            }

            /// <summary>
            /// The run method that runs in a different thread.
            /// </summary>
            public override void Run()
            {
                byte[] buff = new byte[8192];

                while (true)
                {
                    try
                    {
                        int avail = this.sg.istr.Read(buff, 0, buff.Length);
                        lock (this.sg.synchronizer)
                        {
                            if (avail <= 0)
                            {
                                this.sg.isEOF = true;
                            }
                            else
                            {
                                int space_available = this.sg.buffer.Length - this.sg.write_pos;

                                if (space_available < avail)
                                {
                                    // compact/resize buffer
                                    int unread_size = this.sg.write_pos - this.sg.read_pos;
                                    int need_space = unread_size + avail;

                                    byte[] new_buffer = this.sg.buffer;

                                    if (need_space > this.sg.buffer.Length)
                                    {
                                        int inc = need_space / 3;
                                        inc = (inc < 256) ? 256 : inc;
                                        inc = (inc > 8192) ? 8192 : inc;
                                        new_buffer = new byte[need_space + inc];
                                    }

                                    if (unread_size > 0)
                                    {
                                        Array.Copy(this.sg.buffer, this.sg.read_pos, new_buffer, 0, unread_size);
                                    }

                                    this.sg.buffer = new_buffer;

                                    this.sg.read_pos = 0;
                                    this.sg.write_pos = unread_size;
                                }
                                
                                Array.Copy(buff, 0, this.sg.buffer, this.sg.write_pos, avail);
                                this.sg.write_pos += avail;
                            }
                            
                            Monitor.PulseAll(this.sg.synchronizer);

                            // TODO: this is a kludge to avoid a tight loop
                            Thread.Sleep(1);
                        }
                    }
                    catch (IOException e)
                    {
                        lock (this.sg.synchronizer)
                        {
                            this.sg.exception = e;
                            Monitor.PulseAll(this.sg.synchronizer);
                            break;
                        }
                    }
                }
            }
        }
    }
}
