// <copyright file="KexManager.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> KexManager
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;
using System.Runtime.CompilerServices;
using System.Threading;

using Eunomia.SSH2.Crypto;
using Eunomia.SSH2.Crypto.Cipher;
using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.DH;
using Eunomia.SSH2.Log;
using Eunomia.SSH2.Packets;
using Eunomia.SSH2.Signature;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.Transport
{
    /// <summary>
    /// Class that manages a key exchange.
    /// </summary>
    public class KexManager
    {
        private static readonly Logger log = Logger.GetLogger(typeof(KexManager));
        private KexState kxs;
        private int kexCount = 0;
        private KeyMaterial km;
        private byte[] sessionId;
        private ClientServerHello clientServerHello;

        private object accessLock;
        private ConnectionInfo lastConnInfo;

        private bool connectionClosed = false;

        private bool ignore_next_kex_packet = false;

        private TransportManager tm;

        private CryptoWishList nextKEXcryptoWishList;
        private DHGexParameters nextKEXdhgexParameters;

        private IServerHostkeyVerifier verifier;
        private string hostName;
        private int port;
        private SecureRandom random;

        /// <summary>
        /// Initializes a new instance of the KexManager class.
        /// </summary>
        /// <param name="tm">The 'parent' TransportManager.</param>
        /// <param name="clientServerHello">Used for the client, server 'hello'.</param>
        /// <param name="initialCryptoWishList">The initial CryptoWishList.</param>
        /// <param name="hostName">The hostName to connect to.</param>
        /// <param name="port">The port to connect to.</param>
        /// <param name="keyVerifier">The host key verifier to use.</param>
        /// <param name="random">A SecureRandom instance to use.</param>
        public KexManager(
            TransportManager tm,
            ClientServerHello clientServerHello,
            CryptoWishList initialCryptoWishList,
            string hostName,
            int port,
            IServerHostkeyVerifier keyVerifier,
            SecureRandom random)
        {
            this.accessLock = new object();
            this.tm = tm;
            this.clientServerHello = clientServerHello;
            this.nextKEXcryptoWishList = initialCryptoWishList;
            this.nextKEXdhgexParameters = new DHGexParameters();
            this.hostName = hostName;
            this.port = port;
            this.verifier = keyVerifier;
            this.random = random;
        }

        /// <summary>
        /// Gets the possible server host key algorithms.
        /// </summary>
        public static string[] DefaultServerHostkeyAlgorithmList
        {
            get { return new string[] { "ssh-rsa", "ssh-dss" }; }
        }

        /// <summary>
        /// Gets the possible KEX algorithms.
        /// </summary>
        public static string[] DefaultKexAlgorithmList
        {
            get
            {
                return new string[]
                {
                    "diffie-hellman-group-exchange-sha1",
                    "diffie-hellman-group14-sha1",
                    "diffie-hellman-group1-sha1"
                };
            }
        }
        
        /// <summary>
        /// Gets the session id.
        /// </summary>
        public byte[] SessionId
        {
            get { return this.sessionId; }
        }

        /// <summary>
        /// Static method to check that the host key algorithms are sane.
        /// </summary>
        /// <param name="algos">String array of algorithms to check. </param>
        /// <exception cref="ArgumentException">If it finds an unknown algorithm.</exception>
        public static void CheckServerHostkeyAlgorithmsList(string[] algos)
        {
            for (int i = 0; i < algos.Length; i++)
            {
                if (("ssh-rsa".Equals(algos[i]) == false) && ("ssh-dss".Equals(algos[i]) == false))
                {
                    throw new ArgumentException("Unknown server host key algorithm '" + algos[i] + "'");
                }
            }
        }

        /// <summary>
        /// Static method to check that the kex algorithms are sane.
        /// </summary>
        /// <param name="algos">String array of algorithms to check. </param>
        /// <exception cref="ArgumentException">If it finds an unknown algorithm.</exception>
        public static void CheckKexAlgorithmList(string[] algos)
        {
            for (int i = 0; i < algos.Length; i++)
            {
                if ("diffie-hellman-group-exchange-sha1".Equals(algos[i]))
                {
                    continue;
                }

                if ("diffie-hellman-group14-sha1".Equals(algos[i]))
                {
                    continue;
                }

                if ("diffie-hellman-group1-sha1".Equals(algos[i]))
                {
                    continue;
                }

                throw new ArgumentException("Unknown kex algorithm '" + algos[i] + "'");
            }
        }

        /// <summary>
        /// Gets or waits for connection information to be received.
        /// </summary>
        /// <param name="minKexCount">
        /// The minimum number of key exchanges that have to be complete before we return.
        /// </param>
        /// <returns>The last ConnectionInfo class 'received'.</returns>
        public ConnectionInfo GetOrWaitForConnectionInfo(int minKexCount)
        {
            bool wasInterrupted = false;

            try
            {
                lock (this.accessLock)
                {
                    while (true)
                    {
                        if ((this.lastConnInfo != null)
                            && (this.lastConnInfo.KeyExchangeCounter >= minKexCount))
                        {
                            return this.lastConnInfo;
                        }

                        if (this.connectionClosed)
                        {
                            // TODO: get ConnectionClosedCause from tm!
                            throw new IOException("Key exchange was not finished, connection is closed.");
                        }
                        
                        try
                        {
                            Monitor.Wait(this.accessLock);
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    Thread.CurrentThread.Interrupt();
                }
            }
        }

        /// <summary>
        /// Initiate a key exchange.
        /// </summary>
        /// <param name="cryptoWishList">
        /// The CryptoWishList with the possible key exchange algorithms.
        /// </param>
        /// <param name="dhgex">The DHGex parameters to use.</param>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void InitiateKEX(CryptoWishList cryptoWishList, DHGexParameters dhgex)
        {
            this.nextKEXcryptoWishList = cryptoWishList;
            this.nextKEXdhgexParameters = dhgex;

            if (this.kxs == null)
            {
                this.kxs = new KexState();

                this.kxs.DHGexParameters = this.nextKEXdhgexParameters;
                PacketKexInit kp = new PacketKexInit(this.nextKEXcryptoWishList, this.random);
                this.kxs.LocalKEX = kp;
                this.tm.SendKexMessage(kp.Payload);
            }
        }

        /// <summary>
        /// Handle received messages (called from TransportManager).
        /// KexManager does NOT inherit from MessageHandler so this method
        /// signature does not HAVE to be the same....but it is.
        /// </summary>
        /// <param name="msg">The message/packet to process.</param>
        /// <param name="msgLength">The length of the message/packet.</param>
        /// <exception cref="IOException">If the data we receive is bad.</exception>
        /// <exception cref="InvalidOperationException">If the 'state' is not sensible.</exception>
        [MethodImpl(MethodImplOptions.Synchronized)]
        public void HandleMessage(byte[] msg, int msgLength)
        {
            PacketKexInit kip;

            if (msg == null)
            {
                lock (this.accessLock)
                {
                    this.connectionClosed = true;
                    Monitor.PulseAll(this.accessLock);
                    return;
                }
            }

            if ((this.kxs == null) && (msg[0] != Packet.SSHMessageKEXInitialize))
            {
                throw new IOException("Unexpected KEX message (type " + msg[0] + ")");
            }

            if (this.ignore_next_kex_packet)
            {
                this.ignore_next_kex_packet = false;
                return;
            }

            if (msg[0] == Packet.SSHMessageKEXInitialize)
            {
                if ((this.kxs != null) && (this.kxs.State != 0))
                {
                    throw new IOException("Unexpected SSHMessageKEXINIT message during on-going kex exchange!");
                }

                if (this.kxs == null)
                {
                    // Ah, OK, peer wants to do KEX. Let's be nice and play
                    // together.
                    this.kxs = new KexState();
                    this.kxs.DHGexParameters = this.nextKEXdhgexParameters;
                    kip = new PacketKexInit(this.nextKEXcryptoWishList, this.random);
                    this.kxs.LocalKEX = kip;
                    this.tm.SendKexMessage(kip.Payload);
                }

                kip = new PacketKexInit(msg, 0, msgLength);
                this.kxs.RemoteKEX = kip;

                this.kxs.NP = KexManager.MergeKexParameters(
                    this.kxs.LocalKEX.KexParameters,
                    this.kxs.RemoteKEX.KexParameters);

                if (this.kxs.NP.NegotiateSuccessful == false)
                {
                    throw new IOException("Cannot negotiate, proposals do not match.");
                }

                if (this.kxs.RemoteKEX.FirstKexPacketFollows && (this.kxs.NP.GuessOk == false))
                {
                     // Guess was wrong, we need to ignore the next kex packet.
                    this.ignore_next_kex_packet = true;
                }

                if (this.kxs.NP.KexAlgo.Equals("diffie-hellman-group-exchange-sha1"))
                {
                    if (this.kxs.DHGexParameters.MinGroupLength == 0)
                    {
                        PacketKexDHGexRequestOld dhgexreq = new PacketKexDHGexRequestOld(this.kxs.DHGexParameters);
                        this.tm.SendKexMessage(dhgexreq.Payload);
                    }
                    else
                    {
                        PacketKexDHGexRequest dhgexreq = new PacketKexDHGexRequest(this.kxs.DHGexParameters);
                        this.tm.SendKexMessage(dhgexreq.Payload);
                    }
                    
                    this.kxs.State = 1;
                    return;
                }

                if (this.kxs.NP.KexAlgo.Equals("diffie-hellman-group1-sha1")
                    || this.kxs.NP.KexAlgo.Equals("diffie-hellman-group14-sha1"))
                {
                    this.kxs.DHX = new DHExchange();

                    if (this.kxs.NP.KexAlgo.Equals("diffie-hellman-group1-sha1"))
                    {
                        this.kxs.DHX.Init(1, this.random);
                    }
                    else
                    {
                        this.kxs.DHX.Init(14, this.random);
                    }

                    PacketKexDHInit kp = new PacketKexDHInit(this.kxs.DHX.E);
                    this.tm.SendKexMessage(kp.Payload);
                    this.kxs.State = 1;
                    return;
                }

                throw new InvalidOperationException("Unkown KEX method!");
            }

            if (msg[0] == Packet.SSHMessageNewKeys)
            {
                if (this.km == null)
                {
                    throw new IOException("Peer sent SSHMessageNEWKEYS, but I have no key material ready!");
                }

                IBlockCipher cbc;
                MAC mac;

                try
                {
                    cbc = BlockCipherFactory.CreateCipher(
                        this.kxs.NP.EncAlgoServerToClient,
                        false,
                        this.km.EncKeyServerToClient,
                        this.km.InitialIVServerToClient);

                    mac = new MAC(this.kxs.NP.MacAlgoServerToClient, this.km.IntegrityKeyServerToClient);
                }
                catch (ArgumentException)
                {
                    throw new IOException("Fatal error during MAC startup!");
                }

                this.tm.ChangeReceiveCipher(cbc, mac);

                ConnectionInfo sci = new ConnectionInfo();

                this.kexCount++;

                sci.KeyExchangeAlgorithm = this.kxs.NP.KexAlgo;
                sci.KeyExchangeCounter = this.kexCount;
                sci.ClientToServerCryptoAlgorithm = this.kxs.NP.EncAlgoClientToServer;
                sci.ServerToClientCryptoAlgorithm = this.kxs.NP.EncAlgoServerToClient;
                sci.ClientToServerMACAlgorithm = this.kxs.NP.MacAlgoClientTOServer;
                sci.ServerToClientMACAlgorithm = this.kxs.NP.MacAlgoServerToClient;
                sci.ServerHostkeyAlgorithm = this.kxs.NP.ServerHostkeyAlgo;
                sci.ServerHostkey = this.kxs.Hostkey;

                lock (this.accessLock)
                {
                    this.lastConnInfo = sci;
                    Monitor.PulseAll(this.accessLock);
                }

                this.kxs = null;
                return;
            }

            if ((this.kxs == null) || (this.kxs.State == 0))
            {
                throw new IOException("Unexpected Kex submessage!");
            }

            if (this.kxs.NP.KexAlgo.Equals("diffie-hellman-group-exchange-sha1"))
            {
                if (this.kxs.State == 1)
                {
                    PacketKexDHGexGroup dhgexgrp = new PacketKexDHGexGroup(msg, 0, msgLength);
                    this.kxs.DHGX = new DHGroupExchange(dhgexgrp.P, dhgexgrp.G);
                    this.kxs.DHGX.Init(this.random);
                    PacketKexDHGexInit dhgexinit = new PacketKexDHGexInit(this.kxs.DHGX.E);
                    this.tm.SendKexMessage(dhgexinit.Payload);
                    this.kxs.State = 2;
                    return;
                }

                if (this.kxs.State == 2)
                {
                    PacketKexDHGexReply dhgexrpl = new PacketKexDHGexReply(msg, 0, msgLength);

                    this.kxs.Hostkey = dhgexrpl.Hostkey;

                    if (this.verifier != null)
                    {
                        bool vres = false;

                        try
                        {
                            vres = this.verifier.VerifyServerHostkey(
                                this.hostName,
                                this.port,
                                this.kxs.NP.ServerHostkeyAlgo,
                                this.kxs.Hostkey);
                        }
                        catch (Exception e)
                        {
                            throw new IOException("The server hostkey was not accepted by the verifier callback.  " + e.Message);
                        }

                        if (vres == false)
                        {
                            throw new IOException("The server hostkey was not accepted by the verifier callback");
                        }
                    }

                    this.kxs.DHGX.F = dhgexrpl.F;

                    try
                    {
                        this.kxs.H = this.kxs.DHGX.CalculateH(
                            this.clientServerHello.GetClientString(),
                            this.clientServerHello.GetServerString(),
                            this.kxs.LocalKEX.Payload,
                            this.kxs.RemoteKEX.Payload,
                            dhgexrpl.Hostkey,
                            this.kxs.DHGexParameters);
                    }
                    catch (ArgumentException e)
                    {
                        throw new IOException("KEX error.  " + e.Message);
                    }

                    bool res = this.VerifySignature(dhgexrpl.Signature, this.kxs.Hostkey);

                    if (res == false)
                    {
                        throw new IOException("Hostkey signature sent by remote is wrong!");
                    }

                    this.kxs.K = this.kxs.DHGX.K;

                    this.FinishKex();
                    this.kxs.State = -1;
                    return;
                }

                throw new InvalidOperationException("Illegal State in KEX Exchange!");
            }

            if (this.kxs.NP.KexAlgo.Equals("diffie-hellman-group1-sha1")
                || this.kxs.NP.KexAlgo.Equals("diffie-hellman-group14-sha1"))
            {
                if (this.kxs.State == 1)
                {
                    PacketKexDHReply dhr = new PacketKexDHReply(msg, 0, msgLength);

                    this.kxs.Hostkey = dhr.Hostkey;

                    if (this.verifier != null)
                    {
                        bool vres = false;

                        try
                        {
                            vres = this.verifier.VerifyServerHostkey(
                                this.hostName,
                                this.port,
                                this.kxs.NP.ServerHostkeyAlgo,
                                this.kxs.Hostkey);
                        }
                        catch (Exception e)
                        {
                            throw new IOException("The server hostkey was not accepted by the verifier callback.  " + e.Message);
                        }

                        if (vres == false)
                        {
                            throw new IOException("The server hostkey was not accepted by the verifier callback");
                        }
                    }

                    this.kxs.DHX.F = dhr.F;

                    try
                    {
                        this.kxs.H = this.kxs.DHX.CalculateH(
                            this.clientServerHello.GetClientString(),
                            this.clientServerHello.GetServerString(),
                            this.kxs.LocalKEX.Payload,
                            this.kxs.RemoteKEX.Payload,
                            dhr.Hostkey);
                    }
                    catch (ArgumentException e)
                    {
                        throw new IOException("KEX error.  " + e.Message);
                    }

                    bool res = this.VerifySignature(dhr.Signature, this.kxs.Hostkey);

                    if (res == false)
                    {
                        throw new IOException("Hostkey signature sent by remote is wrong!");
                    }

                    this.kxs.K = this.kxs.DHX.K;

                    this.FinishKex();
                    this.kxs.State = -1;
                    return;
                }
            }

            throw new InvalidOperationException("Unkown KEX method! (" + this.kxs.NP.KexAlgo + ")");
        }

        /// <summary>
        /// Get the first match between two string arrays.
        /// </summary>
        /// <param name="client">The client string array.</param>
        /// <param name="server">The server string array.</param>
        /// <returns>The string that matched.</returns>
        /// <exception cref="ArgumentException">If either string array is null.</exception>
        /// <exception cref="NegotiateException">If a match is not found.</exception>
        private static string GetFirstMatch(string[] client, string[] server)
        {
            if (client == null || server == null)
            {
                throw new ArgumentException("GetFirstMatch at least one string array is null.");
            }

            if (client.Length == 0)
            {
                return null;
            }

            for (int i = 0; i < client.Length; i++)
            {
                for (int j = 0; j < server.Length; j++)
                {
                    if (client[i].Equals(server[j]))
                    {
                        return client[i];
                    }
                }
            }
            
            throw new NegotiateException();
        }

        /// <summary>
        /// Check to see if the first strings in two string arrays are equal.
        /// </summary>
        /// <param name="a">The first string array.</param>
        /// <param name="b">The second string array.</param>
        /// <returns>True if the first strings are equal.</returns>
        private static bool CompareFirstOfNameList(string[] a, string[] b)
        {
            if (a == null || b == null)
            {
                throw new ArgumentException("CompareFirstOfNameList at least one string array is null.");
            }

            if ((a.Length == 0) && (b.Length == 0))
            {
                return true;
            }

            if ((a.Length == 0) || (b.Length == 0))
            {
                return false;
            }

            return a[0].Equals(b[0]);
        }

        /// <summary>
        /// Checks to see if the kex algorithm and server host key algorithm guesses
        /// match, (the first choice of each for the client and server match).
        /// </summary>
        /// <param name="cpar">Client kex parameters to check.</param>
        /// <param name="spar">Server kex parameters to check.</param>
        /// <returns>True if the guess matches.</returns>
        private static bool IsGuessOK(KexParameters cpar, KexParameters spar)
        {
            if (KexManager.CompareFirstOfNameList(cpar.KexAlgorithms, spar.KexAlgorithms) == false)
            {
                return false;
            }

            if (KexManager.CompareFirstOfNameList(cpar.ServerHostkeyAlgorithms, spar.ServerHostkeyAlgorithms) == false)
            {
                return false;
            }

            // We do NOT check here if the other algorithms can be agreed on, this
            // is just a check if kex_algorithms and server_host_key_algorithms were
            // guessed right!
            return true;
        }

        /// <summary>
        /// Merge the kex parameters to come up with the NegotiatedParameters that
        /// both client and server can do.
        /// </summary>
        /// <param name="client">Client kex parameters.</param>
        /// <param name="server">Server kex parameters.</param>
        /// <returns>The NegotiatedParameters that will be used.</returns>
        private static NegotiatedParameters MergeKexParameters(KexParameters client, KexParameters server)
        {
            NegotiatedParameters np = new NegotiatedParameters();

            try
            {
                np.KexAlgo = KexManager.GetFirstMatch(client.KexAlgorithms, server.KexAlgorithms);

                log.Log(20, "kex_algo=" + np.KexAlgo);

                np.ServerHostkeyAlgo = KexManager.GetFirstMatch(
                    client.ServerHostkeyAlgorithms,
                    server.ServerHostkeyAlgorithms);

                log.Log(20, "server_host_key_algo=" + np.ServerHostkeyAlgo);

                np.EncAlgoClientToServer = KexManager.GetFirstMatch(
                    client.EncryptionAlgorithmsClientToServer,
                    server.EncryptionAlgorithmsClientToServer);
                np.EncAlgoServerToClient = KexManager.GetFirstMatch(
                    client.EncryptionAlgorithmsServerToClient,
                    server.EncryptionAlgorithmsServerToClient);

                log.Log(20, "enc_algo_client_to_server=" + np.EncAlgoClientToServer);
                log.Log(20, "enc_algo_server_to_client=" + np.EncAlgoServerToClient);

                np.MacAlgoClientTOServer = KexManager.GetFirstMatch(
                    client.MacAlgorithmsClientToServer,
                    server.MacAlgorithmsClientToServer);
                np.MacAlgoServerToClient = KexManager.GetFirstMatch(
                    client.MacAlgorithmsServerToClient,
                    server.MacAlgorithmsServerToClient);

                log.Log(20, "mac_algo_client_to_server=" + np.MacAlgoClientTOServer);
                log.Log(20, "mac_algo_server_to_client=" + np.MacAlgoServerToClient);

                np.CompAlgoClientToServer = KexManager.GetFirstMatch(
                    client.CompressionAlgorithmsClientToServer,
                    server.CompressionAlgorithmsClientToServer);
                np.CompAlgoServerToClient = KexManager.GetFirstMatch(
                    client.CompressionAlgorithmsServerToClient,
                    server.CompressionAlgorithmsServerToClient);

                log.Log(20, "comp_algo_client_to_server=" + np.CompAlgoClientToServer);
                log.Log(20, "comp_algo_server_to_client=" + np.CompAlgoServerToClient);
            }
            catch (NegotiateException)
            {
                np.NegotiateSuccessful = false;
                return np;
            }

            try
            {
                np.LangClientToServer = KexManager.GetFirstMatch(
                    client.LanguagesClientToServer,
                    server.LanguagesClientToServer);
            }
            catch (NegotiateException)
            {
                np.LangClientToServer = null;
            }

            try
            {
                np.LangServerToClient = KexManager.GetFirstMatch(
                    client.LanguagesServerToClient,
                    server.LanguagesServerToClient);
            }
            catch (NegotiateException)
            {
                np.LangServerToClient = null;
            }

            if (KexManager.IsGuessOK(client, server))
            {
                np.GuessOk = true;
            }
            
            np.NegotiateSuccessful = true;

            return np;
        }

        /// <summary>
        /// Establishes the key material.
        /// </summary>
        /// <returns>True if successful.</returns>
        private bool EstablishKeyMaterial()
        {
            try
            {
                int mac_cs_key_len = MAC.GetKeyLen(this.kxs.NP.MacAlgoClientTOServer);
                int enc_cs_key_len = BlockCipherFactory.GetKeySize(this.kxs.NP.EncAlgoClientToServer);
                int enc_cs_block_len = BlockCipherFactory.GetBlockSize(this.kxs.NP.EncAlgoClientToServer);

                int mac_sc_key_len = MAC.GetKeyLen(this.kxs.NP.MacAlgoServerToClient);
                int enc_sc_key_len = BlockCipherFactory.GetKeySize(this.kxs.NP.EncAlgoServerToClient);
                int enc_sc_block_len = BlockCipherFactory.GetBlockSize(this.kxs.NP.EncAlgoServerToClient);

                this.km = KeyMaterial.Create(
                    "SHA1",
                    this.kxs.H,
                    this.kxs.K,
                    this.sessionId,
                    enc_cs_key_len,
                    enc_cs_block_len,
                    mac_cs_key_len,
                    enc_sc_key_len,
                    enc_sc_block_len,
                    mac_sc_key_len);
            }
            catch (ArgumentException)
            {
                return false;
            }
            
            return true;
        }

        /// <summary>
        /// Key exchange done, inform the other side to switch to it.
        /// </summary>
        /// <exception cref="IOException">If we are not successful setting up the new MAC.
        /// </exception>
        private void FinishKex()
        {
            if (this.sessionId == null)
            {
                this.sessionId = this.kxs.H;
            }

            this.EstablishKeyMaterial();

            // Tell the other side that we start using the new material
            PacketNewKeys ign = new PacketNewKeys();
            this.tm.SendKexMessage(ign.Payload);

            IBlockCipher cbc;
            MAC mac;

            try
            {
                cbc = BlockCipherFactory.CreateCipher(
                    this.kxs.NP.EncAlgoClientToServer,
                    true,
                    this.km.EncKeyClientToServer,
                    this.km.InitialIVClientToServer);

                mac = new MAC(this.kxs.NP.MacAlgoClientTOServer, this.km.IntegrityKeyClientToServer);
            }
            catch (ArgumentException)
            {
                throw new IOException("Fatal error during MAC startup!");
            }

            this.tm.ChangeSendCipher(cbc, mac);
            this.tm.KexFinished();
        }

        /// <summary>
        /// Verify the server host key.
        /// </summary>
        /// <param name="sig">The signature to check.</param>
        /// <param name="hostkey">The host key to check.</param>
        /// <returns>True if it verifies.</returns>
        /// <exception cref="IOException">If the host key algorithm is not valid.</exception>
        private bool VerifySignature(byte[] sig, byte[] hostkey)
        {
            if (this.kxs.NP.ServerHostkeyAlgo.Equals("ssh-rsa"))
            {
                RSASignature rs = RSASHA1Verify.DecodeSSHRSASignature(sig);
                RSAPublicKey rpk = RSASHA1Verify.DecodeSSHRSAPublicKey(hostkey);

                log.Log(50, "Verifying ssh-rsa signature");

                return RSASHA1Verify.VerifySignature(this.kxs.H, rs, rpk);
            }

            if (this.kxs.NP.ServerHostkeyAlgo.Equals("ssh-dss"))
            {
                DSASignature ds = DSASHA1Verify.DecodeSSHDSASignature(sig);
                DSAPublicKey dpk = DSASHA1Verify.DecodeSSHDSAPublicKey(hostkey);

                log.Log(50, "Verifying ssh-dss signature");

                return DSASHA1Verify.VerifySignature(this.kxs.H, ds, dpk);
            }

            throw new IOException("Unknown server host key algorithm '" + this.kxs.NP.ServerHostkeyAlgo + "'");
        }
    }
}
