// <copyright file="TransportConnection.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> TransportConnection
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.IO;
using Eunomia.SSH2.Crypto.Cipher;
using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Log;
using Eunomia.SSH2.Packets;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.Transport
{
    /// <summary>
    /// Manages a Transport connection.
    /// </summary>
    public class TransportConnection : IDisposable
    {
        private static Logger log = Logger.GetLogger(typeof(TransportConnection));

        private int send_seq_number = 0;
        private int recv_seq_number = 0;
        private CipherInputStream cis;
        private CipherOutputStream cos;
        private bool useRandomPadding = false;

        // Depends on current MAC and CIPHER
        private MAC send_mac;
        private byte[] send_mac_buffer;
        private int send_padd_blocksize = 8;
        private MAC recv_mac;
        private byte[] recv_mac_buffer;
        private byte[] recv_mac_buffer_cmp;
        private int recv_padd_blocksize = 8;

        // won't change
        private byte[] send_padding_buffer = new byte[256];
        private byte[] send_packet_header_buffer = new byte[5];
        private byte[] recv_padding_buffer = new byte[256];
        private byte[] recv_packet_header_buffer = new byte[5];
        private bool recv_packet_header_present = false;

        // private ClientServerHello csh;
        private SecureRandom rnd;

        /// <summary>
        /// Initializes a new instance of the TransportConnection class.
        /// </summary>
        /// <param name="inStream">The stream used for input.</param>
        /// <param name="outStream">The stream used for output.</param>
        /// <param name="random">The SecureRandom instance to use.</param>
        public TransportConnection(Stream inStream, Stream outStream, SecureRandom random)
        {
            this.cis = new CipherInputStream(new NullCipher(), inStream);
            this.cos = new CipherOutputStream(new NullCipher(), outStream);
            this.rnd = random;
        }

        /// <summary>
        /// Gets an integer with an estimate of the packet overhead (for send operations).
        /// </summary>
        public int PacketOverheadEstimate
        {
            get { return 5 + 4 + (this.send_padd_blocksize - 1) + this.send_mac_buffer.Length; }
        }
        
        /// <summary>
        /// Implement the public IDisposable interface method.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }
        
        /// <summary>
        /// Change the receiving cipher for the connection.
        /// </summary>
        /// <param name="bc">New block cipher to use.</param>
        /// <param name="mac">New MAC to use.</param>
        public void ChangeReceiveCipher(IBlockCipher bc, MAC mac)
        {
            this.cis.ChangeCipher(bc);
            this.recv_mac = mac;
            this.recv_mac_buffer = (mac != null) ? new byte[mac.Size] : null;
            this.recv_mac_buffer_cmp = (mac != null) ? new byte[mac.Size] : null;
            this.recv_padd_blocksize = bc.BlockSize;
            if (this.recv_padd_blocksize < 8)
            {
                this.recv_padd_blocksize = 8;
            }
        }

        /// <summary>
        /// Change the sending cipher to use for the connection.
        /// </summary>
        /// <param name="bc">New block cipher to use.</param>
        /// <param name="mac">New MAC to use.</param>
        public void ChangeSendCipher(IBlockCipher bc, MAC mac)
        {
            if ((bc is NullCipher) == false)
            {
                // Only use zero byte padding for the first few packets
                this.useRandomPadding = true;
                
                // Once we start encrypting, there is no way back
            }

            this.cos.ChangeCipher(bc);
            this.send_mac = mac;
            this.send_mac_buffer = (mac != null) ? new byte[mac.Size] : null;
            this.send_padd_blocksize = bc.BlockSize;
            if (this.send_padd_blocksize < 8)
            {
                this.send_padd_blocksize = 8;
            }
        }

        /// <summary>
        /// Send a message on the outbound channel/stream.
        /// </summary>
        /// <param name="message">The byte array to send.</param>
        public void SendMessage(byte[] message)
        {
            this.SendMessage(message, 0, message.Length, 0);
        }

        /// <summary>
        /// Send a message on the outbound channel/stream.
        /// </summary>
        /// <param name="message">The byte array to send.</param>
        /// <param name="off">The offset to the first byte to send.</param>
        /// <param name="len">The number of bytes to send.</param>
        public void SendMessage(byte[] message, int off, int len)
        {
            this.SendMessage(message, off, len, 0);
        }

        /// <summary>
        /// Send a message on the outbound channel/stream.
        /// </summary>
        /// <param name="message">The byte array to send.</param>
        /// <param name="offset">The offset to the first byte to send.</param>
        /// <param name="length">The number of bytes to send.</param>
        /// <param name="padding">The number of padding bytes.</param>
        public void SendMessage(byte[] message, int offset, int length, int padding)
        {
            if (padding < 4)
            {
                padding = 4;
            }
            else if (padding > 64)
            {
                padding = 64;
            }

            int packet_len = 5 + length + padding; // Minimum allowed padding is 4
            int slack = packet_len % this.send_padd_blocksize;

            if (slack != 0)
            {
                packet_len += this.send_padd_blocksize - slack;
            }

            if (packet_len < 16)
            {
                packet_len = 16;
            }

            int padd_len = packet_len - (5 + length);

            if (this.useRandomPadding)
            {
                for (int i = 0; i < padd_len; i = i + 4)
                {
                    // don't waste calls to rnd.NextInt() (by using only 8 bits of the
                    // output). just believe me: even though we may write here up to 3
                    // bytes which won't be used, there is no "buffer overflow" (i.e.,
                    // arrayindexoutofbounds). the padding buffer is big enough =) (256
                    // bytes, and that is bigger than any current cipher block size + 64).
                    int r = this.rnd.NextInt();
                    unchecked
                    {
                        this.send_padding_buffer[i] = (byte)r;
                        this.send_padding_buffer[i + 1] = (byte)(r >> 8);
                        this.send_padding_buffer[i + 2] = (byte)(r >> 16);
                        this.send_padding_buffer[i + 3] = (byte)(r >> 24);
                    }
                }
            }
            else
            {
                // use zero padding for unencrypted traffic
                for (int i = 0; i < padd_len; i++)
                {
                    this.send_padding_buffer[i] = 0;
                }
                
                // Actually this code is paranoid: we never filled any
                // bytes into the padding buffer so far, therefore it should
                // consist of zeros only.
            }

            unchecked
            {
                this.send_packet_header_buffer[0] = (byte)((packet_len - 4) >> 24);
                this.send_packet_header_buffer[1] = (byte)((packet_len - 4) >> 16);
                this.send_packet_header_buffer[2] = (byte)((packet_len - 4) >> 8);
                this.send_packet_header_buffer[3] = (byte)(packet_len - 4);
                this.send_packet_header_buffer[4] = (byte)padd_len;
            }
            
            this.cos.Write(this.send_packet_header_buffer, 0, 5);
            this.cos.Write(message, offset, length);
            this.cos.Write(this.send_padding_buffer, 0, padd_len);

            if (this.send_mac != null)
            {
                this.send_mac.InitMac(this.send_seq_number);
                this.send_mac.Update(this.send_packet_header_buffer, 0, 5);
                this.send_mac.Update(message, offset, length);
                this.send_mac.Update(this.send_padding_buffer, 0, padd_len);

                this.send_mac.GetMac(this.send_mac_buffer, 0);
                this.cos.WritePlain(this.send_mac_buffer, 0, this.send_mac_buffer.Length);
            }

            this.cos.Flush();

            if (log.Enabled)
            {
                log.Log(90, "Sent " + Packet.GetMessageName(message[offset] & 0xff) + " " + length + " bytes payload");
            }

            this.send_seq_number++;
        }

        /// <summary>
        /// Peek to get the next message length.
        /// </summary>
        /// <returns>The message payload size (bytes).</returns>
        public int PeekNextMessageLength()
        {
            if (this.recv_packet_header_present == false)
            {
                this.cis.Read(this.recv_packet_header_buffer, 0, 5);
                this.recv_packet_header_present = true;
            }

            int packet_length = ((this.recv_packet_header_buffer[0] & 0xff) << 24)
                | ((this.recv_packet_header_buffer[1] & 0xff) << 16)
                | ((this.recv_packet_header_buffer[2] & 0xff) << 8)
                | (this.recv_packet_header_buffer[3] & 0xff);

            int padding_length = this.recv_packet_header_buffer[4] & 0xff;

            if (packet_length > 35000 || packet_length < 12)
            {
                throw new IOException("Illegal packet size! (" + packet_length + ")");
            }

            int payload_length = packet_length - padding_length - 1;

            if (payload_length < 0)
            {
                throw new IOException(
                    "Illegal padding_length in packet from remote (" + padding_length + ")");
            }

            return payload_length;
        }

        /// <summary>
        /// Receive a message/packet.
        /// </summary>
        /// <param name="buffer">The byte array that we put the packet in.</param>
        /// <param name="off">The offset where we begin placing the received bytes.</param>
        /// <param name="len">The number of bytes to receive.</param>
        /// <returns>The number of bytes actually received.</returns>
        public int ReceiveMessage(byte[] buffer, int off, int len)
        {
            if (this.recv_packet_header_present == false)
            {
                this.cis.Read(this.recv_packet_header_buffer, 0, 5);
            }
            else
            {
                this.recv_packet_header_present = false;
            }

            int packet_length = ((this.recv_packet_header_buffer[0] & 0xff) << 24)
                | ((this.recv_packet_header_buffer[1] & 0xff) << 16)
                | ((this.recv_packet_header_buffer[2] & 0xff) << 8)
                | (this.recv_packet_header_buffer[3] & 0xff);

            int padding_length = this.recv_packet_header_buffer[4] & 0xff;

            if (packet_length > 35000 || packet_length < 12)
            {
                throw new IOException("Illegal packet size! (" + packet_length + ")");
            }

            int payload_length = packet_length - padding_length - 1;

            if (payload_length < 0)
            {
                throw new IOException("Illegal padding_length in packet from remote (" + padding_length + ")");
            }

            if (payload_length >= len)
            {
                throw new IOException("Receive buffer too small (" + len + ", need " + payload_length + ")");
            }

            this.cis.Read(buffer, off, payload_length);
            this.cis.Read(this.recv_padding_buffer, 0, padding_length);

            if (this.recv_mac != null)
            {
                this.cis.ReadPlain(this.recv_mac_buffer, 0, this.recv_mac_buffer.Length);

                this.recv_mac.InitMac(this.recv_seq_number);
                this.recv_mac.Update(this.recv_packet_header_buffer, 0, 5);
                this.recv_mac.Update(buffer, off, payload_length);
                this.recv_mac.Update(this.recv_padding_buffer, 0, padding_length);
                this.recv_mac.GetMac(this.recv_mac_buffer_cmp, 0);

                for (int i = 0; i < this.recv_mac_buffer.Length; i++)
                {
                    if (this.recv_mac_buffer[i] != this.recv_mac_buffer_cmp[i])
                    {
                        throw new IOException("Remote sent corrupt MAC.");
                    }
                }
            }

            this.recv_seq_number++;

            if (log.Enabled)
            {
                log.Log(
                    90,
                    "Received " + Packet.GetMessageName(buffer[off] & 0xff) + " " + payload_length + " bytes payload");
            }

            return payload_length;
        }
        
        /// <summary>
        /// Implement the protected IDisposable interface method.
        /// </summary>
        /// <param name="disposing">True if we are disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.cis.Close();
                this.cos.Close();
            }
        }
    }
}
