// <copyright file="TransportManager.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> TransportManager
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Collections;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

using Eunomia.SSH2.Crypto;
using Eunomia.SSH2.Crypto.Cipher;
using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Log;
using Eunomia.SSH2.Packets;
using Eunomia.SSH2.Util;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.Transport
{
    /// <summary>
    /// Yes, the "standard" is a big mess. On one side, they say that arbitary channel
    /// packets are allowed during kex exchange, on the other side we need to blindly
    /// ignore the next _packet_ if the KEX guess was wrong. Where do we know from that
    /// the next packet is not a channel data packet? Yes, we could check if it is in
    /// the KEX range. But the standard says nothing about this. The OpenSSH guys
    /// block local "normal" traffic during KEX. That's fine - however, they assume
    /// that the other side is doing the same. During re-key, if they receive traffic
    /// other than KEX, they become horribly irritated and kill the connection. Since
    /// we are very likely going to communicate with OpenSSH servers, we have to play
    /// the same game - even though we could do better.
    /// <para>
    /// btw: having stdout and stderr on the same channel, with a shared window, is
    /// also a VERY good idea... =(.
    /// </para>
    /// </summary>
    public class TransportManager : IDisposable
    {
        private static Logger log = Logger.GetLogger(typeof(TransportManager));

        private string hostName;
        private int port;
        private Socket sock;
        private object connectionSemaphore;

        private bool flagKexOngoing = false;
        private bool connectionClosed = false;

        private Exception reasonClosedCause;

        private TransportConnection transportConnection;
        private KexManager keyExchangeManager;

        private ArrayList messageHandlers;
        private JavaThread receiveThread;

        private ArrayList connectionMonitors;
        private bool monitorsWereInformed = false;

        private ArrayList asynchronousQueue;
        private JavaThread asynchronousThread = null;

        /// <summary>
        /// Initializes a new instance of the TransportManager class.
        /// </summary>
        /// <param name="host">The host to connect to.</param>
        /// <param name="port">The port to connect to.</param>
        public TransportManager(string host, int port)
        {
            this.asynchronousQueue = new ArrayList(10);
            this.sock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.IP);
            this.connectionSemaphore = new object();
            this.messageHandlers = new ArrayList(10);
            this.connectionMonitors = new ArrayList(10);
            this.hostName = host;
            this.port = port;
        }
        
        /// <summary>
        /// Gets an ArrayList that contains connection monitors.
        /// </summary>
        public ArrayList ConnectionMonitors
        {
            get { return this.connectionMonitors; }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the Socket NoDelay flag should be set.
        /// </summary>
        public bool TcpNoDelay
        {
            get { return this.sock.NoDelay; }
            set { this.sock.NoDelay = value; }
        }

        /// <summary>
        /// Gets or sets the socket receive timeout.
        /// </summary>
        public int SoTimeout
        {
            get { return this.sock.ReceiveTimeout; }
            set { this.sock.ReceiveTimeout = value; }
        }

        /// <summary>
        /// Gets the reasonClosedClause exception.
        /// </summary>
        public Exception ReasonClosedCause
        {
            get
            {
                lock (this.connectionSemaphore)
                {
                    return this.reasonClosedCause;
                }
            }
        }

        /// <summary>
        /// Gets the session ID.
        /// </summary>
        public byte[] SessionIdentifier
        {
            get
            {
                return this.keyExchangeManager.SessionId;
            }
        }

        /// <summary>
        /// Gets the packet overhead estimate.
        /// </summary>
        public int PacketOverheadEstimate
        {
            get { return this.transportConnection.PacketOverheadEstimate; }
        }
        
        /// <summary>
        /// Implement the public IDisposable interface method.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Gets the connection info based on the kex number.
        /// </summary>
        /// <param name="kexNumber">The kex number.</param>
        /// <returns>Returns the ConnectionInfo.</returns>
        public ConnectionInfo GetConnectionInfo(int kexNumber)
        {
            return this.keyExchangeManager.GetOrWaitForConnectionInfo(kexNumber);
        }

        /// <summary>
        /// Close the connection.
        /// </summary>
        /// <param name="cause">The reason the connection was closed.</param>
        /// <param name="useDisconnectPacket">Whether to send a disconnect packet.</param>
        //// TODO: Be more specific about the exceptions that we catch.
        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes",
                         Justification = "This will be a future TODO")]
        public void Close(Exception cause, bool useDisconnectPacket)
        {
            if (useDisconnectPacket == false)
            {
                // OK, hard shutdown - do not aquire the semaphore,
                // perhaps somebody is inside (and waits until the remote
                // side is ready to accept new data).
                try
                {
                    this.sock.Close();
                }
                catch (IOException)
                {
                }

                // OK, whoever tried to send data, should now agree that
                // there is no point in further waiting =)
                // It is safe now to aquire the semaphore.
            }

            lock (this.connectionSemaphore)
            {
                if (this.connectionClosed == false)
                {
                    if (useDisconnectPacket == true)
                    {
                        try
                        {
                            byte[] msg = new PacketDisconnect(
                                Packet.SSHDisconnectByApplication, cause.Message, string.Empty).Payload;
                            if (this.transportConnection != null)
                            {
                                this.transportConnection.SendMessage(msg);
                            }
                        }
                        catch (IOException)
                        {
                        }

                        try
                        {
                            this.sock.Close();
                        }
                        catch (IOException)
                        {
                        }
                    }

                    this.connectionClosed = true;
                    this.reasonClosedCause = cause; /* may be null */
                }

                Monitor.PulseAll(this.connectionSemaphore);
            }

            // No check if we need to inform the monitors
            ArrayList monitors = new ArrayList();

            lock (this)
            {
                // Short term lock to protect "connectionMonitors"
                // and "monitorsWereInformed"
                // (they may be modified concurrently)
                if (this.monitorsWereInformed == false)
                {
                    this.monitorsWereInformed = true;
                    monitors = this.connectionMonitors;
                }
            }

            if (monitors != null)
            {
                for (int i = 0; i < monitors.Count; i++)
                {
                    try
                    {
                        IConnectionMonitor cmon = (IConnectionMonitor)monitors[i];
                        cmon.ConnectionLost(this.reasonClosedCause);
                    }
                    catch (Exception)
                    {
                    }
                }
            }
        }

        /// <summary>
        /// Initialize the TransportManager.
        /// </summary>
        /// <param name="cryptoWishList">The CryptoWishList.</param>
        /// <param name="verifier">The host key verifier to use.</param>
        /// <param name="dhgex">The DHGex paramters to use.</param>
        /// <param name="connectTimeout">
        /// The connection timeout.
        /// In .net there is no connection timeout on sockets.  This is not used.
        /// </param>
        /// <param name="random">A SecureRandom instance.</param>
        /// <param name="proxyData">Information about the proxy to use.</param>
        [SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "connectTimeout",
                         Justification = "Left in to retain some compatibility with ganymed.")]
        public void Initialize(
            CryptoWishList cryptoWishList,
            IServerHostkeyVerifier verifier,
            DHGexParameters dhgex,
            int connectTimeout,
            SecureRandom random,
            ProxyData proxyData)
        {
            // First, establish the TCP connection to the SSH-2 server
            // Note: there is no built-in connection timeout on .Net sockets.
            this.EstablishConnection(proxyData);

            // Parse the server line and say hello - important: this information is later needed for the
            // key exchange (to stop man-in-the-middle attacks) - that is why we wrap it into an object
            // for later use.
            NetworkStream stream = new NetworkStream(this.sock);

            ClientServerHello csh = new ClientServerHello(stream, stream);
            this.transportConnection = new TransportConnection(stream, stream, random);

            this.keyExchangeManager = new KexManager(this, csh, cryptoWishList, this.hostName, this.port, verifier, random);
            this.keyExchangeManager.InitiateKEX(cryptoWishList, dhgex);

            this.receiveThread = new JavaThread(new ThreadStart(new ReceiveClass(this).Run));
            this.receiveThread.IsBackground = true;
            this.receiveThread.Start();
        }

        /// <summary>
        /// This is needed to make FxCop happy and to be used as a substitute for
        /// writing to the ConnectionMonitors property which is now read only.
        /// </summary>
        /// <param name="newConnectionMonitors">
        /// The connection monitors that replace any connection monitors that we had.
        /// </param>
        public void ReplaceConnectionMonitors(ArrayList newConnectionMonitors)
        {
            this.connectionMonitors.Clear();
            this.connectionMonitors.AddRange(newConnectionMonitors);
        }
        
        /// <summary>
        /// Register a message handler.
        /// </summary>
        /// <param name="mh">The message handle to register.</param>
        /// <param name="low">The lowest message type to process.</param>
        /// <param name="high">The highest message type to process.</param>
        public void RegisterMessageHandler(MessageHandler mh, int low, int high)
        {
            HandlerEntry he = new HandlerEntry();
            he.MessageHandler = mh;
            he.Low = low;
            he.High = high;

            lock (this.messageHandlers.SyncRoot)
            {
                this.messageHandlers.Add(he);
            }
        }

        /// <summary>
        /// Remove a message handler.
        /// </summary>
        /// <param name="mh">The message handler to unregister.</param>
        /// <param name="low">The lowest message type to process.</param>
        /// <param name="high">The highest message type to process.</param>
        public void RemoveMessageHandler(MessageHandler mh, int low, int high)
        {
            lock (this.messageHandlers.SyncRoot)
            {
                for (int i = 0; i < this.messageHandlers.Count; i++)
                {
                    HandlerEntry he = (HandlerEntry)this.messageHandlers[i];
                    if ((he.MessageHandler == mh) && (he.Low == low) && (he.High == high))
                    {
                        this.messageHandlers.RemoveAt(i);
                        break;
                    }
                }
            }
        }

        /// <summary>
        /// Send kex exchange message.
        /// </summary>
        /// <param name="msg">The kex message to be sent.</param>
        public void SendKexMessage(byte[] msg)
        {
            lock (this.connectionSemaphore)
            {
                if (this.connectionClosed)
                {
                    throw new IOException("Sorry, this connection is closed.");
                }

                this.flagKexOngoing = true;

                try
                {
                    this.transportConnection.SendMessage(msg);
                }
                catch (IOException e)
                {
                    this.Close(e, false);
                    throw;
                }
            }
        }

        /// <summary>
        /// Key exchange finished.
        /// </summary>
        public void KexFinished()
        {
            lock (this.connectionSemaphore)
            {
                this.flagKexOngoing = false;
                Monitor.PulseAll(this.connectionSemaphore);
            }
        }

        /// <summary>
        /// Force a key exchange.
        /// </summary>
        /// <param name="cryptoWishList">The crypto wish list.</param>
        /// <param name="dhgex">The DHGex paramters.</param>
        public void ForceKeyExchange(CryptoWishList cryptoWishList, DHGexParameters dhgex)
        {
            this.keyExchangeManager.InitiateKEX(cryptoWishList, dhgex);
        }

        /// <summary>
        /// Change the receive cipher.
        /// </summary>
        /// <param name="bc">The block cipher to use.</param>
        /// <param name="mac">The MAC to use.</param>
        public void ChangeReceiveCipher(IBlockCipher bc, MAC mac)
        {
            this.transportConnection.ChangeReceiveCipher(bc, mac);
        }

        /// <summary>
        /// Change the send cipher.
        /// </summary>
        /// <param name="bc">The block cipher to use.</param>
        /// <param name="mac">The MAC to use.</param>
        public void ChangeSendCipher(IBlockCipher bc, MAC mac)
        {
            this.transportConnection.ChangeSendCipher(bc, mac);
        }

        /// <summary>
        /// Send a message asychronously.
        /// </summary>
        /// <param name="msg">The message to send.</param>
        public void SendAsynchronousMessage(byte[] msg)
        {
            lock (this.asynchronousQueue)
            {
                this.asynchronousQueue.Add(msg);

                // This limit should be flexible enough. We need this, otherwise the peer
                // can flood us with global requests (and other stuff where we have to reply
                // with an asynchronous message) and (if the server just sends data and does not
                // read what we send) this will probably put us in a low memory situation
                // (our send queue would grow and grow and...)
                if (this.asynchronousQueue.Count > 100)
                {
                    throw new IOException("Error: the peer is not consuming our asynchronous replies.");
                }

                // Check if we have an asynchronous sending thread
                if (this.asynchronousThread == null)
                {
                    this.asynchronousThread = new AsynchronousWorker(this);
                    this.asynchronousThread.IsBackground = true;
                    this.asynchronousThread.Start();

                    // The thread will stop after 2 seconds of inactivity (i.e., empty queue)
                }
            }
        }

        /// <summary>
        /// Send a message.
        /// </summary>
        /// <param name="msg">The message to send.</param>
        public void SendMessage(byte[] msg)
        {
            if (JavaThread.Current() == this.receiveThread)
            {
                throw new IOException("Assertion error: sendMessage may never be invoked by the receiver thread!");
            }

            bool wasInterrupted = false;

            try
            {
                lock (this.connectionSemaphore)
                {
                    while (true)
                    {
                        if (this.connectionClosed)
                        {
                            throw new IOException("Sorry, this connection is closed.");
                        }

                        if (this.flagKexOngoing == false)
                        {
                            break;
                        }

                        try
                        {
                            Monitor.Wait(this.connectionSemaphore);
                        }
                        catch (ThreadInterruptedException)
                        {
                            wasInterrupted = true;
                        }
                    }

                    try
                    {
                        this.transportConnection.SendMessage(msg);
                    }
                    catch (IOException e)
                    {
                        this.Close(e, false);
                        throw;
                    }
                }
            }
            finally
            {
                if (wasInterrupted)
                {
                    JavaThread.Current().Interrupt();
                }
            }
        }

        /// <summary>
        /// ReceiveLoop to process received messages.
        /// </summary>
        public void ReceiveLoop()
        {
            byte[] msg = new byte[35000];

            while (true)
            {
                int msgLength = this.transportConnection.ReceiveMessage(msg, 0, msg.Length);

                int type = msg[0] & 0xff;

                if (type == Packet.SSHMessageIgnore)
                {
                    continue;
                }

                if (type == Packet.SSHMessageDebug)
                {
                    if (log.Enabled)
                    {
                        TypesReader tr = new TypesReader(msg, 0, msgLength);
                        tr.ReadByte();
                        tr.ReadBoolean();
                        StringBuilder debugMessageBuffer = new StringBuilder();
                        debugMessageBuffer.Append(tr.ReadString("UTF-8"));
                        char[] charMessageBuffer = debugMessageBuffer.ToString().ToCharArray();

                        for (int i = 0; i < charMessageBuffer.Length; i++)
                        {
                            char c = charMessageBuffer[i];

                            if ((c >= 32) && (c <= 126))
                            {
                                continue;
                            }
                            
                            charMessageBuffer[i] = '\uFFFD';
                        }

                        log.Log(50, "DEBUG Message from remote: '" + charMessageBuffer.ToString() + "'");
                    }
                    
                    continue;
                }

                if (type == Packet.SSHMessageUnimplemented)
                {
                    throw new IOException("Peer sent UNIMPLEMENTED message, that should not happen.");
                }

                if (type == Packet.SSHMessageDisconnect)
                {
                    TypesReader tr = new TypesReader(msg, 0, msgLength);
                    tr.ReadByte();
                    int reason_code = (int)tr.ReadUINT32();
                    StringBuilder reasonBuffer = new StringBuilder();
                    reasonBuffer.Append(tr.ReadString("UTF-8"));

                    // Do not get fooled by servers that send abnormal long error
                    // messages
                    if (reasonBuffer.Length > 255)
                    {
                        reasonBuffer.Capacity = 255;
                        reasonBuffer[254] = '.';
                        reasonBuffer[253] = '.';
                        reasonBuffer[252] = '.';
                    }

                    // Also, check that the server did not send characters that may
                    // screw up the receiver -> restrict to reasonable US-ASCII
                    // subset -> "printable characters" (ASCII 32 - 126). Replace
                    // all others with 0xFFFD (UNICODE replacement character).
                    for (int i = 0; i < reasonBuffer.Length; i++)
                    {
                        char c = reasonBuffer[i];

                        if ((c >= 32) && (c <= 126))
                        {
                            continue;
                        }
                        
                        reasonBuffer[i] = '\uFFFD';
                    }

                    throw new IOException("Peer sent DISCONNECT message (reason code " + reason_code + "): "
                                          + reasonBuffer.ToString());
                }

                // Is it a KEX Packet?
                if ((type == Packet.SSHMessageKEXInitialize) || (type == Packet.SSHMessageNewKeys)
                    || ((type >= 30) && (type <= 49)))
                {
                    this.keyExchangeManager.HandleMessage(msg, msgLength);
                    continue;
                }

                MessageHandler mh = null;

                for (int i = 0; i < this.messageHandlers.Count; i++)
                {
                    HandlerEntry he = (HandlerEntry)this.messageHandlers[i];
                    if ((he.Low <= type) && (type <= he.High))
                    {
                        mh = he.MessageHandler;
                        break;
                    }
                }

                if (mh == null)
                {
                    throw new IOException("Unexpected SSH message (type " + type + ")");
                }

                mh.HandleMessage(msg, msgLength);
            }
        }
        
        /// <summary>
        /// Implement the protected IDisposable interface method.
        /// </summary>
        /// <param name="disposing">True if we are disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.sock.Close();
                this.transportConnection.Dispose();
            }
        }

        /// <summary>
        /// Gets an IPAddress from a hostName or ip address in a string.
        /// </summary>
        /// <param name="host">The ip address to resolve.</param>
        /// <returns>The IPAddress.</returns>
        private static IPAddress CreateIPAddress(string host)
        {
            return Dns.GetHostEntry(host).AddressList[0];
        }

        /// <summary>
        /// Establish a connection.
        /// </summary>
        /// <param name="proxyData">The proxy to use.</param>
        /// <remarks>
        /// Because .net sockets do not allow for a connection timeout, this method
        /// does not match the signature of the ganymed establishConnection method...
        /// </remarks>
        [SuppressMessage("Microsoft.Performance", "CA1800:DoNotCastUnnecessarily",
                         Justification = "FxCop is confused.  The casting is appropriate.")]
        private void EstablishConnection(ProxyData proxyData)
        {
            if (proxyData == null)
            {
                IPAddress addr = TransportManager.CreateIPAddress(this.hostName);
                this.sock.ReceiveTimeout = 0;
                this.sock.Connect(addr, this.port);
                return;
            }

            if (proxyData is HTTPProxyData)
            {
                HTTPProxyData pd = (HTTPProxyData)proxyData;

                // At the moment, we only support HTTP proxies
                IPAddress addr = TransportManager.CreateIPAddress(pd.ProxyHost);
                this.sock.ReceiveTimeout = 0;
                this.sock.Connect(addr, pd.ProxyPort);

                // OK, now tell the proxy where we actually want to connect to
                StringBuilder sb = new StringBuilder();

                sb.Append("CONNECT ");
                sb.Append(this.hostName);
                sb.Append(':');
                sb.Append(this.port);
                sb.Append(" HTTP/1.0\r\n");

                if ((pd.ProxyUser != null) && (pd.ProxyPass != null))
                {
                    string credentials = pd.ProxyUser + ":" + pd.ProxyPass;

                    char[] encoded = Convert.ToBase64String(StringEncoder.GetBytes(credentials)).ToCharArray();
                    sb.Append("Proxy-Authorization: Basic ");
                    sb.Append(encoded);
                    sb.Append("\r\n");
                }

                if (pd.RequestHeaderLines != null)
                {
                    for (int i = 0; i < pd.RequestHeaderLines.Length; i++)
                    {
                        if (pd.RequestHeaderLines[i] != null)
                        {
                            sb.Append(pd.RequestHeaderLines[i]);
                            sb.Append("\r\n");
                        }
                    }
                }

                sb.Append("\r\n");

                NetworkStream stream = new NetworkStream(this.sock);
                byte[] outbytes = StringEncoder.GetBytes(sb.ToString());

                stream.Write(outbytes, 0, outbytes.Length);
                stream.Flush();

                // Now parse the HTTP response
                byte[] buffer = new byte[1024];

                int len = ClientServerHello.ReadLineRN(stream, buffer);

                string httpReponse = StringEncoder.GetString(buffer, 0, len);

                if (httpReponse.StartsWith("HTTP/", StringComparison.Ordinal) == false)
                {
                    throw new IOException("The proxy did not send back a valid HTTP response.");
                }

                // "HTTP/1.X XYZ X" => 14 characters minimum
                if ((httpReponse.Length < 14) || (httpReponse[8] != ' ') || (httpReponse[12] != ' '))
                {
                    throw new IOException("The proxy did not send back a valid HTTP response.");
                }

                int errorCode = 0;

                try
                {
                    errorCode = Int32.Parse(httpReponse.Substring(9, 12), CultureInfo.InvariantCulture);
                }
                catch (FormatException)
                {
                    throw new IOException("The proxy did not send back a valid HTTP response.");
                }

                if ((errorCode < 0) || (errorCode > 999))
                {
                    throw new IOException("The proxy did not send back a valid HTTP response.");
                }

                if (errorCode != 200)
                {
                    throw new HttpProxyException(httpReponse.Substring(13), errorCode);
                }

                // OK, read until empty line
                while (true)
                {
                    len = ClientServerHello.ReadLineRN(stream, buffer);
                    if (len == 0)
                    {
                        break;
                    }
                }
                
                return;
            }

            throw new IOException("Unsupported ProxyData");
        }

        /// <summary>
        /// Class to store information about a message handler.
        /// </summary>
        private class HandlerEntry
        {
            private MessageHandler messageHandler;
            private int low;
            private int high;
            
            /// <summary>
            /// Gets or sets the MessageHandler.
            /// </summary>
            public MessageHandler MessageHandler
            {
                get { return this.messageHandler; }
                set { this.messageHandler = value; }
            }
            
            /// <summary>
            /// Gets or sets the (numerically) lowest message type to process.
            /// </summary>
            public int Low
            {
                get { return this.low; }
                set { this.low = value; }
            }
            
            /// <summary>
            /// Gets or sets the (numerically) highest message type to process.
            /// </summary>
            public int High
            {
                get { return this.high; }
                set { this.high = value; }
            }
        }

        /// <summary>
        /// Class used to run the ReceiveLoop in a seperate thread.
        /// </summary>
        private class ReceiveClass : JavaThread
        {
            private TransportManager tm;

            /// <summary>
            /// Initializes a new instance of the ReceiveClass class.
            /// </summary>
            /// <param name="tm">The transportation manager.</param>
            public ReceiveClass(TransportManager tm)
            {
                this.tm = tm;
            }

            /// <summary>
            /// Run method that calls the ReceiveLoop.
            /// </summary>
            //// TODO: Be more specific about the exceptions that we catch.
            [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes",
                            Justification = "This will be a future TODO")]
            public override void Run()
            {
                try
                {
                    this.tm.ReceiveLoop();
                }
                catch (IOException e)
                {
                    this.tm.Close(e, false);
                    if (log.Enabled)
                    {
                        log.Log(10, "Receive thread: error in receiveLoop: " + e.Message);
                    }
                }

                if (log.Enabled)
                {
                    log.Log(50, "Receive thread: back from receiveLoop");
                }

                // Tell all handlers that it is time to say goodbye
                if (this.tm.keyExchangeManager != null)
                {
                    try
                    {
                        this.tm.keyExchangeManager.HandleMessage(null, 0);
                    }
                    catch (IOException)
                    {
                    }
                }

                for (int i = 0; i < this.tm.messageHandlers.Count; i++)
                {
                    HandlerEntry he = (HandlerEntry)this.tm.messageHandlers[i];
                    try
                    {
                        he.MessageHandler.HandleMessage(null, 0);
                    }
                    catch (Exception)
                    {
                    }
                }
            }
        }
        
        /// <summary>
        /// AsynchronousWorker class used for sending messages Asynchronously.
        /// </summary>
        private class AsynchronousWorker : JavaThread
        {
            private TransportManager tm;

            /// <summary>
            /// Initializes a new instance of the AsynchronousWorker class.
            /// </summary>
            /// <param name="tm">The transportation manager.</param>
            public AsynchronousWorker(TransportManager tm)
            {
                this.tm = tm;
            }

            /// <summary>
            /// Process the asynchronousQueue.
            /// </summary>
            public override void Run()
            {
                while (true)
                {
                    byte[] msg = null;

                    lock (this.tm.asynchronousQueue)
                    {
                        if (this.tm.asynchronousQueue.Count == 0)
                        {
                            // After the queue is empty for about 2 seconds, stop this thread
                            try
                            {
                                Monitor.Wait(this.tm.asynchronousQueue, TimeSpan.FromMilliseconds(2000));
                            }
                            catch (ThreadInterruptedException)
                            {
                            }

                            if (this.tm.asynchronousQueue.Count == 0)
                            {
                                this.tm.asynchronousThread = null;
                                return;
                            }
                        }

                        object temp;
                        temp = this.tm.asynchronousQueue[0];
                        this.tm.asynchronousQueue.RemoveAt(0);
                        msg = (byte[])temp;
                    }

                    // The following invocation may throw an IOException.
                    // There is no point in handling it - it simply means
                    // that the connection has a problem and we should stop
                    // sending asynchronously messages. We do not need to signal that
                    // we have exited (asynchronousThread = null): further
                    // messages in the queue cannot be sent by this or any
                    // other thread.
                    // Other threads will sooner or later (when receiving or
                    // sending the next message) get the same IOException and
                    // get to the same conclusion.
                    try
                    {
                        this.tm.SendMessage(msg);
                    }
                    catch (IOException)
                    {
                        return;
                    }
                }
            }
        }
    }
}
