// <copyright file="TimeoutService.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> TimeoutService
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Collections;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Threading;

using Eunomia.SSH2.Log;

namespace Eunomia.SSH2.Util
{
    /// <summary>
    /// TimeoutService. Here you can register a timeout.
    /// <para>
    /// Implemented having large scale programs in mind: if you open many concurrent SSH connections
    /// that rely on timeouts, then there will be only one timeout thread. Once all timeouts
    /// have expired/are cancelled, the thread will (sooner or later) exit.
    /// Only after new timeouts arrive a new thread (singleton) will be instantiated.</para>
    /// </summary>
    public static class TimeoutService
    {
        private static readonly Logger log = Logger.GetLogger(typeof(TimeoutService));

        // The list object is also used for locking purposes
        private static readonly ArrayList todolist = new ArrayList();
        private static JavaThread timeoutThread = null;

        /// <summary>
        /// Adds a timeout handler.  It is assumed that the passed handler will not execute for a long time.
        /// </summary>
        /// <param name="runtime">The runtime (or timeout) in milliseconds.</param>
        /// <param name="handler">The handler to run.</param>
        /// <returns>A TimeoutToken that has the runtime and handler stored in it.</returns>
        public static TimeoutToken AddTimeoutHandler(long runtime, IThreadRunnable handler)
        {
            TimeoutToken token = new TimeoutToken(runtime, handler);

            lock (todolist)
            {
                todolist.Add(token);
                todolist.Sort();

                if (timeoutThread != null)
                {
                    timeoutThread.Interrupt();
                }
                else
                {
                    timeoutThread = new TimeoutThread();
                    timeoutThread.IsBackground = true;
                    timeoutThread.Start();
                }
            }

            return token;
        }

        /// <summary>
        /// Removes a TimeoutHandler from our service.
        /// </summary>
        /// <param name="token">The token of the service that we want to remove.</param>
        public static void CancelTimeoutHandler(TimeoutToken token)
        {
            lock (todolist)
            {
                todolist.Remove(token);

                if (timeoutThread != null)
                {
                    timeoutThread.Interrupt();
                }
            }
        }

        /// <summary>
        /// The TimeoutThread used to implement the TimeoutService.
        /// </summary>
        private class TimeoutThread : JavaThread
        {
            /// <summary>
            /// Override the Run method of JavaThread.
            /// </summary>
            //// TODO: Be more specific about the exceptions that we catch.
            [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes",
                            Justification = "This will be a future TODO")]
            public override void Run()
            {
                lock (TimeoutService.todolist.SyncRoot)
                {
                    while (true)
                    {
                        if (TimeoutService.todolist.Count == 0)
                        {
                            TimeoutService.timeoutThread = null;
                            return;
                        }

                        // divide ticks by 10000 to get milliseconds
                        long now = DateTime.Now.Ticks / 10000;

                        TimeoutToken tt = (TimeoutToken)TimeoutService.todolist[0];

                        if (tt.Runtime > now)
                        {
                            // Not ready yet, sleep a little bit
                            try
                            {
                                Monitor.Wait(TimeoutService.todolist, TimeSpan.FromMilliseconds(tt.Runtime - now));
                            }
                            catch (ThreadInterruptedException)
                            {
                            }

                            // We cannot simply go on, since it could be that the token
                            // was removed (cancelled) or another one has been inserted in
                            // the meantime.
                            continue;
                        }

                        TimeoutService.todolist.RemoveAt(0);

                        try
                        {
                            tt.Handler.Run();
                        }
                        catch (Exception e)
                        {
                            StringWriter sw = new StringWriter(CultureInfo.InvariantCulture);
                            TimeoutService.log.Log(20, "Exeception in Timeout handler:" + e.Message + "(" + sw.ToString() + ")");
                        }
                    }
                }
            }
        }
    }
}
