﻿// <copyright file="Basic.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> Basic
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.5.0, 12/17/2010
// </version>

using System;
using System.Collections;
using System.Globalization;
using System.IO;

namespace Eunomia.SSH2.Examples
{
    /// <summary>
    /// Program class...just a basic console appliction.
    /// </summary>
    public static class Program
    {
        private static string hostName;
        private static string userName;
        private static string password;

        /// <summary>
        /// The one and only static Main method.
        /// </summary>
        public static void Main()
        {
            Console.Write("Enter host name: ");
            hostName = Console.ReadLine();
            
            Console.Write("Enter username: ");
            userName = Console.ReadLine();
            
            Console.Write("Enter password: ");
            password = Console.ReadLine();
            
            try
            {
                TrySSH();
            }
            catch (IOException e)
            {
                Console.WriteLine(e.StackTrace);
            }

            try
            {
                TrySFTP("BasicTest.bin");
            }
            catch (IOException e)
            {
                Console.WriteLine(e.StackTrace);
            }

            try
            {
                TrySCP("BasicTest.bin");
            }
            catch (IOException e)
            {
                Console.WriteLine(e.StackTrace);
            }

            try
            {
                TryLocalForward();
            }
            catch (IOException e)
            {
                Console.WriteLine(e.StackTrace);
            }

            Console.Write("Hit a key to close this program: ");
            Console.ReadKey();
        }
        
        /// <summary>
        /// Try using ssh.
        /// </summary>
        public static void TrySSH()
        {
            Console.WriteLine("===============> Now testing ssh <====================");
            Connection conn = new Connection(hostName);
            conn.Connect();
            bool isAuthenticated = conn.AuthenticateWithPassword(userName, password);
            if (isAuthenticated == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }

            Session sess = conn.OpenSession();
            sess.ExecCommand("uname -a && date");
            Console.WriteLine("Here is some information about the remote host:");

            // This basic example does not handle stderr, which is sometimes dangerous
            // (please read the FAQ).
            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            while (true)
            {
                string line = br.ReadLine();
                if (line == null)
                {
                    break;
                }

                Console.WriteLine(line);
            }

            // Show exit status, if available (otherwise "null")
            Console.WriteLine("ExitCode: " + sess.ExitStatus);

            // Close this session
            sess.Close();

            // Close the connection
            conn.Close();
        }

        /// <summary>
        /// Try SFTP.  We test ls, and sending and receiving a file.
        /// </summary>
        /// <param name="fileName">The file name to use for the test file.</param>
        public static void TrySFTP(string fileName)
        {
            InitFileTransfers(fileName);
            Console.WriteLine("===============> Now testing sftp <====================");
            Connection conn = new Connection(hostName);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }

            SFTPv3Client client = new SFTPv3Client(conn);
            ArrayList files = client.Ls(".");
            Console.WriteLine(files.Count.ToString(CultureInfo.InvariantCulture));
            SFTPv3DirectoryEntry de = (SFTPv3DirectoryEntry)files[files.Count - 1];
            Console.WriteLine("Last file name: " + de.FileName);

            // Read our test file and transfer it to the ssh server.
            Console.WriteLine("Sending file to server with sftp.");
            SFTPv3FileHandle fh = client.CreateFileTruncate(fileName);
            FileStream localFile = new FileStream(fileName, FileMode.Open);
            byte[] buffer = new byte[32768];
            int offset = 0;
            int count = 0;
            while ((count = localFile.Read(buffer, 0, buffer.Length)) > 0)
            {
                client.Write(fh, offset, buffer, 0, count);
                Console.Write(".");
                offset += count;
            }
            
            localFile.Close();
            Console.WriteLine(string.Empty);
            
            // Read the file back from the ssh server and put it in our current working directory.
            Console.WriteLine("Receiving file from server with sftp.");
            fh = client.OpenFileRO(fileName);
            localFile = new FileStream(fileName, FileMode.OpenOrCreate);
            offset = 0;
            count = 0;
            while (!fh.Closed)
            {
                count = client.Read(fh, offset, buffer, 0, 32768);
                if (count < 0)
                {
                    break;
                }

                localFile.Write(buffer, 0, count);
                Console.Write(".");
                offset = offset + count;
            }
            
            localFile.Close();
            Console.WriteLine(string.Empty);
        }
        
        /// <summary>
        /// Try SCP.  We try putting and getting a file.
        /// </summary>
        /// <param name="fileName">The file name to use for the test file.</param>
        public static void TrySCP(string fileName)
        {
            InitFileTransfers(fileName);
            Console.WriteLine("===============> Now testing scp <====================");
            string cwd = Directory.GetCurrentDirectory();
            Connection conn = new Connection(hostName);
            Console.WriteLine("current working directory: " + cwd);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }
            
            SCPClient client = new SCPClient(conn);
            Console.WriteLine("Putting the file to the server");
            client.Put(fileName, ".");
            Console.WriteLine("Getting the file from the server");
            client.Get(fileName, cwd);
            conn.Close();
        }

        /// <summary>
        /// Try a local forward.
        /// </summary>
        public static void TryLocalForward()
        {
            Console.WriteLine("===============> Now testing forwarding <====================");
            Connection conn = new Connection(hostName);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }
            
            conn.CreateLocalPortForwarder(2200, hostName, 22);

            Console.Write("Hit a key after testing port forward: ");
            Console.ReadKey();
            Console.WriteLine();
            
            conn.Close();
        }

        /// <summary>
        /// Method called by SFTP and SCP tests.  Creates a random 4MB file to
        /// test file transfers to/from an ssh server.
        /// <param name="fileName">The file name to use for the test file.</param>
        private static void InitFileTransfers(string fileName)
        {
            if (File.Exists(fileName))
            {
                File.Delete(fileName);
            }
            
            if (File.Exists("org" + fileName))
            {
                File.Delete("org" + fileName);
            }

            // Create a test file (4MB of random data)
            System.Random rnd = new System.Random();
            FileStream localFile = new FileStream(fileName, FileMode.Create, FileAccess.Write);
            byte[] buffer = new byte[1024];
            for (int i = 0; i < 1024; i++)
            {
                for (int j = 0; j < 4; j++)
                {
                    rnd.NextBytes(buffer);
                    localFile.Write(buffer, 0, 1024);
                }
            }
            
            localFile.Close();
            
            // save a copy of the original file.
            File.Copy(fileName, "org" + fileName);
        }
    }
}